#!/usr/bin/python
# -*- encoding: utf-8; py-indent-offset: 4 -*-
# .------------------------------------------------------------------------.
# |                ____ _               _        __  __ _  __              |
# |               / ___| |__   ___  ___| | __   |  \/  | |/ /              |
# |              | |   | '_ \ / _ \/ __| |/ /   | |\/| | ' /               |
# |              | |___| | | |  __/ (__|   <    | |  | | . \               |
# |               \____|_| |_|\___|\___|_|\_\___|_|  |_|_|\_\              |
# |                                        |_____|                         |
# |             _____       _                       _                      |
# |            | ____|_ __ | |_ ___ _ __ _ __  _ __(_)___  ___             |
# |            |  _| | '_ \| __/ _ \ '__| '_ \| '__| / __|/ _ \            |
# |            | |___| | | | ||  __/ |  | |_) | |  | \__ \  __/            |
# |            |_____|_| |_|\__\___|_|  | .__/|_|  |_|___/\___|            |
# |                                     |_|                                |
# |                     _____    _ _ _   _                                 |
# |                    | ____|__| (_) |_(_) ___  _ __                      |
# |                    |  _| / _` | | __| |/ _ \| '_ \                     |
# |                    | |__| (_| | | |_| | (_) | | | |                    |
# |                    |_____\__,_|_|\__|_|\___/|_| |_|                    |
# |                                                                        |
# | mathias-kettner.com                                 mathias-kettner.de |
# '------------------------------------------------------------------------'
#  This file is part of the Check_MK Enterprise Edition (CEE).
#  Copyright by Mathias Kettner and Mathias Kettner GmbH.  All rights reserved.
#
#  Distributed under the Check_MK Enterprise License.
#
#  You should have  received  a copy of the Check_MK Enterprise License
#  along with Check_MK. If not, email to mk@mathias-kettner.de
#  or write to the postal address provided at www.mathias-kettner.de

import shutil
import tempfile

try:
    from hashlib import md5
except ImportError:
    from md5 import md5 # deprecated with python 2.5

# Some constants
all_opsyses = [
    "linux_rpm",
    "linux_deb",
    "linux_tgz",
    "aix_tgz",
    "windows_msi",
]
baked_agents_dir = var_dir + "/agents"
agent_file_header = "# Created by Check_MK Agent Bakery.\n" \
                    "# This file is managed via WATO, do not edit manually or you \n" \
                    "# lose your changes next time when you update the agent.\n\n"

# Global variables that are filled by the bake plugins
# Not all rulesets have the default match type "first"
bakery_info             = {}

def load_baking_plugins():
    for path in [ agents_dir + "/bakery", omd_root + "/local/share/check_mk/agents/bakery" ]:
        if os.path.exists(path):
            for entry in os.listdir(path):
                if not entry.startswith("."):
                    execfile(path + "/" + entry, globals(), globals())

def do_bake_agents(args = None):
    load_baking_plugins()
    global g_baked_paths
    g_baked_paths = set([])
    global g_used_hashes
    g_used_hashes = set([])

    if not args:
        hostnames = all_configured_realhosts()
    else:
        hostnames = args

    # Make sure tmp_dir is existing. This can be missing directly
    # after an omd update
    if not os.path.exists(tmp_dir):
        os.makedirs(tmp_dir)

    # Create vanilla agent
    if opt_verbose:
        sys.stdout.write("Vanilla agent...")
    bake_agent(None)
    if opt_verbose:
        sys.stdout.write("OK\n")

    # Create generic agent
    if opt_verbose:
        sys.stdout.write("Agent for generic unknown host...")
    bake_agent(True)
    if opt_verbose:
        sys.stdout.write("OK\n")

    # Create host specific agents
    for hostname in hostnames:
        if opt_verbose:
            sys.stdout.write("%s..." % hostname)
            sys.stdout.flush()

        if not is_tcp_host(hostname):
            if opt_verbose:
                sys.stdout.write("does not use agent, skipping.\n")
            continue

        try:
            bake_agent(hostname)
            if opt_verbose:
                sys.stdout.write("OK\n")
        except Exception, e:
            if opt_debug:
                raise
            raise MKGeneralException("Error creating agent for %s: %s." %
                            (format_bakery_hostname_for_humans(hostname), e))

    # Remove any hashes that are not referred to any longer. But only if we have baked
    # for *all* hosts.
    if not args:
        cleanup_orphaned_packages()


def supports_msi_building():
    for path in os.environ["PATH"].split(":"):
        if os.path.exists(path + "/msi-update"):
            return True
    return False


# hostname may be None -> vanilla agent or True -> generic host
def bake_agent(hostname):
    # Collect configuration for that host. Start with setting some default
    # values. These can be overridden with rules. Note the ruleset agent_ports
    # has originally been created for the configuration of the Check_MK engine.
    # We simply use the same configuration here: why should the agent's port
    # and the configured one differ?

    agconf = {
        "check_mk_version" : check_mk_version,
        "agent_port"       : 6556,
        "is_ipv6_primary"  : False,
        "agent_paths"      : {
            "bin"    : "/usr/bin",
            "config" : "/etc/check_mk",
            "lib"    : "/usr/lib/check_mk_agent",
            "var"    : "/var/lib/check_mk_agent",
        },
        "file_hashes"     : {},
    }

    # Now add values from rulesets
    if hostname:
        agconf["is_ipv6_primary"] = is_ipv6_primary(hostname)

        for varname, ruleset in agent_config.items() + [ ("agent_port", agent_ports) ]:
            matchtype = bakery_info.get(varname, {}).get("matchtype", "first")

            if hostname == True:
                entries = generic_host_extra_conf(ruleset)
            else:
                entries = host_extra_conf(hostname, ruleset)

            if varname == "agent_paths":
                matchtype = "dict"
            elif varname == "custom_files":
                matchtype = "list"

            if matchtype == "first":
                if len(entries):
                    agconf[varname] = entries[0]

            elif matchtype == "list":
                if entries:
                    l = []
                    for e in entries:
                        l += e
                    agconf[varname] = l

            elif matchtype == "dict":
                for e in entries[::-1]:
                    agconf[varname].update(e)

            elif matchtype == "all":
                if entries:
                    agconf[varname] = entries

    if "custom_files" in agconf:
        add_custom_file_hashes_to_config(agconf)

    # Create unique has value of this config, since we just want one copy
    # of the agent for each different configuration.
    aghash = agent_config_hash(agconf)
    global g_used_hashes
    g_used_hashes.add(aghash)

    # Some platforms, like CentOS 5.x, do not support building MSI packages.
    # Excluding the Windows MSI build target to prevent errors
    if not supports_msi_building():
        try:
            all_opsyses.remove('windows_msi')
        except ValueError:
            pass

    for opsys in all_opsyses:
        try:
            bake_agent_for_os(hostname, opsys, agconf, aghash)
        except Exception, e:
            if opt_debug:
                raise
            raise MKGeneralException("Error creating %s/%s agent for %s: %s." %
                (aghash, opsys, format_bakery_hostname_for_humans(hostname), e))


def add_custom_file_hashes_to_config(agconf):
    custom_dir = custom_file_base_path()
    for path in get_custom_file_paths(agconf):
        agconf["file_hashes"][("custom", path)] = file_hash(custom_dir + path)


def bake_agent_for_os(hostname, opsys, agconf, aghash):
    if opt_verbose:
        sys.stdout.write(opsys + ":")
    dirpath = baked_agents_dir + "/" + opsys
    pac_dirpath = dirpath + "/_PACKAGES"
    if not os.path.exists(pac_dirpath):
        os.makedirs(pac_dirpath)

    # Make sure that nobody has access to that directory except our OMD site
    os.chmod(dirpath, 0750)
    os.chmod(pac_dirpath, 0750)

    activate_agent_for_host(dirpath, hostname, aghash)

    path = pac_dirpath + "/" + aghash
    if os.path.exists(path) and not opt_force:
        if opt_verbose:
            sys.stdout.write("uptodate...")
            sys.stdout.flush()

    elif path not in g_baked_paths:
        g_baked_paths.add(path)
        if opt_verbose:
            sys.stdout.write("baking...")
            sys.stdout.flush()

        # Write the config file in all cases (useful for debugging when baking fails)
        func = globals()["bake_agent_for_" + opsys]
        package_path = dirpath + "/_PACKAGES/" + aghash

        # Now do the actual baking
        file(package_path + ".conf", "w").write(pprint.pformat(agconf) + "\n")
        os.chmod(package_path + ".conf", 0640)

        func(package_path, agconf, aghash)
        os.chmod(package_path, 0640)

    else: # already baked
        if opt_verbose:
            sys.stdout.write("copying...")
            sys.stdout.flush()


def format_bakery_hostname_for_humans(hostname):
    return format_bakery_hostname(hostname).lstrip('_')


def format_bakery_hostname(hostname):
    if hostname == None:
        return '_VANILLA'
    elif hostname == True:
        return '_GENERIC'
    else:
        return hostname


def activate_agent_for_host(dirpath, hostname, aghash):
    hostpath = dirpath + "/" + format_bakery_hostname(hostname)
    if os.path.lexists(hostpath):
        os.remove(hostpath)
    os.symlink("_PACKAGES/" + aghash, hostpath)


def bake_agent_for_windows_msi(path, agconf, aghash):
    msi_container = agents_dir + "/windows/baked_container.msi"

    # Create working directory
    work_dir = tempfile.mkdtemp(prefix = "msi.workdir.", dir = tmp_dir)
    os.chmod(work_dir, 0700)
    old_pwd = os.getcwd()
    try:
        shutil.copy(agents_dir + "/windows/check_mk_agent.exe", work_dir)
        shutil.copy(agents_dir + "/windows/check_mk_agent-64.exe", work_dir)

        # Now let the plugins work. They may create:
        # 1. plugin files
        # 2. configuration files for the plugins
        # 3. configuration snippets for check_mk.ini. The latter have the format check_mk.ini.SECTION.PLUGIN,
        #    e.g. check_mk.ini.global.onlyfrom
        os.chdir(work_dir)
        os.mkdir(work_dir + "/parts")
        os.mkdir("parts/plugins")
        os.mkdir("parts/config")
        os.mkdir("parts/bin")

        for name, info in bakery_info.items():
            if "windows" in info["os"] and agconf.get(name) != None:
                info["bake_function"]("windows", agconf[name], \
                    work_dir + "/parts/config", work_dir + "/parts/plugins")

        sections = {}
        for filename in os.listdir("parts/config"):
            if filename.startswith("check_mk.ini."):
                snippet = filename[13:]
                section = snippet.split(".")[0]
                sections.setdefault(section, "")
                sections[section] += file("parts/config/check_mk.ini." + snippet).read() + "\r\n"
                os.remove("parts/config/" + filename) # do not pack into plugins.cap

        check_mk_ini = file(work_dir + "/check_mk.ini", "w")
        check_mk_ini.write(agent_file_header.replace("\n", "\r\n"))
        check_mk_ini.write("[global]\r\n")
        check_mk_ini.write(sections.get("global", ""))
        check_mk_ini.write("\r\n")
        section_names = sections.keys()
        section_names.sort()
        for section_name in section_names:
            if section_name != "global":
                check_mk_ini.write("[%s]\r\n" % section_name)
                check_mk_ini.write(sections[section_name])
                check_mk_ini.write("\r\n")

        # Writes data to disk, otherwise the check_mk.ini is empty
        check_mk_ini.close()

        if agconf.get("win_openhardwaremonitor"):
            source_path = agents_dir + "/windows/ohm"
            target_path = work_dir + "/parts/bin"
            for filename in ["OpenHardwareMonitorCLI.exe",
                             "OpenHardwareMonitorLib.dll"]:
                add_file_for_windows(source_path, target_path, [filename])

        add_custom_files_for_windows(agconf, work_dir)

        # Debug check_mk.ini:
        # print
        # print file(work_dir + "/check_mk.ini", "r").read()

        # Create CAP file with all plugins and their configuration files
        # For debugging use verbose mode when packing the plugins:
        # file(work_dir + "/plugins.cap", "w").write(cap_pack("parts", verb=True))
        file(work_dir + "/plugins.cap", "w").write(cap_pack("parts"))

        p = os.popen("msi-update -v %s/windows/baked_container.msi . %s %s 2>&1" % (
              (agents_dir, 12345, check_mk_version)))
        output = p.read()
        if p.close():
            raise MKGeneralException("Cannot create MSI file: %s" % output)
        else:
            shutil.copy("check_mk_agent.msi", path)

    except:
        shutil.rmtree(work_dir)
        raise

    os.chdir(old_pwd)
    shutil.rmtree(work_dir)


def custom_file_base_path():
    return omd_root + "/local/share/check_mk/agents/custom/"


def get_custom_file_paths(agconf):
    paths = []
    custom_dir = custom_file_base_path()
    for package in set(agconf.get("custom_files", [])):
        custom_path = custom_dir + package
        if not os.path.isdir(custom_path):
            continue

        for logical_path in os.listdir(custom_path):
            # We only include from the logical paths lib, bin, var, config
            if logical_path not in [ "lib", "bin", "var", "config" ] or \
               not os.path.isdir("%s/%s" % (custom_path, logical_path)):
                continue

            src_path = "%s/%s/" % (custom_path, logical_path)
            for dirpath, dirnames, filenames in os.walk(src_path):
                rel_to_logical_path = dirpath[len(src_path):]
                for filename in filenames:
                    paths.append("%s/%s/%s/%s" % (package, logical_path,
                                                  rel_to_logical_path, filename))
    return paths


def add_file_for_windows(source_base, target_base, rel_path):
    rel_path = "/".join(rel_path)
    source_path = "%s/%s" % (source_base, rel_path)
    target_path = "%s/%s" % (target_base, rel_path)

    target_dir  = os.path.dirname(target_path)
    if not os.path.exists(target_dir):
        os.mkdir(target_dir)

    shutil.copy(source_path, target_path)


# For windows packages the custom files are put to the agent package relative to
# the agent base directory. The package and logical paths are stripped off.
def add_custom_files_for_windows(agconf, work_dir):
    custom_dir = custom_file_base_path()
    for path in get_custom_file_paths(agconf):
        # strip package name and logical path component (first two path elements in path)
        pack_path = "/".join(path.split("/")[:2])
        rel_path  = path.split("/")[2:]
        add_file_for_windows(custom_dir + "/" + pack_path, work_dir + "/parts", rel_path)


def bake_agent_for_linux_deb(path, agconf, aghash):
    deb_topdir = tempfile.mkdtemp(prefix = "deb.topdir.", dir = tmp_dir)
    file(deb_topdir + "/debian-binary", "w").write("2.0\n")

    os.mkdir(deb_topdir + "/agentfiles")
    create_linux_agent_files(deb_topdir + "/agentfiles", agconf)

    # Prepare control directory first (md5sums will be created soon)
    os.mkdir(deb_topdir + "/control")

    # The actual data. We create these files as normal user, but
    # pack them into the tarball as root:root.
    os.mkdir(deb_topdir + "/data")
    config_dir = agconf["agent_paths"]["config"]
    lib_dir = agconf["agent_paths"]["lib"]
    bin_dir = agconf["agent_paths"]["bin"]
    var_dir = agconf["agent_paths"]["var"]

    if os.system("""set -e
cd '%(deb_topdir)s/data'
mkdir -p .%(lib_dir)s/local
mkdir -p .%(lib_dir)s/plugins
mkdir -p .%(bin_dir)s
install -m 755 ../agentfiles/bin/* .%(bin_dir)s/
mkdir -p .%(config_dir)s
cp -r ../agentfiles/plugins/* .%(lib_dir)s/plugins 2>/dev/null || true
install -m 640 ../agentfiles/pluginconfig/* .%(config_dir)s 2>/dev/null || true
mkdir -p etc/xinetd.d
install -m 644 ../agentfiles/xinetd.conf etc/xinetd.d/check_mk_agent
mkdir -p .%(var_dir)s/job
mkdir -p .%(var_dir)s/cache

# Create MD5 Sums right now
find * -type f | xargs md5sum > ../control/md5sums

""" % vars()) or \
       os.system("tar czf %s/data.tar.gz --owner=root --group=root -C %s/data ." % (
                deb_topdir, deb_topdir)):
        raise MKGeneralException("Cannot create data.tar.gz")

    # conffiles: lists all configuration files. We do not specify any file
    # as configuration file since alle files are handled by WATO
    file(deb_topdir + "/control/conffiles", "w").write("")

    # control: general package information
    version = check_mk_version
    configpprint = pprint.pformat(agconf)
    file(deb_topdir + "/control/control", "w").write("""
Package: check-mk-agent
Version: %(version)s-%(aghash)s
Architecture: all
Maintainer: WATO - Check_MK's Web Administration Tool
Installed-Size: 1
Section: System/Monitoring
Priority: extra
Suggests: time
Description: Check_MK Agent for Linux, config hash %(aghash)s
 The Check_MK Agent for Linux has been created by the Check_MK Agent Bakery. It
 is created specifically for a set of hosts sharing the same configuration:
 .
 %(configpprint)s
 .
 NOTE: Do not modify any of the configuration files manually. They are maintained
 via WATO and will be replaced on the next update of the agent.
""" % vars())

    # Skripts
    file(deb_topdir + "/control/postinst", "w").write("""#!/bin/sh
if [ -x /etc/init.d/xinetd ] ; then
    if pgrep -x xinetd >/dev/null ; then
        echo "Reloading xinetd..."
        /etc/init.d/xinetd reload
    else echo "Starting xinetd..."
        /etc/init.d/xinetd start
    fi
fi
""")
    shutil.copy(deb_topdir + "/control/postinst", deb_topdir + "/control/postrm")
    # Note: we just cleanup the var directory if the user has not changed
    # the paths! Otherwise he might delete some important directory
    if agconf["agent_paths"]["var"] == "/var/lib/check_mk_agent":
        file(deb_topdir + "/control/prerm", "w").write("""#!/bin/sh\n
if [ "$1" = remove ] ; then
    rm -rf /var/lib/check_mk_agent
fi
""")
    else:
        file(deb_topdir + "/control/prerm", "w").write("#!/bin/sh\n")

    file(deb_topdir + "/control/preinst", "w").write("""#!/bin/sh
if [ ! -x /etc/init.d/xinetd ] ; then
    echo
    echo "---------------------------------------------"
    echo "WARNING"
    echo
    echo "This package needs xinetd to be installed. "
    echo "Currently you do not have installed xinetd. "
    echo "Please install and start xinetd or install "
    echo "and setup another inetd manually."
    echo ""
    echo "It's also possible to monitor via SSH without "
    echo "an inetd."
    echo "---------------------------------------------"
    echo
fi
""")
    for script in [ "postinst", "postrm", "preinst", "prerm" ]:
        os.chmod(deb_topdir + "/control/" + script, 0755)

    # pack control.tar.gz
    if os.system("tar czf %s/control.tar.gz --owner=root --group=root -C %s/control ." % (
                deb_topdir, deb_topdir)):
        raise MKGeneralException("Cannot create control.tar.gz")

    if os.system("cd %s ; ar cr %s debian-binary *.tar.gz" % (deb_topdir, path)):
        raise MKGeneralException("Cannot create ar archive.")
    shutil.rmtree(deb_topdir)


def bake_agent_for_linux_rpm(path, agconf, aghash):
    rpm_topdir = tempfile.mkdtemp(prefix = "rpm.topdir.", dir = tmp_dir)
    try:
        for subdir in [ "RPMS", "SRPMS", "SOURCES", "BUILD", "SPECS", "buildroot" ]:
            os.mkdir(rpm_topdir + "/" + subdir)

        # Create tarball with shipped agent parts
        if os.system("tar czf %s/SOURCES/agent.tar.gz -C %s ." % (rpm_topdir, agents_dir)):
            raise MKGeneralException("Cannot create tarball of agent sources")

        if local_agents_dir \
              and os.path.exists(local_agents_dir) \
              and os.system("tar czf %s/SOURCES/agent_local.tar.gz -C %s ." % (rpm_topdir, local_agents_dir)):
            raise MKGeneralException("Cannot create tarball of local agent sources")

        # Now create spec file.
        spec_path = rpm_topdir + "/check-mk-agent.spec"
        release = aghash
        version = check_mk_version.replace("-", "_")
        configpprint = pprint.pformat(agconf)
        bin_dir = agconf["agent_paths"]["bin"]
        config_dir = agconf["agent_paths"]["config"]
        lib_dir = agconf["agent_paths"]["lib"]
        var_dir = agconf["agent_paths"]["var"]

        # Create configuration for xinetd
        create_linux_agent_files(rpm_topdir + "/SOURCES", agconf)

        file(spec_path, "w").write("""# Spec file created by Check_MK Agent Bakery
Summary:   Check_MK Agent for Linux, config hash %(aghash)s
Name:      check-mk-agent
Version:   %(version)s
Release:   %(release)s
License:   GPL
Group:     System/Monitoring
URL:       http://mathias-kettner.de/check_mk
BuildRoot: %(rpm_topdir)s/buildroot
AutoReq:   off
AutoProv:  off
BuildArch: noarch
Obsoletes: check_mk-agent check_mk_agent check_mk-agent-logwatch
Provides:  check_mk-agent check_mk_agent

%%description
The Check_MK Agent for Linux has been created by the Check_MK Agent Bakery. It
is created specifically for a set of hosts sharing the same configuration.

NOTE: Do not modify any of the configuration files manually. They are maintained
via WATO and will be replaced on the next update of the agent.

%%prep
echo '====================== >>> prep <<< ========================'
mkdir agents
tar xzf ../SOURCES/agent.tar.gz -C agents
if [ -e ../SOURCES/agent_local.tar.gz ] ; then
    tar xzf ../SOURCES/agent_local.tar.gz -C agents
fi
cp ../SOURCES/xinetd.conf .
cp -r ../SOURCES/plugins .
cp -r ../SOURCES/pluginconfig .

%%install
echo '====================== >>> install <<< ========================'
# Overcome nasty bug in RPM/Ubuntu. It breaks the permissions of
# plugins and changes it to 644. Nobody knows why...
chmod 755 ../BUILD/agents/plugins

R=$RPM_BUILD_ROOT
rm -rf $R
mkdir -p $R%(lib_dir)s/local
mkdir -p $R%(lib_dir)s/plugins
mkdir -p $R%(bin_dir)s
install -m 755 ../SOURCES/bin/* $R%(bin_dir)s/
mkdir -p $R%(config_dir)s
cp -r plugins/* $R%(lib_dir)s/plugins || true
install -m 644 pluginconfig/* $R%(config_dir)s || true
chmod 640 $R%(config_dir)s/* || true
mkdir -p $R/etc/xinetd.d
install -m 644 xinetd.conf $R/etc/xinetd.d/check_mk_agent
mkdir -p $R%(var_dir)s/job
mkdir -p $R%(var_dir)s/cache

%%clean
rm -rf $RPM_BUILD_ROOT

%%files
%%defattr(-,root,root)
%(bin_dir)s/*
/etc/xinetd.d/*
%(config_dir)s
%(lib_dir)s
%%dir %(var_dir)s
%%dir %(var_dir)s/job
%%dir %(var_dir)s/cache

%%define reload_xinetd if [ -x /etc/init.d/xinetd ] ; then if pgrep -x xinetd >/dev/null ; then echo "Reloading xinetd..." ; /etc/init.d/xinetd reload ; else echo "Starting xinetd..." ; /etc/init.d/xinetd start ; fi ; fi
%%define activate_xinetd if which chkconfig >/dev/null 2>&1 ; then echo "Activating startscript of xinetd" ; chkconfig xinetd on ; fi

%%pre
if [ ! -x /etc/init.d/xinetd ] ; then
    echo
    echo "---------------------------------------------"
    echo "WARNING"
    echo
    echo "This package needs xinetd to be installed. "
    echo "Currently you do not have installed xinetd. "
    echo "Please install and start xinetd or install "
    echo "and setup another inetd manually."
    echo ""
    echo "It's also possible to monitor via SSH without "
    echo "an inetd."
    echo "---------------------------------------------"
    echo
fi

%%post
%%activate_xinetd
%%reload_xinetd

%%postun
%%reload_xinetd


""" % vars())

        p = os.popen('rpmbuild -ba --buildroot %s/buildroot --define "_binary_payload w9.gzdio" --define "_topdir %s" --define "_source_filedigest_algorithm  1" --define "_binary_filedigest_algorithm  1" %s/check-mk-agent.spec 2>&1' % (
                          rpm_topdir, rpm_topdir, rpm_topdir))
        output = p.read()
        exit_code = p.close()
        if exit_code:
            if opt_debug:
                sys.stderr.write(output)
            raise MKGeneralException("Error building RPM packet.\n%s" % output)

        rpm_path = rpm_topdir + "/RPMS/noarch"
        rpm_name = "check-mk-agent-" + version + "-" + release + ".noarch.rpm"
        file(path, "w").write(file(rpm_path + "/" + rpm_name).read())

    except:
        # shutil.rmtree(rpm_topdir)
        raise

    try:
        shutil.rmtree(rpm_topdir)
    except:
        if opt_debug:
            raise


def fixup_paths_in_agent(agconf, source_path):
    content = ""
    paths = agconf["agent_paths"]
    for line in file(source_path):
        if line.startswith("export MK_LIBDIR="):
            line = "export MK_LIBDIR=%s\n" % quote_shell_string(paths["lib"])
        if line.startswith("export MK_VARDIR="):
            line = "export MK_VARDIR=%s\n" % quote_shell_string(paths["var"])
        elif line.startswith("export MK_CONFDIR="):
            line = "export MK_CONFDIR=%s\n" % quote_shell_string(paths["config"])
        elif line.startswith("PATH=") and paths["bin"] not in [ "/bin", "/usr/bin", "/usr/local/bin" ]:
            line = "PATH=%s:$PATH:/usr/local/bin\n" % quote_shell_string(paths["bin"])
        content += line
    return content


def create_linux_agent_files(targetdir, agconf):
    # Copy agent binaries and modify paths on the fly
    os.mkdir(targetdir + "/bin")
    for binfile in [ "waitmax", "mk-job", "check_mk_agent.linux", "check_mk_caching_agent.linux" ]:
        if local_agents_dir and os.path.exists(local_agents_dir + "/" + binfile):
            source_path = local_agents_dir + "/" + binfile
        else:
            source_path = agents_dir + "/" + binfile
        if binfile.startswith("check_mk_") or binfile == "mk-job":
            content = fixup_paths_in_agent(agconf, source_path)
            if binfile.endswith(".linux"):
                binfile = binfile[:-6]
            file(targetdir + "/bin/" + binfile, "w").write(content)
            os.chmod(targetdir + "/bin/" + binfile, 0755)
        else:
            shutil.copy(source_path, targetdir + "/bin/" + binfile)

    # Always install agent updater into /usr/bin/
    shutil.copy(agents_dir + "/plugins/cmk-update-agent", targetdir + "/bin/cmk-update-agent")
    os.chmod(targetdir + "/bin/cmk-update-agent", 0755)

    # Configuration for xinetd. Once we have a configuration ruleset
    # for "Monitor via SSH", we can create an empty file instead.

    # TODO: - switch for caching agent
    agent_port = agconf["agent_port"] # must be set
    only_from = agconf.get("only_from")
    agent_user = agconf.get("agent_user", "root")
    if only_from:
        only_from_txt = "\tonly_from      = %s\n" % " ".join(only_from)
    else:
        only_from_txt = ""
    bin_dir = agconf["agent_paths"]["bin"]

    if agconf["is_ipv6_primary"]:
        flags_txt = "\tflags          = IPv6\n"
    else:
        flags_txt =""

    file(targetdir + "/xinetd.conf", "w").write(agent_file_header + """
service check_mk_agent
{
	type           = UNLISTED
	port           = %(agent_port)d
	socket_type    = stream
	protocol       = tcp
	wait           = no
	user           = %(agent_user)s
	server         = %(bin_dir)s/check_mk_agent
	log_on_success =
	disable        = no
%(flags_txt)s%(only_from_txt)s}
""" % vars())

    create_unix_agent_plugin_files("linux", targetdir, agconf)


def create_unix_agent_plugin_files(flavour, targetdir, agconf):
    os.makedirs(targetdir + "/pluginconfig")
    os.makedirs(targetdir + "/plugins")

    for name, info in bakery_info.items():
        if "linux" in info["os"] and agconf.get(name) != None:
            info["bake_function"]("linux", agconf[name], \
                targetdir + "/pluginconfig", targetdir + "/plugins")


def bake_agent_for_linux_tgz(path, agconf, aghash):
    bake_agent_for_unix_tgz("linux", path, agconf, aghash)


def bake_agent_for_aix_tgz(path, agconf, aghash):
    bake_agent_for_unix_tgz("aix", path, agconf, aghash)


def bake_agent_for_unix_tgz(flavour, path, agconf, aghash):
    topdir = tempfile.mkdtemp(prefix = "tgz.topdir.", dir = tmp_dir)
    tarroot = topdir + "/tarroot"
    paths = agconf["agent_paths"]
    first_dirs = set([])
    for dirpath in paths.values():
        if not os.path.exists(tarroot + dirpath):
            os.makedirs(tarroot + dirpath)
        first_dirs.add(quote_shell_string(dirpath.split("/")[1]))
    os.makedirs(tarroot + paths["lib"] + "/plugins")
    os.makedirs(tarroot + paths["lib"] + "/local")
    os.makedirs(tarroot + paths["var"] + "/job")
    os.makedirs(tarroot + paths["var"] + "/cache")

    if flavour == "linux":
        create_linux_agent_files(topdir, agconf)
    else: # aix
        create_aix_agent_files(topdir, agconf)

    for binfile in os.listdir(topdir + "/bin"):
        shutil.move(topdir + "/bin/" + binfile, tarroot + paths["bin"])
    if os.path.exists(topdir + "/pluginconfig"):
        for pluginconfig in os.listdir(topdir + "/pluginconfig"):
            shutil.move(topdir + "/pluginconfig/" + pluginconfig, tarroot + paths["config"])
    if os.path.exists(topdir + "/plugins"):
        for plugin in os.listdir(topdir + "/plugins"):
            shutil.move(topdir + "/plugins/" + plugin, tarroot + paths["lib"] + "/plugins")

    if flavour == "linux":
        os.makedirs(tarroot + "/etc/xinetd.d")
        shutil.move(topdir + "/xinetd.conf", tarroot + "/etc/xinetd.d/check_mk_agent")

    pipe = os.popen("tar cvzf %s -C %s %s 2>&1" % (
          quote_shell_string(path),
          quote_shell_string(tarroot),
          " ".join(first_dirs)))
    output = pipe.read()
    if pipe.close():
        raise MKGeneralException("Error creating tar: %s" % output)
    shutil.rmtree(topdir)

def create_aix_agent_files(targetdir, agconf):
    os.mkdir(targetdir + "/bin")
    if local_agents_dir and os.path.exists(local_agents_dir + "/check_mk_agent.aix"):
        source_path = local_agents_dir + "/check_mk_agent.aix"
    else:
        source_path = agents_dir + "/check_mk_agent.aix"
    content = fixup_paths_in_agent(agconf, source_path)
    file(targetdir + "/bin/check_mk_agent", "w").write(content)
    os.chmod(targetdir + "/bin/check_mk_agent", 0755)

    create_unix_agent_plugin_files("aix", targetdir, agconf)


# Create a unique hash string of the agent configuration. This
# is used as file name for creating that agent. Note: we do not
# need cryptographic security, we rather prefer short filenames.
# So we cut down the length a bit... 12 hex digits make up for
# 18,446,744,073,709,551,616 different agents, that is approx.
# 1.8 * 10^29.
def agent_config_hash(agconf):
    return md5(repr(agconf)).hexdigest()[:16]


# Code duplication with plugins/wato/agent_bakery.py. Sorry.
def file_hash(path):
    h = md5()
    f = file(path)
    buf = f.read(65536)
    while len(buf) > 0:
        h.update(buf)
        buf = f.read(65536)
    return h.hexdigest()


def cleanup_orphaned_packages():
    for opsys in all_opsyses:
        dirpath = baked_agents_dir + "/" + opsys
        cleanup_symlinks_of_vanished_hosts(dirpath)
        cleanup_unused_packages(dirpath)


def cleanup_symlinks_of_vanished_hosts(dirpath):
    for filename in os.listdir(dirpath):
        if not filename.startswith("_") \
           and not filename in ['.', '..'] \
           and not filename in all_configured_realhosts():
            os.remove(dirpath + "/" + filename)


def cleanup_unused_packages(dirpath):
    pac_dirpath = dirpath + "/_PACKAGES"
    for filename in os.listdir(pac_dirpath):
        if filename not in ['.', '..'] \
           and not filename.endswith('.conf') \
           and not filename.endswith('.sig') \
           and filename not in g_used_hashes:

            for candidate in [ filename, filename + ".conf", filename + ".sig" ]:
                path = pac_dirpath + "/" + candidate
                try:
                    os.remove(path)
                except OSError, e:
                    if e.errno == 2: # No such file or directory
                        pass
                    else:
                        raise
