#!/usr/bin/python
# -*- encoding: utf-8; py-indent-offset: 4 -*-
# .------------------------------------------------------------------------.
# |                ____ _               _        __  __ _  __              |
# |               / ___| |__   ___  ___| | __   |  \/  | |/ /              |
# |              | |   | '_ \ / _ \/ __| |/ /   | |\/| | ' /               |
# |              | |___| | | |  __/ (__|   <    | |  | | . \               |
# |               \____|_| |_|\___|\___|_|\_\___|_|  |_|_|\_\              |
# |                                        |_____|                         |
# |             _____       _                       _                      |
# |            | ____|_ __ | |_ ___ _ __ _ __  _ __(_)___  ___             |
# |            |  _| | '_ \| __/ _ \ '__| '_ \| '__| / __|/ _ \            |
# |            | |___| | | | ||  __/ |  | |_) | |  | \__ \  __/            |
# |            |_____|_| |_|\__\___|_|  | .__/|_|  |_|___/\___|            |
# |                                     |_|                                |
# |                     _____    _ _ _   _                                 |
# |                    | ____|__| (_) |_(_) ___  _ __                      |
# |                    |  _| / _` | | __| |/ _ \| '_ \                     |
# |                    | |__| (_| | | |_| | (_) | | | |                    |
# |                    |_____\__,_|_|\__|_|\___/|_| |_|                    |
# |                                                                        |
# | mathias-kettner.com                                 mathias-kettner.de |
# '------------------------------------------------------------------------'
#  This file is part of the Check_MK Enterprise Edition (CEE).
#  Copyright by Mathias Kettner and Mathias Kettner GmbH.  All rights reserved.
#
#  Distributed under the Check_MK Enterprise License.
#
#  You should have  received  a copy of the Check_MK Enterprise License
#  along with Check_MK. If not, email to mk@mathias-kettner.de
#  or write to the postal address provided at www.mathias-kettner.de

alert_log_path        = log_dir + "/alerts.log"
alert_logging         = 1
alert_handler_dir     = omd_root + "/local/share/check_mk/alert_handlers"
alert_handler_timeout = 60


# Keep track of the currently running handlers. Each entry is
# a dict. Example:
# g_running_handlers = [
#   {
#        "pid"        : 18124,
#        "pipe_fd"    : 8,
#        "output"     : "This output has so far been read...",
#        "context"    : { var -> value },
#        "start_time" : 1434381063.12,
#    }
# ]
g_running_handlers = []

# Keep track of inline handlers (implemented in Python, called via
# function call, not via process execution).
g_inline_handlers = {}


def alert_log(message):
    if alert_logging >= 1:
        event_log(alert_log_path, message)


def alert_log_debug(message):
    if alert_logging >= 2:
        alert_log(message)


# TODO: Make use of the generic do_keepalive() mechanism
def do_handle_alerts(args):
    alert_log("----------------------------------------------------------------------")
    load_inline_alert_handlers()
    register_keepalive_sigint_handler()
    event_keepalive(
        event_function    = process_alert,
        log_function      = alert_log,
        call_every_loop   = collect_finished_alert_handlers,
        loop_interval     = 2, # check handlers every 2 seconds
        shutdown_function = shutdown_alert_handlers,
    )


def load_inline_alert_handlers():
    for rule in alert_handler_rules:
        handler_name, handler_arguments = rule["alert_handler"]
        if handler_name not in g_inline_handlers and handler_is_inline(handler_name):
            if handler_arguments:
                alert_log("Warning: inline handlers take no arguments. Arguments %s for %s are ignored" % (
                    ", ".join(handler_arguments), handler_name))
            load_inline_alert_handler(handler_name)



# Detect inline handlers. They have the line "# Bulk: yes" as line 2 or 3
def handler_is_inline(handler_name):
    handler_path = alert_handler_dir + "/" + handler_name
    try:
        lines = file(handler_path).readlines()
    except Exception, e:
        alert_log("Cannot load alert handler %s: %s" % (handler_path, e))
        if opt_debug:
            raise
        return False

    for line in lines[:3]:
        if line.strip().lower().replace(" ", "") == "#inline:yes":
            return True
    return False


def load_inline_alert_handler(handler_name):
    handler_path = alert_handler_dir + "/" + handler_name
    g_inline_handlers[handler_name] = None # remember we've tried in case of an error

    namespace = {}
    add_global_variables_to_handler(handler_name, namespace)

    try:
        execfile(handler_path, namespace, namespace)
    except Exception, e:
        alert_log("Cannot load inline alert handler %s: %s" % (handler_path, e))
        if opt_debug:
            raise
        return

    implemented_phases = []
    for phase in [ "init", "alert", "reload", "shutdown" ]:
        if "handle_" + phase in namespace:
            implemented_phases.append(phase)

    if not implemented_phases:
        alert_log("Cannot load inline alert handler %s: it does not contain any valid handle_*() function." %
                  handler_name)
        return

    if "alert" not in implemented_phases:
        alert_log("Warnng: inline handler %s does not implement handle_alert()" % handler_name)

    g_inline_handlers[handler_name] = namespace
    alert_log("Successfully loaded inline alert handler %s. Implemented phases: %s" % (
        handler_path, ", ".join(implemented_phases)))

    run_inline_alert_handler(handler_name, context=None, phase="init")


def add_global_variables_to_handler(handler_name, namespace):
    namespace["log"] = create_handler_log_function(handler_name)
    namespace["omd_root"] = omd_root
    namespace["omd_site"] = omd_site


def create_handler_log_function(handler_name):
    return lambda text: alert_log("Inline handler %s: %s" % (handler_name, text))


def process_alert(raw_context):
    alert_log("Got raw alert (%s) context with %s variables" % (
        find_host_service_in_context(raw_context), len(raw_context)))

    if alert_logging >= 2:
        alert_log(render_context_dump(raw_context))

    complete_raw_context(raw_context, with_dump = alert_logging >= 2, event_log = alert_log)
    execute_alert_rules(raw_context)


def execute_alert_rules(context):
    for rule in alert_handler_rules:
        alert_log("Rule '%s'..." % rule["description"])
        why_not = alert_rule_matches(rule, context)
        if why_not:
            alert_log(" -> does not match: %s" % why_not)
        else:
            alert_log(" -> matches!")
            execute_alert_handler_rule(rule, context)


def alert_rule_matches(rule, context):
    if rule.get("disabled"):
        return "This rule is disabled"

    return event_match_rule(rule, context) or \
           alert_match_host_event(rule, context) or \
           alert_match_service_event(rule, context) or \
           alert_match_attempt(rule, context)

def alert_match_host_event(rule, context):
    if "match_host_event" in rule:
        if context["WHAT"] != "HOST":
            if "match_service_event" not in rule:
                return "This is a service alert, but the rule just matches host events"
            else:
                return # Let this be handled by match_service_event
        allowed_events = rule["match_host_event"]
        state          = context["HOSTSTATE"]
        last_state     = context["LASTHOSTSTATE"]
        events         = { "UP" : 'r', "DOWN" : 'd', "UNREACHABLE" : 'u' }
        return alert_match_event(context, state, last_state, events, allowed_events)


def alert_match_service_event(rule, context):
    if "match_service_event" in rule:
        if context["WHAT"] != "SERVICE":
            if "match_host_event" not in rule:
                return "This is a host alert, but the rule just matches service events"
            else:
                return # Let this be handled by match_host_event
        allowed_events = rule["match_service_event"]
        state          = context["SERVICESTATE"]
        last_state     = context["LASTSERVICESTATE"]
        events         = { "OK" : 'r', "WARNING" : 'w', "CRITICAL" : 'c', "UNKNOWN" : 'u' }
        return alert_match_event(context, state, last_state, events, allowed_events)


def alert_match_event(context, state, last_state, events, allowed_events):
    event = events.get(last_state, '?') + events.get(state, '?')

    alert_log("Event type is %s" % event)

    # Now go through the allowed events. Handle '?' has matching all types!
    for allowed in allowed_events:
        if event == allowed or \
            (allowed[0] == '?' and len(event) > 1 and event[1] == allowed[1]) or \
            (event[0] == '?' and len(allowed) > 1 and event[1] == allowed[1]):
            return

    return "Event type '%s' not handled by this rule. Allowed are: %s" % (
            event, ", ".join(allowed_events))


def alert_match_attempt(rule, context):
    if "match_attempt" in rule:
        from_attempt, to_attempt = rule["match_attempt"]
        if context["WHAT"] == "HOST":
            check_attempt = int(context.get("HOSTATTEMPT", 1))
        else:
            check_attempt = int(context.get("SERVICEATTEMPT", 1))
        if check_attempt < from_attempt or check_attempt > to_attempt:
            return "The check attempt %d does not lie in range %d ... %d" % (
                    check_attempt, from_attempt, to_attempt)


def execute_alert_handler_rule(rule, context):
    handler_name, handler_params = rule["alert_handler"]
    alert_log("Executing alert handler %s for %s" % (handler_name, find_host_service_in_context(context)))

    if handler_name in g_inline_handlers:
        run_inline_alert_handler(handler_name, context, phase="alert")
    else:
        run_process_alert_handler(handler_name, handler_params, context)


def run_inline_alert_handler(handler_name, context, phase):
    namespace = g_inline_handlers[handler_name]
    if namespace == None:
        alert_log("Cannot execute handler: not correctly loaded.")
        return

    function_name = "handle_" + phase
    if function_name in namespace:
        try:
            if phase == "alert":
                namespace[function_name](context)
            else:
                namespace[function_name]()

        except Exception, e:
            alert_log("Error executing inline handler %s function %s(): %s" % (
                handler_name, function_name, e))
            if opt_debug:
                raise


def run_process_alert_handler(handler_name, handler_params, context):
    handler_path = alert_handler_dir + "/" + handler_name
    if not os.path.exists(handler_path):
        raise MKGeneralException("Cannot find alert handler %s" % handler_path)

    handler_pid, handler_pipe = run_alert_handler_async(handler_path, [ handler_name ] + handler_params, context)
    g_running_handlers.append({
        "pid"        : handler_pid,
        "pipe_fd"    : handler_pipe,
        "output"     : "",
        "context"    : context,
        "start_time" : time.time(),
    })
    alert_log("Spawned event handler with PID %d" % handler_pid)


def run_alert_handler_async(handler_path, command_line_arguments, plugin_context):
    read_end, write_end = os.pipe()
    pid = os.fork()
    if not pid:
        os.close(read_end)
        os.dup2(write_end, 1)
        os.dup2(write_end, 2)
        os.close(write_end)
        dev_null = os.open("/dev/null", os.O_RDONLY)
        os.dup2(dev_null, 0)
        os.close(dev_null)
        for fd in range(3, 1024):
            try:
                os.close(fd)
            except:
                pass
        add_context_to_environment(plugin_context, "ALERT_")

        try:
            os.execv(handler_path, command_line_arguments)
        except OSError, e:
            if e.errno == 13: # [Errno 13] Permission denied
                if opt_debug:
                    raise
                raise MKGeneralException("Failed to execute alert handler %s: %s" % (handler_path, e))
            else:
                raise

    os.close(write_end)
    old_flags = fcntl.fcntl(read_end, fcntl.F_GETFL)
    fcntl.fcntl(read_end, fcntl.F_SETFL, old_flags | os.O_NONBLOCK)
    return pid, read_end


def collect_finished_alert_handlers():
    dump_running_alert_handlers()
    remaining = []

    for handler in g_running_handlers:
        if not collect_finished_alert_handler(handler):
            remaining.append(handler)
            check_alert_handler_timeout(handler)
        else:
            try:
                os.close(handler["pipe_fd"])
            except OSError, e:
                if e.errno == 9: # Bad file descriptor
                    # In case the process has terminated or was never started correctly,
                    # the fd is already closed
                    pass
                else:
                    raise

    g_running_handlers[:] = remaining


def collect_finished_alert_handler(handler):
    while try_read_alert_output(handler):
        pass

    try:
        pid, status = os.waitpid(handler["pid"], os.WNOHANG)
    except Exception, e:
        alert_log("Error collection handler processes with PID %d: %s" % (
            handler["pid"], e))
        if opt_debug:
            raise
        return True

    if pid:
        alert_log_debug("Handler [%d] has exited." % handler["pid"])
        try_read_alert_output(handler)
        handler["status"] = status
        log_finished_alert_handler(handler)
        return True
    else:
        return False


def check_alert_handler_timeout(handler):
    now = time.time()
    age = now - handler["start_time"]

    if age > 2 * alert_handler_timeout:
        os.kill(handler["pid"], 9) # SIGKILL
        alert_log("Killed handler [%d] with signal KILL after %d sec timeout" % (
                handler["pid"], 2 * alert_handler_timeout))

    elif age > alert_handler_timeout:
        os.kill(handler["pid"], 15) # SIGTERM
        alert_log("Killed handler [%d] with signal TERM after %d sec timeout" % (
                handler["pid"], alert_handler_timeout))


def try_read_alert_output(handler):
    try:
        script_output = os.read(handler["pipe_fd"], 4096)
        if script_output:
            alert_log_debug("Read '%s' from alert handler %d" % (script_output, handler["pid"]))
            handler["output"] += script_output
            return True
    except:
        pass
    return False


def log_finished_alert_handler(handler):
    # TODO: We could output the elapsed time!
    alert_log("Handler [%d] for %s exited with %s." % (
             handler["pid"],
             find_host_service_in_context(handler["context"]),
             render_exit_status(handler["status"])))
    alert_log("Output: %s" % handler["output"].rstrip())


def render_exit_status(status):
    if status & 0x7f:
        return "signal %d" % (status & 0x7f)
    else:
        return "exit code %d" % (status >> 8)


def shutdown_alert_handlers():
    for handler_name in g_inline_handlers:
        run_inline_alert_handler(handler_name, context=None, phase="shutdown")

    while g_running_handlers:
        collect_finished_alert_handlers()
        if g_running_handlers:
            alert_log("Still waiting for %d alert handlers to finish.." % len(g_running_handlers))
            time.sleep(2)


def dump_running_alert_handlers():
    if g_running_handlers:
        alert_log("%d running alert handlers:" % len(g_running_handlers))
        for handler in g_running_handlers:
            alert_log("PID: %d, object: %s" % (handler["pid"], find_host_service_in_context(handler["context"])))
