#!/usr/bin/python
# -*- encoding: utf-8; py-indent-offset: 4 -*-
# .------------------------------------------------------------------------.
# |                ____ _               _        __  __ _  __              |
# |               / ___| |__   ___  ___| | __   |  \/  | |/ /              |
# |              | |   | '_ \ / _ \/ __| |/ /   | |\/| | ' /               |
# |              | |___| | | |  __/ (__|   <    | |  | | . \               |
# |               \____|_| |_|\___|\___|_|\_\___|_|  |_|_|\_\              |
# |                                        |_____|                         |
# |             _____       _                       _                      |
# |            | ____|_ __ | |_ ___ _ __ _ __  _ __(_)___  ___             |
# |            |  _| | '_ \| __/ _ \ '__| '_ \| '__| / __|/ _ \            |
# |            | |___| | | | ||  __/ |  | |_) | |  | \__ \  __/            |
# |            |_____|_| |_|\__\___|_|  | .__/|_|  |_|___/\___|            |
# |                                     |_|                                |
# |                     _____    _ _ _   _                                 |
# |                    | ____|__| (_) |_(_) ___  _ __                      |
# |                    |  _| / _` | | __| |/ _ \| '_ \                     |
# |                    | |__| (_| | | |_| | (_) | | | |                    |
# |                    |_____\__,_|_|\__|_|\___/|_| |_|                    |
# |                                                                        |
# | mathias-kettner.com                                 mathias-kettner.de |
# '------------------------------------------------------------------------'
#  This file is part of the Check_MK Enterprise Edition (CEE).
#  Copyright by Mathias Kettner and Mathias Kettner GmbH.  All rights reserved.
#
#  Distributed under the Check_MK Enterprise License.
#
#  You should have  received  a copy of the Check_MK Enterprise License
#  along with Check_MK. If not, email to mk@mathias-kettner.de
#  or write to the postal address provided at www.mathias-kettner.de

# Code for implementing the simplest possible archive format: CAP.  CAP is the
# Check_MK Agent Package format. Needed for the Windows agent. It's a bit like
# tar. But has absolutely no features other then packing files. We have an
# own format since we need to implement this in our standalone C Windows agent.

# Format:
# CAP ::= FILE FILE FILE ...
# FILE ::= NAME CONTENT
# NAME ::= NAME_LENGTH(1 Byte) NAME_CHARS(NAME_LENGTH Bytes)
# CONTENT ::= CONTENT_LENGTH(4 Bytes LE) CONTENT_BYTES(CONTENT_LENGTH Bytes)
# Notes:
# - An empty file is a valid CAP file.
# - No mode or permission bits are stored.
# - the file name is a relative path, relative to the Check_MK Agent directory
# - the file name may contain /: This is the path separator (not \)


# Implement option --cap (just for testing)
def do_cap(args):
    if len(args) < 2 or len(args) > 3:
        cap_usage()

    mode = args[0]
    if mode not in ("pack", "unpack", "list"):
        bail_out("Please specify pack, unpack or list, not '%s'" % mode)

    packfile = args[1]
    if mode == "unpack" and not os.path.exists(packfile):
        bail_out("Cannot open %s: no such file." % packfile)

    if len(args) == 3:
        dirpath = args[2]
    else:
        dirpath = "."

    if mode == "pack":
        try:
            content = cap_pack(dirpath, verb=opt_verbose)
            file(packfile, "w").write(content)
            verbose("Wrote %s (%d bytes).\n" % (packfile, len(content)))
        except Exception, e:
            if opt_debug:
                raise
            bail_out("Cannot create cap file: %s" % e)

    else:
        try:
            content = file(packfile).read()
            cap_unpack(dirpath, content, mode=="list", verb=opt_verbose)
        except Exception, e:
            if opt_debug:
                raise
            bail_out("Cannot extract cap file: %s" % e)


def cap_usage():
    bail_out("Usage: cmk --cap pack|unpack FILENAME.cap [DIR]")


# Pack a directory and return the byte stream of the CAP.
def cap_pack(install_basedir, verb=False):
    cap = ""
    old_cwd = os.path.abspath('.')
    os.chdir(install_basedir)
    for relative_dir, dirs, files in os.walk("."):
        for filename in files:
            path = (relative_dir + "/" + filename)[2:]
            cap += cap_entry(path, verb)
    os.chdir(old_cwd)
    return cap


def cap_entry(relpath, verb):
    entry = chr(len(relpath)) + relpath
    content = file(relpath).read()
    entry += cap_filesize(len(content)) + content
    if verb:
        sys.stdout.write("%s (%d)\n" % (relpath, len(content)))
    return entry


def cap_filesize(l):
    return chr(l     & 0xff) + \
           chr(l>>8  & 0xff) + \
           chr(l>>16 & 0xff) + \
           chr(l>>24 & 0xff)


# Unpack into a directory
def cap_unpack(install_basedir, cap, only_list, verb=False):
    old_cwd = os.path.abspath('.')
    os.chdir(install_basedir)
    while cap:
        pathname, file_content, cap = cap_next_entry(cap)
        dirname = os.path.dirname(pathname)
        if dirname and not os.path.exists(dirname):
            os.makedirs(dirname)
        if only_list or verb:
            sys.stdout.write("%s (%d)\n" % (pathname, len(file_content)))
        if not only_list:
            file(pathname, "w").write(file_content)
    os.chdir(old_cwd)


def cap_next_entry(cap):
    pathname_length = ord(cap[0])
    cap = cap[1:]
    pathname = cap[:pathname_length]
    cap = cap[pathname_length:]
    filesize = cap_get_filesize(cap[:4])
    cap = cap[4:]
    file_content = cap[:filesize]
    cap = cap[filesize:]
    return pathname, file_content, cap


def cap_get_filesize(b):
    return ord(b[0]) + \
           (ord(b[1]) * 0x100)+ \
           (ord(b[2]) * 0x10000) + \
           (ord(b[3]) * 0x1000000)

