#!/usr/bin/python
# -*- encoding: utf-8; py-indent-offset: 4 -*-
# .------------------------------------------------------------------------.
# |                ____ _               _        __  __ _  __              |
# |               / ___| |__   ___  ___| | __   |  \/  | |/ /              |
# |              | |   | '_ \ / _ \/ __| |/ /   | |\/| | ' /               |
# |              | |___| | | |  __/ (__|   <    | |  | | . \               |
# |               \____|_| |_|\___|\___|_|\_\___|_|  |_|_|\_\              |
# |                                        |_____|                         |
# |             _____       _                       _                      |
# |            | ____|_ __ | |_ ___ _ __ _ __  _ __(_)___  ___             |
# |            |  _| | '_ \| __/ _ \ '__| '_ \| '__| / __|/ _ \            |
# |            | |___| | | | ||  __/ |  | |_) | |  | \__ \  __/            |
# |            |_____|_| |_|\__\___|_|  | .__/|_|  |_|___/\___|            |
# |                                     |_|                                |
# |                     _____    _ _ _   _                                 |
# |                    | ____|__| (_) |_(_) ___  _ __                      |
# |                    |  _| / _` | | __| |/ _ \| '_ \                     |
# |                    | |__| (_| | | |_| | (_) | | | |                    |
# |                    |_____\__,_|_|\__|_|\___/|_| |_|                    |
# |                                                                        |
# | mathias-kettner.com                                 mathias-kettner.de |
# '------------------------------------------------------------------------'
#  This file is part of the Check_MK Enterprise Edition (CEE).
#  Copyright by Mathias Kettner and Mathias Kettner GmbH.  All rights reserved.
#
#  Distributed under the Check_MK Enterprise License.
#
#  You should have  received  a copy of the Check_MK Enterprise License
#  along with Check_MK. If not, email to mk@mathias-kettner.de
#  or write to the postal address provided at www.mathias-kettner.de

from struct import *

# Defaults settings for global configuration
cmc_check_helpers            = 5
cmc_cmk_helpers              = 20
cmc_pnp_update_delay         = 3600
cmc_pnp_update_on_restart    = False
cmc_state_retention_interval = 600 # sec
cmc_log_rotation_method      = 3 # weekly
cmc_log_limit                = 20 * 1024 * 1024 # for rotation at this size
cmc_check_timeout            = 60
cmc_flap_settings            = (3.0, 5.0, 0.1) # low/high threshold, decay per minute
cmc_initial_scheduling       = { "burst" : 10, "spread_cmk" : 1200, "spread_generic" : 150 }
cmc_housekeeping_interval    = 1000 # ms
cmc_smartping_check_interval = 500 # ms
cmc_smartping_settings       = [] # This is a ruleset for hosts
cmc_smartping_tuning         = { "omit_payload" : False, "num_sockets" : 8, "ignore_rst" : False }
cmc_dump_core                = False
cmc_log_level                = 5
cmc_log_cmk_helpers          = { "log_level": 0, "debug": False }
cmc_livestatus_debug         = 0
cmc_log_microtime            = False
cmc_livestatus_threads       = 20
cmc_livestatus_logcache_size = 500000
cmc_livestatus_lines_per_file = 1000000 # maximum allows lines per logfile
cmc_statehist_cache          = { "horizon": 63072000, "max_core_downtime": 30 }
cmc_turbo_mode               = False
cmc_livestatus_delay         = None
cmc_debug_alerts             = False
alert_handler_event_types    = [ "statechange" ]
cmc_debug_notifications      = False
cmc_authorization            = { "host": 0, "group": 0 }
cmc_timeperiod_horizon       = 2 * 365
cmc_import_nagios_state      = True
cmc_graphite                 = [] # was previously None or a single dict
cmc_host_limit               = None # Do not allow more than this number of hosts
cmc_service_limit            = None # Do not allow more than this number of services
shadow_hosts                 = {}
cmc_real_time_checks         = None

cmc_smartping_omit_payload   = False # No longer used, here for compatibility reasons

# Match rrd_format_t in CMC
RRD_FORMAT_NONE         = 0
RRD_FORMAT_PNP_MULTIPLE = 1
RRD_FORMAT_CMC_SINGLE   = 2

# Features of CEE that do not (only) belong to the core
host_recurring_downtimes     = []
service_recurring_downtimes  = []
alert_handler_rules          = []

alert_handler_event_type_bits = {
    "statechange" : 1,
    "checkresult" : 2,
}

def do_create_cmc_config(relfilename):
    if opt_verbose:
        sys.stdout.write("\n")

    # Open output file for writing right now, at the beginning.
    # That way others will see that we are already working on it.
    cmc_dir = var_dir + "/core"
    if not os.path.exists(cmc_dir):
        os.makedirs(cmc_dir)
    out_path = "%s/%s" % (cmc_dir, relfilename)
    out = file(out_path + ".new", "w")

    load_resource_cfg()

    global define_contactgroups
    if define_contactgroups == None:
        define_contactgroups = {}

    if not omd_root:
        raise MKGeneralException("Sorry. You cannot use the Check_MK Micro Core without OMD.")

    global cmc_strings, contactlists_to_define, commands_to_define, cmc_contactgroups
    cmc_strings = {}
    contactlists_to_define = {}
    cmc_contactgroups = {} # Cache for mapping group->contacts
    commands_to_define = set([])

    # Note: we need to *compute* the configuration not in the
    # same order as we need to *output* it. Therefore some
    # temporary variables for storing part of the config

    global_config            = cmc_global()
    timeperiods_config       = cmc_timeperiods()
    contacts_config          = cmc_contacts() # makes also sure define_contactgroups is complete
    hosts_config             = cmc_all_hosts()
    object_groups_config     = cmc_groups(define_hostgroups)
    object_groups_config    += cmc_groups(define_servicegroups)
    contact_groups_config    = cmc_groups(define_contactgroups)
    commands_config          = cmc_stringlist(commands_to_define)
    contactlists_config      = cmc_contactlists()

    out.write(cmc_string_config())
    out.write(global_config)
    out.write(timeperiods_config)
    out.write(contact_groups_config)
    out.write(contacts_config)
    out.write(contactlists_config)
    out.write(hosts_config)
    out.write(object_groups_config)
    out.write(commands_config)

    out.close()

    if cmc_import_nagios_state:
        import_nagios_state()

    os.rename(out_path + ".new", out_path)
    if opt_verbose:
        sys.stdout.write("%s written.\n" % out_path)


def cmc_global():
    con  = cmc_string(var_dir + "/core")                    # data directory
    con += cmc_string(omd_root + "/tmp/run/cmc.pid")        # PID file path
    con += cmc_string(omd_root + "/var/log/cmc.log")        # daemon log file
    con += cmc_string(omd_root + "/tmp/run/nagios.cmd")     # Command pipe path
    con += cmc_int(cmc_log_level)                           # log level for daemon log
    con += cmc_int(cmc_log_cmk_helpers["log_level"])        # 0: off, 1: moderate, 2: full
    con += cmc_bool(cmc_log_cmk_helpers["debug"])
    con += cmc_int(cmc_livestatus_debug)                    # Debug level for Livestatus (0, 1, 2)
    con += cmc_bool(cmc_log_microtime)
    con += cmc_bool(cmc_dump_core)                          # dump core files on segfault
    con += cmc_int(cmc_check_timeout)                       # in seconds
    con += cmc_int(cmc_housekeeping_interval * 1000)        # Housekeeping interval in microseconds
    con += cmc_int(cmc_smartping_check_interval * 1000)     # Smartping check interval in microseconds
    con += cmc_bool(cmc_smartping_tuning["omit_payload"])   # Do not put payload into ICMP packet
    con += cmc_int(cmc_smartping_tuning["num_sockets"])
    con += cmc_bool(cmc_smartping_tuning.get("ignore_rst", False))
    throttling = cmc_smartping_tuning.get("throttling", (0, 1000))
    con += cmc_int(throttling[0])
    con += cmc_int(throttling[1])
    con += cmc_int(cmc_state_retention_interval)
    con += cmc_int(cmc_log_rotation_method)
    con += cmc_int(cmc_log_limit)
    con += cmc_string(var_dir + "/core/history")            # monitoring history
    con += cmc_string(var_dir + "/core/archive")            # monitoring history archive
    con += cmc_string(omd_root + "/lib/cmc/checkhelper")    # check helper
    con += cmc_string(omd_root + "/lib/cmc/icmpsender")     # PING sender helper
    con += cmc_string(omd_root + "/lib/cmc/icmpreceiver")   # PING receiver helper
    con += cmc_string(omd_root + "/bin/cmk")                # Check_MK binary
    con += cmc_int(cmc_check_helpers)                       # Number of generic check helpers
    con += cmc_int(cmc_cmk_helpers)                         # Number of cmk check helpers
    con += cmc_bool(cmc_turbo_mode)                         # Schedule as fast as possible
    if cmc_livestatus_delay:
        con += cmc_int(cmc_livestatus_delay["delay"])
        con += cmc_int(cmc_livestatus_delay["randomize"])
    else:
        con += cmc_int(0)
        con += cmc_int(0)

    con += cmc_bool(cmc_debug_alerts)
    con += cmc_bool(cmc_debug_notifications)
    con += cmc_bool(enable_rulebased_notifications)
    con += cmc_bool(len(alert_handler_rules) > 0)           # enable alert handlers if there are rules
    con += cmc_alert_handler_event_types()
    con += cmc_int(cmc_authorization["host"])               # 0: loose, 1: strict
    con += cmc_int(cmc_authorization["group"])              # 0: loose, 1: strict
    con += cmc_double(cmc_flap_settings[0])                 # low flap threshold
    con += cmc_double(cmc_flap_settings[1])                 # high flap threshold
    con += cmc_double(cmc_flap_settings[2])                 # flap decay
    con += cmc_int(cmc_initial_scheduling["burst"])
    con += cmc_int(cmc_initial_scheduling["spread_cmk"])
    con += cmc_int(cmc_initial_scheduling["spread_generic"])

    # processing of perf data
    # Beware: due to a bug in rrdcached (or is it a nastiness in PNP4Nagios?)
    # we cannot use the correct logical rrd_path with "/omd/sites/...". we
    # need to use "/opt/omd/sites/...". Otherwise our UPDATE commands will have
    # different literal path names then the FLUSH commands from PNP4Nagios and
    # the rrdcached will not flush.
    con += cmc_string("/opt/omd/sites/" + omd_site + "/var/pnp4nagios/perfdata") # base directory for RRDs
    con += cmc_int(cmc_pnp_update_delay)                                         # XML_UPDATE_DELAY for PNP
    con += cmc_bool(cmc_pnp_update_on_restart)                                   # Enable/disable update after reload/restart
    con += cmc_string("/opt/omd/sites/" + omd_site + "/var/check_mk/inventory")  # base directory for HW/SW-Inventory
    con += cmc_string(omd_root + "/tmp/run/rrdcached.sock")                      # Path to RRD cache socket
    con += cmc_string("/opt/omd/sites/" + omd_site + "/var/check_mk/rrd")       # Base for new CMC style RRD files

    # Graphite integration
    con += cmc_graphite_config()

    # Real-Time Checks
    if cmc_real_time_checks:
        con += cmc_int(cmc_real_time_checks["port"])
    else:
        con += cmc_int(0)

    # Livestatus
    con += cmc_string(livestatus_unix_socket)               # livestatus socket
    con += cmc_int(65536)                                   # Livestatus: stack size
    con += cmc_int(cmc_livestatus_threads)                  # Number of Livestatus threads
    con += cmc_int(cmc_livestatus_logcache_size)            # Livestatus: max cached log messages
    con += cmc_int(cmc_livestatus_lines_per_file)           # Livestatus: never read more than that number of lines

    if cmc_statehist_cache:
        con += cmc_int(cmc_statehist_cache["horizon"])      # Horizon for new statehist cache
        con += cmc_int(cmc_statehist_cache["max_core_downtime"]) # Tolerate core downtimes shorted than this
        con += cmc_int(cmc_statehist_cache.get("tarpit", 0)) # Artificial slowdown of parsing
    else:
        con += cmc_int(0)
        con += cmc_int(0)
        con += cmc_int(0)
    return con


def cmc_graphite_config():
    # convert legacy single-connection configuration which was None/dict
    if cmc_graphite == None:
        connections = []
    elif type(cmc_graphite) == dict:
        connections = [cmc_graphite]
    else:
        connections = cmc_graphite

    con = cmc_int(len(connections))
    for connection in connections:
        con += parse_ipv4_address(lookup_ipv4_address(connection["host"]))
        con += cmc_int(connection["port"])
        con += cmc_string(connection["prefix"])

    return con


def cmc_alert_handler_event_types():
    bitmask = 0
    for event_type in alert_handler_event_types:
        bitmask += alert_handler_event_type_bits[event_type]
    return cmc_int(bitmask)

def cmc_timeperiods():
    # Compute DST change offset
    global dst_shift, g_now
    g_now = int(time.time())
    # is_dst = time.localtime(g_now).tm_isdst

    # dst_shift: seconds shifted to the west when DST is on
    if time.daylight:
        dst_shift = time.timezone - time.altzone
    else:
        dst_shift = 0

    con = ""
    # Compute timeperiods in a bottom-up order. They refer to
    # each other.

    sequences = {}
    tps_to_do = timeperiods.keys()
    while tps_to_do:
        tpname = tps_to_do[0]
        serialize_timeperiod(tpname, sequences)
        tps_to_do = [ tpname for tpname in tps_to_do if tpname not in sequences ]

    if "24X7" not in sequences:
        sequences["24X7"] = [ 0 ]
        timeperiods["24X7"] = { "alias" : "Always" }

    con += cmc_int(len(sequences))
    for tpname, sequence in sequences.iteritems():
        con += cmc_string(tpname)
        con += cmc_string(timeperiods[tpname].get("alias", tpname))
        con += cmc_int(len(sequence))
        for x in sequence:
            con += cmc_int(x)
    return con

def cmc_groups(d):
    if type(d) != dict:
        return cmc_int(0)
    else:
        con = cmc_int(len(d))
        for name, alias in d.iteritems():
            con += cmc_string(name)
            con += cmc_string(alias)
        return con

# Example for resource.cfg:
# ############################################
# # OMD settings, please use them to make your config
# # portable, but dont change them
# $USER1$=/omd/sites/prod/lib/nagios/plugins
# $USER2$=/omd/sites/prod/local/lib/nagios/plugins
# $USER3$=prod
# $USER4$=/omd/sites/prod
# ############################################
# # set your own macros here:
# $USER32$=wrdlpfrmpt

def load_resource_cfg():
    global resource_macros
    resource_macros = {}

    nagios_etc_dir = omd_root + "/etc/nagios"
    path = nagios_etc_dir + "/resource.cfg"
    if os.path.exists(path):
        try:
            for line in file(path):
                line = line.strip()
                if not line or line[0] == '#':
                    continue
                varname, value = line.split('=', 1)
                resource_macros[varname] = value
        except Exception, e:
            configuration_warning("Cannot parse %s: %s" % (path, e))


def replace_macros(s):
    if type(s) == str:
        s = s.decode("utf-8")

    for key, value in macros.iteritems():
        if type(value) in (int, long, float):
            value = str(value) # e.g. in _EC_SL (service level)
        try:
            s = s.replace(key, value)
        except: # Might have failed due to binary UTF-8 encoding in value
            try:
                s = s.replace(key, value.decode("utf-8"))
            except:
                # If this does not help, do not replace
                if opt_debug:
                    raise

    for varname, value in resource_macros.iteritems():
        s = s.replace(varname, value)
    return s


def expand_macros_in(attrs):
    for what in [ "notes_url", "action_url", "notes", "icon_image" ]:
        if what in attrs:
            attrs[what + "_expanded"] = replace_macros(attrs[what])

# Attributes that must be removed from the extra conf, because
# they are already send as explicit values.
explicit_attributes = set([
   "max_check_attempts",
   "check_interval",
   "notification_interval",
   "first_notification_delay",
   "check_period",
   "service_period",
   "notification_period",
   "retry_interval",
   "notifications_enabled",
   "flap_detection_enabled",
])

def cmc_all_hosts():
    global g_cmc_num_hosts, g_cmc_parents
    g_cmc_num_hosts = 0
    g_cmc_parents = []
    hosts_config = cmc_hosts() + cmc_shadow_hosts()
    return cmc_int(g_cmc_num_hosts) + hosts_config + cmc_parents()


def cmc_hosts():
    global g_cmc_num_services, g_cmc_num_hosts
    g_cmc_num_services = 0

    con = ""
    hostnames = all_active_hosts()
    if cmc_host_limit != None and len(hostnames) > cmc_host_limit:
        raise MKGeneralException("There is a limit for monitoring up to %d hosts. "
                                 "Your configuration contains %d hosts." % (cmc_host_limit, len(hostnames)))

    for hostname in hostnames:
        g_cmc_num_hosts += 1
        con += cmc_host(hostname)

    if cmc_service_limit != None and g_cmc_num_services > cmc_service_limit:
        raise MKGeneralException("There is a limit for monitoring up to %d services. "
                                 "Your configuration contains %d services." % (cmc_service_limit, g_cmc_num_services))

    return con


def cmc_parents():
    con = ""
    # parent declarations afterwards - avoid dependency problems
    hostset = set(all_active_hosts() + shadow_hosts.keys())
    con += cmc_int(len(g_cmc_parents))
    for hostname, parentlist in g_cmc_parents:
        if type(parentlist) == list:
            parents = parentlist # shadow hosts create lists
        else:
            parents = parentlist.split(",")
        for parent in parents:
            if parent not in hostset:
                configuration_warning("The parent '%s' of the host '%s' is "
                                      "not defined on this site." % (parent, hostname))
                parents.remove(parent)
        con += cmc_string(hostname)
        con += cmc_stringlist(parents)
    return con


def cmc_host(hostname):
    con = ""

    is_clust = is_cluster(hostname)

    attrs = get_host_attributes(hostname)

    tags = tags_of_host(hostname)
    attrs["_TAGS"] = " ".join(tags)
    path = host_paths.get(hostname)
    if path:
        attrs["_FILENAME"] = path

    # Add custom user icons and actions
    actions = icons_and_actions_of('host', hostname)
    if actions:
        attrs["_ACTIONS"] = ','.join(actions)

    global macros
    macros = get_basic_host_macros_from_attributes(hostname, attrs)

    if not is_clust:
        parents_list = attrs.get("parents", "")
        if parents_list:
            g_cmc_parents.append((hostname, parents_list))

    else:
        parents_list = get_cluster_nodes_for_config(hostname)
        attrs.setdefault("alias", "cluster of %s" % ", ".join(parents_list))

        attrs.update(get_cluster_attributes(hostname, parents_list))
        for suffix in [ "", "_4", "_6" ]:
            macros["$_NODEIPS%s$" % suffix] = attrs["_NODEIPS%s" % suffix]
            macros["$_HOSTNODEIPS%s$" % suffix] = attrs["_NODEIPS%s" % suffix]

        g_cmc_parents.append((hostname, ",".join(parents_list)))


    global hostcheck_commands_to_define
    hostcheck_commands_to_define = []
    cc_definition = host_check_command(hostname, attrs["address"], is_clust)
    ipv6_primary = is_ipv6_primary(hostname)
    if cc_definition == None or ipv6_primary:
        base_cmd = "check-mk-host-ping"
        if is_clust:
            base_cmd += "-cluster"

        if ipv6_primary:
            cc_definition = base_cmd + "!" + check_icmp_arguments_of(hostname, add_defaults=True)
        else:
            cc_definition = base_cmd

    if '!' in cc_definition:
        cc_name, cc_args = cc_definition.split('!', 1)
    else:
        cc_name = cc_definition
        cc_args = None

    cc_line = command_command_line(cc_name, cc_args, use_smart_ping_host(hostname, attrs))

    attrs["check_command"] = cc_name
    commands_to_define.add(cc_name)
    attrs["check_command_expanded"] = cc_name + "!" + (cc_args or "")
    process_perfdata = not cc_line.startswith('@')
    if cc_line == "@smartping":
        default_check_interval = 0.1
    else:
        default_check_interval = 1.0

    attrs.setdefault("display_name", attrs["alias"])

    # Groups
    hgs = hostgroups_of(hostname)
    if not hgs:
        hgs = [ default_host_group ]
    cgrs = host_contactgroups_of([hostname])
    contactlist_id = cmc_contactlist(cgrs)

    max_check_attempts       = int(attrs.get("max_check_attempts", 1))
    check_interval           = float(attrs.get("check_interval", default_check_interval))
    retry_interval           = float(attrs.get("retry_interval", default_check_interval))
    notification_interval    = float(attrs.get("notification_interval", 0.0))
    first_notification_delay = float(attrs.get("first_notification_delay", 0.0))

    # Validate referenced timeperiods
    tp_names = timeperiods.keys() + [ "24X7" ]
    for what in ( 'check', 'service', 'notification' ):
        varname = what + "_period"
        period = attrs.get(varname)
        if period and period not in tp_names:
            configuration_warning("The %s period '%s' used by host '%s' does not exist. Falling back to 24x7" %
                                  (what, period, hostname))
            del attrs[varname]

    check_period             = attrs.get("check_period", "24X7")
    service_period           = attrs.get("service_period", "24X7")
    notification_period      = attrs.get("notification_period", "24X7")


    notification_options     = notification_option_bitmask(attrs.get("notification_options"), "")
    notifications_enabled    = bool(int(attrs.get("notifications_enabled", True)))
    flap_detection_enabled   = bool(int(attrs.get("flap_detection_enabled", True)))
    expand_macros_in(attrs)

    # Smartping settings
    smartping_settings = { "timeout" : 2.5 }
    smartping_settings.update(host_extra_conf_merged(hostname, cmc_smartping_settings))
    smartping_timeout = smartping_settings["timeout"]
    if type(smartping_timeout) == float:
        smartping_timeout = int(60 * 1000 * check_interval * smartping_timeout) # in ms
    else:
        smartping_timeout = smartping_timeout * 1000 # in ms

    # The order of the elements is fixed an well known

    # Base information (C++ class Entity)
    con += cmc_string(hostname)
    con += cmc_stringdict(attrs, explicit_attributes)

    # Specific information (C++ class Object)
    con += cmc_bool(False) # real object, no shadow
    con += cmc_string(cc_name)
    con += cmc_string(cc_line)
    con += cmc_bool(notifications_enabled)
    con += cmc_bool(flap_detection_enabled)
    con += cmc_bool(process_perfdata)
    con += cmc_double(check_interval)
    con += cmc_double(retry_interval)
    con += cmc_double(notification_interval)
    con += cmc_double(first_notification_delay)
    con += cmc_int(max_check_attempts)
    con += cmc_string(check_period)
    con += cmc_string(service_period)
    con += cmc_string(notification_period)
    con += cmc_int(notification_options)
    con += cmc_int(0) # freshness state, not supported for hosts
    con += cmc_string("") # freshness output, not supported for hosts
    con += cmc_rrd_format(hostname)
    con += cmc_stringlist(hgs)  # host groups
    con += cmc_int(contactlist_id)
    con += cmc_recurring_downtimes(hostname)
    con += cmc_int(smartping_timeout)
    con += get_cmc_ipv4_address(attrs["_ADDRESS_4"])
    con += cmc_services(hostname, tags, attrs, contactlist_id)
    return con


def use_smart_ping_host(hostname, attrs):
    ipv6_primary = is_ipv6_primary(hostname)
    if ipv6_primary:
        # SmartPING can not be used with IPV6 at the moment, sorry.
        return False
    else:
        try:
            parse_ipv4_address(attrs["_ADDRESS_4"])
            return True
        except ValueError:
            # This seems to be a host- or dns name. Fallback to check_icmp
            return False


def get_cmc_ipv4_address(ipv4_address):
    try:
        return parse_ipv4_address(ipv4_address)
    except ValueError:
        return "\0\0\0\0"


def cmc_services(hostname, tags, host_attrs, host_contactlistid):
    global g_cmc_num_services

    if service_dependencies:
        configuration_warning("Sorry, the Check_MK Micro Core does not support service_dependencies. They are ignored.")

    con = ""

    host_checks = get_check_table(hostname, remove_duplicates=True).items()
    host_checks.sort() # Create deterministic order
    have_at_least_one_service = False
    used_descriptions = {}

    def do_omit_service(hostname, description):
        if service_ignored(hostname, None, description):
            return True
        if hostname != host_of_clustered_service(hostname, description):
            return True
        return False

    # Since we need to send the number of services first, we store all
    # services in a list before serializing them. This list has tuples
    # with the following entries:
    # 1. Service description
    # 2. command name
    # 3. command line (only for active checks, otherwise None)
    # 4. has_perfdata (True or False)
    # 5. check_interval
    # 6. additional attributes
    services_for_cmc = []

    check_mk_attrs = extra_service_attributes(hostname, "Check_MK")
    default_check_interval = interval_of(check_mk_attrs)

    # - 1 - Check_MK Services
    for ((checkname, item), (params, description, deps)) in host_checks:
        if checkname not in check_info:
            configuration_warning("Skipping unimplemented check type %s on host %s" % \
                  (checkname, hostname))
            continue

        # Make sure, the service description is unique on this host
        if description in used_descriptions:
            cn, it = used_descriptions[description]
            configuration_warning("Duplicate service description '%s' for host '%s'. "
                                  "First occurrance: checktype = %s, item = %r, "
                                  "Second occurrance: checktype = %s, item = %r" %
                    (description, hostname, cn, it, checkname, item))

            continue

        used_descriptions[description] = ( checkname, item )

        # Add the check interval of either the Check_MK service or
        # (if configured) the snmp_check_interval for snmp based checks
        check_interval = default_check_interval
        value = check_interval_of(hostname, checkname)
        if value is not None:
            check_interval = value

        # string dict with extra configuration
        attrs = extra_service_attributes(hostname, description, checkname, params)
        if checkname == "logwatch":
            attrs["notes_url"] = logwatch_notes_url % (urllib.quote(hostname), urllib.quote(make_utf8(item)))

        expand_macros_in(attrs)
        command = "check_mk-" + checkname
        commands_to_define.add(command)

        services_for_cmc.append((description, command, "", check_info[checkname].get("has_perfdata", False),
                   check_interval, None, attrs))

    # Active check for check_mk
    if services_for_cmc:
        commands_to_define.add("check-mk")
        # "@cmk ..." use special check helper
        services_for_cmc.append(("Check_MK", "check-mk", "@cmk %s %s" % (hostname, host_attrs["address"]),
                                 1, default_check_interval, None, check_mk_attrs))

    # Legacy checks are not supported by CMC!
    legchecks = host_extra_conf(hostname, legacy_checks)
    if legchecks:
        configuration_warning("Sorry, the Check_MK Micro Core does not support legacy_checks. "
                           "Please use custom_checks or active_checks instead.")

    # active_checks
    actchecks = []
    for check_type, rules in active_checks.items():
        entries = host_extra_conf(hostname, rules)
        if entries:
            # Skip Check_MK HW/SW Inventory for all ping hosts, even when the user has enabled
            # the inventory for ping only hosts
            if check_type == "cmk_inv" and is_ping_host(hostname):
                continue

            act_info = active_check_info[check_type]
            for params in entries:
                # Make hostname available as global variable in argument functions
                global g_hostname
                g_hostname = hostname

                has_perfdata = act_info.get('has_perfdata', False)
                description = active_check_service_description(act_info, params)
                if do_omit_service(hostname, description):
                    continue

                args = act_info["argument_function"](params)

                if description in used_descriptions:
                    cn, it = used_descriptions[description]
                    # If we have the same active check again with the same description,
                    # then we do not regard this as an error, but simply ignore the
                    # second one. That way one can override a check with other settings.
                    if cn == "active(%s)" % check_type:
                        continue

                    configuration_warning(
                            "Ignoring duplicat active service '%s' for host '%s'!\n"
                            " - 1st occurrance: checktype = %s, item = %r\n"
                            " - 2nd occurrance: checktype = active(%s), item = None\n" %
                            (description, hostname, cn, it, check_type))
                    continue

                else:
                    used_descriptions[description] = ( "active(" + check_type+ ")", description )

                command_line = replace_macros(act_info["command_line"].replace("$ARG1$", args))
                attrs = extra_service_attributes(hostname, description)

                if host_attrs["address"] in [ "0.0.0.0", "::" ]:
                    command = "check-mk-custom"
                    command_line = "echo \"Failed to lookup IP address and no explicit IP address configured\" ; exit 3"
                else:
                    command = "check_mk_active-%s" % check_type

                commands_to_define.add(command)
                attrs["check_command_expanded"] = command + "!" + replace_macros(args)
                services_for_cmc.append((description, command + "!" + args, command_line,
                                         has_perfdata, interval_of(attrs), None, attrs))

    # Legacy checks via custom_checks
    for entry in host_extra_conf(hostname, custom_checks):
        # entries are dicts with the following keys:
        # "service_description"        Service description to use
        # "command_line"  (optional)   Unix command line for executing the check
        #                              If this is missing, we create a passive check
        # "command_name"  (optional)   Name of Monitoring command to define. If missing,
        #                              we use "check-mk-custom"
        # "has_perfdata"  (optional)   If present and True, we activate perf_data
        description = sanitize_service_description(entry["service_description"])
        command_name = entry.get("command_name", "check-mk-custom")
        if description in used_descriptions:
            cn, it = used_descriptions[description]
            # If we have the same active check again with the same description,
            # then we do not regard this as an error, but simply ignore the
            # second one.
            if cn == "custom(%s)" % command_name:
                continue
            configuration_warning(
                    "ERROR: Duplicate service description (custom check) '%s' for host '%s'! "
                    "First occurrance: checktype = %s, item = %r "
                    "Second occurrance: checktype = custom(%s), item = %r" %
                    (description, hostname, cn, it, command_name, description))
            continue

        else:
            used_descriptions[description] = ( "custom(%s)" % command_name, description )

        if do_omit_service(hostname, description):
            continue

        has_perfdata = entry.get("has_perfdata", False)
        command_line = entry.get("command_line", "")
        command_line_expanded = replace_macros(command_line)

        if command_line:
            plugin_name = command_line_expanded.split()[0]
            if not command_line_expanded.startswith("/"):
                try:
                    for dir in [ "/local", "" ]:
                        path = omd_root + dir + "/lib/nagios/plugins/"
                        if os.path.exists(path + plugin_name):
                            command_line_expanded = path + command_line_expanded
                            break
                except:
                    pass

        if "freshness" in entry:
            fr_interval = entry["freshness"]["interval"]
            fr_state    = entry["freshness"]["state"]
            fr_output   = entry["freshness"]["output"]
            freshness = (fr_interval, fr_state, fr_output)
        else:
            freshness = None


        attrs = extra_service_attributes(hostname, description)
        attrs["check_command_expanded"] = command_name + "!" + command_line_expanded
        commands_to_define.add(command_name)
        services_for_cmc.append((description, command_name + "!" + command_line, command_line_expanded, has_perfdata,
                                interval_of(attrs), freshness, attrs))

    # No check_mk service, no legacy service -> create PING service
    if not services_for_cmc:
        cmc_add_ping_service(hostname, host_attrs["address"],
                             is_ipv6_primary(hostname) and 6 or 4,
                            "PING", host_attrs.get("_NODEIPS"), services_for_cmc)

    # Dual-stack host, create a PING service for the secondary address family
    if is_ipv4v6_host(hostname):
        if is_ipv6_primary(hostname):
            cmc_add_ping_service(hostname, host_attrs["_ADDRESS_4"], 4, "PING IPv4", host_attrs.get("_NODEIPS_4"), services_for_cmc)
        else:
            cmc_add_ping_service(hostname, host_attrs["_ADDRESS_6"], 6, "PING IPv6", host_attrs.get("_NODEIPS_6"), services_for_cmc)

    # Discovery check - if user has configured them. Not for clusters.
    # Note: now it's also added if there are *no* Check_MK services
    # yet - as long as it's not a ping-only host.

    # FIXME: Remove old name one day
    service_discovery_name = 'Check_MK inventory'
    if 'cmk-inventory' in use_new_descriptions_for:
        service_discovery_name = 'Check_MK Discovery'

    disc_check_params = discovery_check_parameters(hostname)
    if disc_check_params \
        and not service_ignored(hostname, None, service_discovery_name) \
        and "ping" not in tags: # FIXME/TODO: Why not user is_ping_host()?
        attrs = extra_service_attributes(hostname, service_discovery_name)
        attrs["check_command_expanded"] = "check-mk-inventory!"
        commands_to_define.add("check-mk-inventory")
        services_for_cmc.append((service_discovery_name, "check-mk-inventory",
            "@cmk --cache --check-discovery %s %s" % (hostname, host_attrs["address"]), # use special check helper
            0, disc_check_params["check_interval"], None, attrs))

    # Finally output all services
    con += cmc_int(len(services_for_cmc))
    for description, command_name, command_line, has_perfdata, interval, freshness, attrs in services_for_cmc:
        if freshness:
            interval = freshness[0] # is already in minutes
        groups = service_extra_conf(hostname, description, service_groups)
        contact_groups = service_extra_conf(hostname, description, service_contactgroups)
        if contact_groups:
            contactlist_id = cmc_contactlist(contact_groups)
        else:
            contactlist_id = host_contactlistid
        max_check_attempts       = int(attrs.get("max_check_attempts", 1))
        # Make the retry interval default to the normal interval for the "Check_MK Discovery" check
        default_retry_interval = 1.0
        if command_name == 'check-mk-inventory':
            default_retry_interval = interval
        retry_interval           = float(attrs.get("retry_interval", default_retry_interval))
        notification_interval    = float(attrs.get("notification_interval", 0.0))
        first_notification_delay = float(attrs.get("first_notification_delay", 0.0))

        # Validate referenced timeperiods
        tp_names = timeperiods.keys() + [ "24X7" ]
        for what in ( 'check', 'service', 'notification' ):
            varname = what + "_period"
            period = attrs.get(varname)
            if period and (period not in tp_names):
                configuration_warning("The %s period '%s' used by '%s' on host '%s' does not exist. Falling back to 24x7" %
                                      (what, period, description, hostname))
                del attrs[varname]

        check_period             = attrs.get("check_period", "24X7")
        service_period           = attrs.get("service_period", "24X7")
        notification_period      = attrs.get("notification_period", "24X7")

        notification_options     = notification_option_bitmask("", attrs.get("notification_options"))
        notifications_enabled    = bool(int(attrs.get("notifications_enabled", True)))
        flap_detection_enabled   = bool(int(attrs.get("flap_detection_enabled", True)))
        attrs.setdefault("display_name", description)
        macros["$SERVICEDESC$"] = description
        expand_macros_in(attrs)

        g_cmc_num_services += 1

        # Base information (C++ class Entity)
        con += cmc_string(description)
        con += cmc_stringdict(attrs, explicit_attributes)

        process_perfdata = has_perfdata and not attrs.get("process_perf_data") == "0"

        # Specific information (C++ class Object)
        con += cmc_bool(False) # real object, no shadow
        con += cmc_string(command_name)
        con += cmc_string(command_line)
        con += cmc_bool(notifications_enabled)
        con += cmc_bool(flap_detection_enabled)
        con += cmc_bool(process_perfdata)
        con += cmc_double(interval)
        con += cmc_double(retry_interval)
        con += cmc_double(notification_interval)
        con += cmc_double(first_notification_delay)
        con += cmc_int(max_check_attempts)
        con += cmc_string(check_period)
        con += cmc_string(service_period)
        con += cmc_string(notification_period)
        con += cmc_int(notification_options)
        if freshness:
            con += cmc_int(freshness[1]) # state
            con += cmc_string(freshness[2]) # output
        else:
            con += cmc_int(0)
            con += cmc_string("")
        con += cmc_rrd_format(hostname, description)
        con += cmc_stringlist(groups)
        con += cmc_int(contactlist_id)

        con += cmc_recurring_downtimes(hostname, description)
    return con


def cmc_add_ping_service(hostname, ipaddress, family, descr, node_ips, services_for_cmc):
    attrs = extra_service_attributes(hostname, descr)
    arguments = check_icmp_arguments_of(hostname, family=family) + ' '

    command_name = "check-mk-ping"
    if node_ips:
        command_name += "-cluster"
        arguments += ' -m 1 ' + node_ips
    else:
        arguments += ipaddress

    ping_command = omd_root + '/lib/nagios/plugins/check_icmp ' + arguments

    command_expanded = command_name + "!" + arguments
    attrs["check_command_expanded"] = command_expanded

    commands_to_define.add(command_name)
    services_for_cmc.append((descr, command_expanded, ping_command, 1,
                             interval_of(attrs), None, attrs))


def cmc_shadow_hosts():
    global g_cmc_num_hosts

    con = ""
    for host_name, host in shadow_hosts.items():
        g_cmc_num_hosts += 1
        con += cmc_shadow_host(host_name, host)

    return con

def cmc_shadow_host(host_name, host):
    con = cmc_string(host_name)
    con += cmc_shadow_generic_columns(host, host_name, None)

    # Host specific columns
    con += cmc_int(0) # smartping_timeout, not supported
    con += parse_ipv4_address(host["address"])
    con += cmc_shadow_services(host["services"], host_name)
    g_cmc_parents.append((host_name, host["parents"]))

    return con

# Add all columns of C++ class Object
def cmc_shadow_generic_columns(entry, hostname, service_description):
    con  = cmc_stringdict(cmc_shadow_custom_variables(entry))
    con += cmc_bool(True) # shadow object
    con += cmc_string(entry["check_command"])
    con += cmc_string(entry["check_command_expanded"])
    con += cmc_bool(entry["notifications_enabled"]) # notifications_enabled
    con += cmc_bool(entry["flap_detection_enabled"])
    con += cmc_bool(entry["process_performance_data"])
    con += cmc_double(entry["check_interval"])
    con += cmc_double(entry["retry_interval"])
    con += cmc_double(entry["notification_interval"])
    con += cmc_double(entry["first_notification_delay"])
    con += cmc_int(entry["max_check_attempts"])
    con += cmc_string(entry["check_period"])
    con += cmc_string(entry["service_period"])
    con += cmc_string(entry["notification_period"])
    con += cmc_int(0) # notification_options
    con += cmc_int(0) # freshness state, not supported for entrys
    con += cmc_string("") # freshness output, not supported for entrys
    con += cmc_rrd_format(hostname, service_description)
    con += cmc_stringlist(entry["groups"])  # entry groups
    con += cmc_int(cmc_contactlist(entry["contact_groups"]))
    con += cmc_int(0) # recurring_downtimes, not supported
    return con


def cmc_shadow_custom_variables(entry):
    d = {}
    if "alias" in entry:
        d["alias"] = entry["alias"]
    for varname, value in entry["custom_variables"].items():
        d["_" + varname] = value
    return d


def cmc_shadow_services(services, hostname):
    global g_cmc_num_services
    g_cmc_num_services += len(services)
    con = cmc_int(len(services))
    for service in services:
        con += cmc_shadow_service(service, hostname)
    return con


def cmc_shadow_service(service, hostname):
    con = cmc_string(service["description"])
    con += cmc_shadow_generic_columns(service, hostname, service["description"])
    return con


notify_host_bits = [
  ( 'r',  1 ), # UP ("RECOVERY")
  ( 'd',  2 ), # DOWN
  ( 'u',  4 ), # UNREACHABLE
  ( 'f', 16 ), # FLAPPING
  ( 's', 32 ), # DOWNTIME
]

notify_service_bits = [
  ( 'r',  1 * 256 ), # OK ("RECOVERY")
  ( 'w',  2 * 256 ), # WARN
  ( 'c',  4 * 256 ), # CRIT
  ( 'u',  8 * 256 ), # UNKNOWN
  ( 'f', 16 * 256 ), # FLAPPING
  ( 's', 32 * 256 ), # DOWNTIME
]


def cmc_recurring_downtimes(hostname, service_description=None):
    if service_description:
        downtimes = service_extra_conf(hostname, service_description, service_recurring_downtimes)
    else:
        downtimes = host_extra_conf(hostname, host_recurring_downtimes)

    con = cmc_int(len(downtimes))
    for downtime in downtimes:
        con += cmc_recurring_downtime(downtime)
    return con


def cmc_recurring_downtime(downtime):
    con  = cmc_string(downtime["comment"])
    con += cmc_int(int(downtime["start_time"]))
    con += cmc_int(downtime["duration"])
    con += cmc_int(downtime.get("flex_length", 0))
    con += cmc_int(downtime["interval"])
    return con


def cmc_rrd_format(host_name, service_description = None):
    if service_description:
        rrdconf = service_extra_conf(host_name, service_description, cmc_service_rrd_config)
    else:
        rrdconf = host_extra_conf(host_name, cmc_host_rrd_config)
    if rrdconf:
        rrdconf = rrdconf[0]
        if rrdconf.get("format") == "cmc_single":
            return cmc_int(2) # RRD_FORMAT_CMC_SINGLE

    return cmc_int(1) # RRD_FORMAT_PNP_MULTIPLE


def notification_option_bitmask(hostchars, svcchars):
    bitmask = 0
    for char, bit in notify_host_bits:
        if hostchars == None or char in hostchars:
            bitmask += bit
    for char, bit in notify_service_bits:
        if svcchars == None or char in svcchars:
            bitmask += bit
    return bitmask

def cmc_contacts():
    con = ""

    to_define = [ e for e in contacts.iteritems() if e[1].get("contactgroups") ]
    con += cmc_int(len(to_define))
    for name, contact in to_define:
        con += cmc_string(name)

        # Write attribute dictionary
        attr = dict([(k,v) for (k,v) in contact.iteritems()
                if k in [ "alias", "email", "pager"] or k.startswith("address") or k.startswith("_") ])
        con += cmc_stringdict(attr)

        # Contact groups
        groups = contact.get("contactgroups", [])
        con += cmc_stringlist(groups)
        for group in groups:
            cmc_contactgroups.setdefault(group, []).append(name)
            define_contactgroups.setdefault(group, group) # Make sure this group is defined!

        # Notification settings
        con += cmc_bool(contact.get("notifications_enabled", True))
        period = contact.get("notification_period", "24X7")
        if period not in timeperiods:
            configuration_warning("The notification period '%s' used by '%s' does not exist. "
                                  "Falling back to 24x7. " % (period, name))
            period = "24X7"

        con += cmc_string(period)
        con += cmc_int(notification_option_bitmask(
           contact.get("host_notification_options", ""),
           contact.get("service_notification_options", "")))
    return con

# Return the unique contactlist id that belongs to a
# certain combination of contactgroups
def cmc_contactlist(groups):
    groups.sort()
    for g in groups:
        define_contactgroups.setdefault(g, g) # make sure also empty groups exist
    key = tuple(groups)
    if key in contactlists_to_define:
        return contactlists_to_define[key]
    else:
        nextid = len(contactlists_to_define)
        contactlists_to_define[key] = nextid
        return nextid

def cmc_contactlists():
    con = cmc_int(len(contactlists_to_define))
    entries = contactlists_to_define.items()
    entries.sort(cmp = lambda a,b: cmp(a[1], b[1]))
    for groups, listid in entries:
        contactnames = contacts_of_groups(groups)
        con += cmc_stringlist(contactnames)
        con += cmc_stringlist(groups)
    return con

def contacts_of_groups(groups):
    contacts = set([])
    for group in groups:
        contacts.update(cmc_contactgroups.get(group, []))
    return contacts

#   .--Helpers-------------------------------------------------------------.
#   |                  _   _      _                                        |
#   |                 | | | | ___| |_ __   ___ _ __ ___                    |
#   |                 | |_| |/ _ \ | '_ \ / _ \ '__/ __|                   |
#   |                 |  _  |  __/ | |_) |  __/ |  \__ \                   |
#   |                 |_| |_|\___|_| .__/ \___|_|  |___/                   |
#   |                              |_|                                     |
#   +----------------------------------------------------------------------+
#   |  Helper functions                                                    |
#   '----------------------------------------------------------------------'

def cmc_uchar(t):
    return pack('B', t)

def cmc_int(u):
    return pack('I', u)

def cmc_bool(u):
    return pack('I', u and 1 or 0)

def cmc_double(d):
    return pack('d', d)

def cmc_stringdict(d, exclude=None):
    con = ""
    if exclude:
        for e in exclude.intersection(d.keys()):
            del d[e]
    con += cmc_int(len(d))
    for k, v in d.iteritems():
        con += cmc_string(k)
        if type(v) in [ int, float ]:
            v = str(v)
        elif type(v) not in [ str, unicode ]:
            raise Exception("Invalid type \"%r\" in key \"%s\"" % (v, k))
        con += cmc_string(v)
    return con

def cmc_stringlist(l):
    con = cmc_int(len(l))
    for e in l:
        con += cmc_string(e)
    return con

# Write a string. We do this by postponing the
# configuration output and rather store the string
# in a dictionary with an ID. Identical strings share
# the same ID. We just write the ID into the config.
def cmc_string(s):
    try:
        nr = cmc_strings[s]
    except KeyError:
        nr = len(cmc_strings)
        cmc_strings[s] = nr
    return pack('I', nr)

# Direct inline string, used for creating status file
def cmc_inline_string(s):
    l = len(s)
    data = cmc_int(l) + s
    if l & 3:
        data += "\000" * (4 - l&3)
    return data

# Write list of all strings in order of indices
def cmc_string_config():
    items = cmc_strings.items()
    config = pack('I', len(items))
    items.sort(cmp = lambda a,b: cmp(a[1], b[1]))
    for s, nr in items:
        if type(s) == unicode:
            s = s.encode("utf-8")
        l = len(s)
        config += pack('I', l)
        config += s
        # pad to 4-border
        p = l % 4
        if p:
            config += '\xff\xff\xff'[0:4-p]
    return config


def extra_service_attributes(hostname, description, checkname=None, params=None):
    attrs = {}
    for key, conflist in extra_service_conf.items():
        values = service_extra_conf(hostname, description, conflist)
        if values:
            if key[0] == "_":
                key = key.upper()
            attrs[key] = values[0]

    # Add custom user icons and actions
    actions = icons_and_actions_of('service', hostname, description, checkname, params)
    if actions:
        attrs["_ACTIONS"] = ','.join(actions)

    return attrs


def interval_of(attrs):
    try:
        return float(attrs['check_interval'])
    except KeyError:
        return 1.0


# Directly taken from check_mk_templates.cfg
command_templates = {
    'check-mk-ping'               : omd_root + '/lib/nagios/plugins/check_icmp $ARG1$ $HOSTADDRESS$',
    'check-mk-ping-cluster'       : omd_root + '/lib/nagios/plugins/check_icmp -m 1 $ARG1$ $_HOSTNODEIPS$',
    'check-mk-host-ping'          : omd_root + '/lib/nagios/plugins/check_icmp $ARG1$ $HOSTADDRESS$',
    'check-mk-host-smart'         : omd_root + '/lib/nagios/plugins/check_icmp $ARG1$ $HOSTADDRESS$',
    'check-mk-host-ping-cluster'  : "@parents",
    'check-mk-host-ok'            : "@up",
    'check-mk-host-tcp'           : omd_root + '/lib/nagios/plugins/check_tcp -H $HOSTADDRESS$ -p $ARG1$',
    'check-mk-host-service'       : omd_root + '/lib/nagios/plugins/check_lql_service "$HOSTNAME$" "$ARG1$"',
}

def command_command_line(command, extraargs, use_smart_ping):
    if extraargs == None:
        extraargs = ""

    if command == "check-mk-host-smart" and use_smart_ping:
        return "@smartping"
    elif command == "check-mk-custom":
        return replace_macros(extraargs)
    else:
        return replace_macros(command_templates[command]).replace("$ARG1$", extraargs)

def parse_ipv4_address(a):
    a = a.strip()
    if not a:
        return '\0\0\0\0'
    address = "".join(map(chr, map(int, a.split('.'))))
    if len(address) != 4:
        configuration_warning("Invalid IP address '%s'. Falling back to 0.0.0.0" % a)
        return '\0\0\0\0'
    return address


#.
#   .--Timeperiods---------------------------------------------------------.
#   |      _____ _                                _           _            |
#   |     |_   _(_)_ __ ___   ___ _ __   ___ _ __(_) ___   __| |___        |
#   |       | | | | '_ ` _ \ / _ \ '_ \ / _ \ '__| |/ _ \ / _` / __|       |
#   |       | | | | | | | | |  __/ |_) |  __/ |  | | (_) | (_| \__ \       |
#   |       |_| |_|_| |_| |_|\___| .__/ \___|_|  |_|\___/ \__,_|___/       |
#   |                            |_|                                       |
#   +----------------------------------------------------------------------+
#   |  Computation of timeperiods                                          |
#   '----------------------------------------------------------------------'


# Convert timeperiod definition into a boolean (whether it is
# currently active) and a sequence of timestamps whenever
# the timeperiod toggles.
daynames = [ "monday", "tuesday", "wednesday", "thursday", "friday", "saturday", "sunday" ]

def serialize_timeperiod(tpname, serialized):
    tp = timeperiods[tpname]
    excludes = tp.get("exclude", [])
    ex_sequences = []
    for ex in excludes:
        if ex not in serialized:
            serialize_timeperiod(ex, serialized) # infinite recursion not regarded here
        ex_sequences.append(serialized[ex])
    sequence = timeperiod_sequence_bare(tpname)
    for ex_seq in ex_sequences:
        sequence = timeperiod_substract(sequence, ex_seq)
    serialized[tpname] = sequence

def timeperiod_substract(sequence, ex_sequence):
    result = []
    seq_act = False   # whether the TP is currently active
    ex_act = False    # whether the exclude sequence is currently active
    seq = sequence[:] # do not modify
    ex  = ex_sequence[:]

    while ex or seq:
        # Case 1: next entry is in sequence
        if seq and (not ex or ex[0] >= seq[0]):
            t = seq[0]
            del seq[0]
            seq_act = not seq_act
            if not ex_act:
                if result and result[0] == t:
                    del result[-1] # remove redundancy!
                else:
                    result.append(t)
        elif ex and (not seq or seq[0] >= ex[0]):
            t = ex[0]
            del ex[0]
            ex_act = not ex_act
            if seq_act:
                if result and result[0] == t:
                    del result[-1] # remove redundancy!
                else:
                    result.append(t)

    return result


# Compute bare UTC sequence of timezone. First interprete
# everything without DST. Then shift accordingly
def timeperiod_sequence_bare(tpname):
    timeperiod = timeperiods[tpname]

    per_day = []
    exceptions = {}

    for weekday in daynames:
        per_day.append(day_sequence(timeperiod.get(weekday, [])))

    for key, value in timeperiod.items():
        if key not in daynames and key not in [ "alias", "exclude" ]:
            absdate = parse_date(key)
            exceptions[absdate] = day_sequence(value)

    horizon = g_now + cmc_timeperiod_horizon * 86400

    # We go into the past a bit. If the AV cache is enabled then we go
    # at least that far into the past. If not we go 30 days into the
    # back
    backview = 30 * 24 * 3600
    if cmc_statehist_cache:
        backview = max(backview, cmc_statehist_cache["horizon"])
    startingpoint = g_now - backview
    broken = time.localtime(startingpoint)
    wday = broken.tm_wday
    that_morning_broken = broken[:3] + (0,0,0,0,0) + (0,) # Without DST
    sequence = []
    # t always points to the beginning of a day - to 00:00:00.
    t = int(time.mktime(that_morning_broken))
    while t < horizon:
        exception = exceptions.get(t)
        if exception:
            seq = exception
        else:
            seq = per_day[wday]

        # Join together, if last and first match
        if sequence and seq and sequence[-1] == seq[-1]:
            del sequence[-1]
            sequence += map(lambda x: x + t, seq[1:])
        else:
            sequence += map(lambda x: x + t, seq)

        wday = (wday + 1) % 7
        t += 86400

    # We now have a sequence of timestamps where the TP switches.
    # We begin with OFF.
    shifted = []
    previous = 0
    for x in sequence:
        if timestamp_in_dst(x):
            x -= dst_shift
            if x < previous:
                x = previous # no negative steps!
        previous = x
        shifted.append(x)

    return shifted

def timestamp_in_dst(ts):
    return time.localtime(ts).tm_isdst

# Parse date, interprete without DST
def parse_date(datestr):
    try:
        broken = time.strptime(datestr, "%Y-%m-%d")
        broken = broken[:-1] + (0,)
        return int(time.mktime(broken))
    except:
        configuration_warning("Sorry, the Check_MK Micro Core can handle timeperiod exceptions only "
                              "in the format YYYY-MM-DD currently. Cannot handle '%s', assuming 1970-01-01" % datestr)
        return 0


def day_sequence(hourlist):
    seq = []
    for begin, end in hourlist:
        seq.append(parse_time(begin))
        seq.append(parse_time(end))
    return seq

def parse_time(s):
    hours, minutes = map(int, s.split(":"))
    return hours * 3600 + minutes * 60

#.
#   .--State---------------------------------------------------------------.
#   |                       ____  _        _                               |
#   |                      / ___|| |_ __ _| |_ ___                         |
#   |                      \___ \| __/ _` | __/ _ \                        |
#   |                       ___) | || (_| | ||  __/                        |
#   |                      |____/ \__\__,_|\__\___|                        |
#   |                                                                      |
#   +----------------------------------------------------------------------+
#   |  We import parts of the previous state of the Nagios core here.     |
#   '----------------------------------------------------------------------'

def import_nagios_state():
    retention = omd_root + "/var/nagios/retention.dat"
    if not os.path.exists(retention):
        if opt_verbose:
            sys.stdout.write("Not importing state from Nagios, %s not found.\n" % retention)
        return
    if os.path.exists(var_dir + "/core/state"):
        mtime_config = os.stat(var_dir + "/core/state").st_mtime
        mtime_retention = os.stat(retention).st_mtime
        if mtime_retention < mtime_config:
            if opt_verbose:
                sys.stdout.write("CMC state is newer than retention.dat. Not importing state.\n")
            return
    if opt_verbose:
        sys.stdout.write("Importing previous Nagios state...\n")

    entries = {}
    try:
        entry = {}
        entry_type = None
        for line in file(retention):
            line = line.strip()
            if line[-1] == '{':
                entry_type = line[:-2]
                entry = {}
            elif line == '}':
                entries.setdefault(entry_type, []).append(entry)
                entry_type = None
            elif line[0] != '#':
                varname, value = line.split("=", 1)
                if entry_type:
                    entry[varname] = value

        program_state = entries["program"][0]
        comments = entries.get("hostcomment", []) + entries.get("servicecomment", [])
        downtimes = entries.get("hostdowntime", []) + entries.get("servicedowntime", [])
        state_file_path = var_dir + "/core/state"
        out = file(state_file_path, "w")
        out.write(cmc_int(3))
        out.write(cmc_bool(program_state["enable_notifications"] == "1"))
        out.write(cmc_bool(program_state["active_service_checks_enabled"] == "1"))
        out.write(cmc_bool(program_state["active_host_checks_enabled"] == "1"))
        out.write(cmc_bool(program_state["enable_flap_detection"] == "1"))
        out.write(cmc_bool(program_state["process_performance_data"] == "1"))

        out.write(cmc_int(0)) # Hosts

        # CMC uses one pool of IDs for comments *and* downtimes. We
        # use the original Nagios comment ids - since the downtimes refer
        # to them - and choose new (higher) ids for the downtimes
        max_comment_id = 0
        for comment in comments:
            max_comment_id = max(max_comment_id, int(comment["comment_id"]))
        max_downtime_id = max_comment_id + len(downtimes)
        out.write(cmc_int(max_downtime_id + 1)) # next note ID
        out.write(cmc_int(0)) # next problem ID
        out.write(cmc_int(0)) # last logfile rotation

        # Comments
        out.write(cmc_int(len(comments)))
        for comment in comments:
            out.write(cmc_comment(comment))

        # Downtimes
        out.write(cmc_int(len(downtimes)))
        did = max_comment_id = 1
        for downtime in downtimes:
            out.write(cmc_downtime(did, downtime))
            did += 1

        out.write(cmc_int(0xfea1dead)) # Magic
        out.close()
        if opt_verbose:
            sys.stdout.write(state_file_path + " written.\n")

    except Exception, e:
        if opt_debug:
            raise
        sys.stdout.write("Warning: Cannot import previous Nagios state: %s\n" % e)

def cmc_note(nid, note):
    if "service_description" in note:
        note_type = 2
    else:
        note_type = 1

    object_ref = \
        cmc_inline_string(note["host_name"]) + \
        cmc_inline_string(note.get("service_description", ""))

    note_data = \
        cmc_int(nid) + \
        cmc_int(int(note["entry_time"])) + \
        cmc_inline_string(note["author"]) + \
        cmc_inline_string(note.get("comment", note.get("comment_data"))) + \
        cmc_int(note_type)

    return object_ref + note_data

def cmc_comment(comment):
    comment_data = \
        cmc_int(int(comment["entry_type"])) + \
        cmc_int(int(comment["expires"])) + \
        cmc_int(int(comment["expire_time"])) + \
        cmc_bool(comment["persistent"] == "1") + \
        cmc_bool(comment.get("sticky") == "1") + \
        cmc_int(int(comment["source"]))

    return cmc_note(int(comment["comment_id"]), comment) + comment_data

def cmc_downtime(did, downtime):
    downtime_data = \
        cmc_int(int(downtime["start_time"])) + \
        cmc_int(int(downtime["end_time"])) + \
        cmc_int(int(downtime.get("flex_downtime_start", 0))) + \
        cmc_bool(downtime["fixed"] == "1") + \
        cmc_int(int(downtime["duration"])) + \
        cmc_bool(downtime["is_in_effect"] == "1")

    return cmc_note(did, downtime) + downtime_data


