#!/usr/bin/python
# -*- encoding: utf-8; py-indent-offset: 4 -*-
# .------------------------------------------------------------------------.
# |                ____ _               _        __  __ _  __              |
# |               / ___| |__   ___  ___| | __   |  \/  | |/ /              |
# |              | |   | '_ \ / _ \/ __| |/ /   | |\/| | ' /               |
# |              | |___| | | |  __/ (__|   <    | |  | | . \               |
# |               \____|_| |_|\___|\___|_|\_\___|_|  |_|_|\_\              |
# |                                        |_____|                         |
# |             _____       _                       _                      |
# |            | ____|_ __ | |_ ___ _ __ _ __  _ __(_)___  ___             |
# |            |  _| | '_ \| __/ _ \ '__| '_ \| '__| / __|/ _ \            |
# |            | |___| | | | ||  __/ |  | |_) | |  | \__ \  __/            |
# |            |_____|_| |_|\__\___|_|  | .__/|_|  |_|___/\___|            |
# |                                     |_|                                |
# |                     _____    _ _ _   _                                 |
# |                    | ____|__| (_) |_(_) ___  _ __                      |
# |                    |  _| / _` | | __| |/ _ \| '_ \                     |
# |                    | |__| (_| | | |_| | (_) | | | |                    |
# |                    |_____\__,_|_|\__|_|\___/|_| |_|                    |
# |                                                                        |
# | mathias-kettner.com                                 mathias-kettner.de |
# '------------------------------------------------------------------------'
#  This file is part of the Check_MK Enterprise Edition (CEE).
#  Copyright by Mathias Kettner and Mathias Kettner GmbH.  All rights reserved.
#
#  Distributed under the Check_MK Enterprise License.
#
#  You should have  received  a copy of the Check_MK Enterprise License
#  along with Check_MK. If not, email to mk@mathias-kettner.de
#  or write to the postal address provided at www.mathias-kettner.de

# Implements the more performant inline SNMP mechanism

try:
    import netsnmp
    netsnmp.verbose = 0
    has_inline_snmp = True
except:
    # In this case, inline_snmp is not possible. As we kept the classic snmp code,
    # we can deal with it, so dont raise errors.
    has_inline_snmp = False

import signal

g_snmp_sessions    = {}
g_snmp_stats       = {}
g_snmp_table_cache = {}

def cleanup_inline_snmp_globals():
    global g_snmp_sessions
    g_snmp_sessions = {}
    global g_snmp_table_cache
    g_snmp_table_cache = {}

def deactivate_check_mk_timeout():
    if not g_timeout:
        return
    signal.alarm(0) # Disable alarm

def reactivate_check_mk_timeout():
    if not g_timeout:
        return
    alarm_time_remaining = int(math.ceil(g_timeout - (os.times()[4] - g_initial_times[4])))
    if alarm_time_remaining > 0:
        signal.alarm(alarm_time_remaining)
    else:
        raise MKTimeout()


def oid_range_limits_of(hostname):
    return host_extra_conf(hostname, snmp_limit_oid_range)

def oid_range_limits(hostname, check_type):
    for ty, rules in oid_range_limits_of(hostname):
        if ty == check_type:
            return rules
    return []

def do_show_snmp_stats():
    path = var_dir + '/snmp.stats'
    if not os.path.exists(path):
        sys.stderr.write('The SNMP statistics file %s is missing.\n' % path)
        sys.exit(1)

    stats = dict(eval(file(path).read()))
    def s(k):
        return stats.get(k, 0)

    sys.stdout.write('Started: %s, Ended: %s\n' % (time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(stats['start'])),
                                                 time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(stats['end']))))
    sys.stdout.write('Duration: %d seconds (might be non consecutive)\n' % (stats['end'] - stats['start']))
    sys.stdout.write('\n')

    for title, pre, id in [ ('Bulkwalk', 'bulk-', 'bulkwalks'), ('Walk', 'walk-', 'walks') ]:
        sys.stdout.write(title+'s:\n')
        sys.stdout.write('  Performed:               %10d\n' % s(id))
        if id == 'bulkwalks':
            pkgs_per_walk_avg = s(id) and float(s(pre+'requests'))/s(id) or 0
            sys.stdout.write('  Request packets sent:    %10d (avg. %0.2f, pkgs/bulkwalk)\n' % (s(pre+'requests'), pkgs_per_walk_avg))
        avg_oids_per_walk = s(id) and float(s(pre+'answered'))/s(id) or 0
        avg_oids_per_req  = s(pre+'requests') and float(s(pre+'answered'))/s(pre+'requests') or 0
        sys.stdout.write('  OIDs received:           %10d (Avg. %0.2f/req, %0.2f/%s)\n' %
            (s(pre+'answered'), avg_oids_per_req, avg_oids_per_walk, title))
        sys.stdout.write('    Used:                  %10d\n' % s(pre+'used'))
        perc_overflow = s(pre+'answered') and float(s(pre+'dropped'))/s(pre+'answered')*100 or 0
        sys.stdout.write('    Dropped (overflow):    %10d (%0.2f%% of OIDs)\n' % (s(pre+'dropped'), perc_overflow))
        sys.stdout.write('\n')

    sys.stdout.write('Gets:\n')
    sys.stdout.write('  Get requests:            %10d\n' % s('get'))
    sys.stdout.write('  GetNext requests:        %10d\n' % s('getnext'))
    sys.stdout.write('\n')
    get_answers = s('getnext-wrong-answer') + s('getnext-null-answer') + s('getnext-answered')
    sys.stdout.write('  Answers:                 %10d\n' % get_answers)
    perc_overflow = get_answers and float(s('getnext-wrong-answer'))/get_answers*100 or 0
    sys.stdout.write('    Dropped (overflow):    %10d (%0.2f%% of answers)\n' %
                                        (s('getnext-wrong-answer'), perc_overflow))
    perc_null = get_answers and float(s('getnext-null-answer'))/get_answers*100 or 0
    sys.stdout.write('    Dropped (NULL):        %10d (%0.2f%% of answers)\n' % (s('getnext-null-answer'), perc_null))
    sys.stdout.write('    Used answer:           %10d\n' % s('getnext-answered'))
    sys.stdout.write('\n')

def init_snmp_stats():
    g_snmp_stats['start'] = time.time()

def count_snmp_stats(key):
    if not record_inline_snmp_stats:
        return
    g_snmp_stats.setdefault(key, 0)
    g_snmp_stats[key] += 1

def save_snmp_stats():
    global g_snmp_stats
    import fcntl, pprint, traceback
    path = var_dir + '/snmp.stats'

    file(path, 'a+') # create if not existant

    fd = os.open(path, os.O_RDWR)
    fcntl.fcntl(fd, fcntl.F_SETFD, fcntl.FD_CLOEXEC)
    fcntl.flock(fd, fcntl.LOCK_EX)

    try:
        content = os.read(fd, os.fstat(fd).st_size)
        if content: # merge old with new stats
            try:
                old_stats = eval(content)
            except SyntaxError:
                old_stats = []
            for k, v in old_stats:
                if k in [ 'start', 'end' ]:
                    g_snmp_stats[k] = v
                else:
                    g_snmp_stats.setdefault(k, 0)
                    g_snmp_stats[k] += v

        g_snmp_stats['end'] = time.time()

        os.lseek(fd, 0, os.SEEK_SET)
        os.ftruncate(fd, 0)
        os.write(fd, pprint.pformat(sorted(g_snmp_stats.items())) + '\n')
    except:
        sys.stderr.write('Problem while saving snmp stats: %s\n' % traceback.format_exc())

    fcntl.flock(fd, fcntl.LOCK_UN)
    os.close(fd)

    g_snmp_stats = {} # reset the statistics

def init_snmp_host(hostname, context_name=None):
    try:
        return g_snmp_sessions[(hostname, context_name)]
    except KeyError:
        pass

    if record_inline_snmp_stats:
        init_snmp_stats()

    credentials = snmp_credentials_of(hostname)
    version = 1
    if type(credentials) in (str, unicode):
        if is_bulkwalk_host(hostname) or is_snmpv2c_host(hostname):
            version = 2
    else:
        version = 3
        if len(credentials) == 6:
            sec_level, auth_proto, sec_name, auth_pass, priv_proto, priv_pass = credentials
        elif len(credentials) == 4:
            sec_level, auth_proto, sec_name, auth_pass = credentials
            priv_proto = 'DEFAULT'
            priv_pass  = ''
        elif len(credentials) == 2:
            sec_level, sec_name = credentials
            auth_proto = 'DEFAULT'
            auth_pass  = ''
            priv_proto = 'DEFAULT'
            priv_pass  = ''
        else:
            raise MKGeneralException("Invalid SNMP credentials '%r' for host %s: "
                              "must be string, 2-tuple, 4-tuple or 6-tuple" % (credentials, hostname))

    try:
        ipaddress = lookup_ip_address(hostname)
        if ipaddress == None:
            raise MKGeneralException("Failed to resolve \"%s\" into IP address" % (hostname))
    except Exception, e:
        raise MKGeneralException("Failed to resolve \"%s\" into IP address (%s)" % (hostname, e))

    timing  = snmp_timing_of(hostname)
    timeout = int(timing.get("timeout", 1) * 1000000) # default: 1 second
    retries = timing.get('retries', 3)

    port = snmp_port_of(hostname)
    if port:
        address = "%s:%d" % (ipaddress, port)
    else:
        address = ipaddress

    if is_ipv6_primary(hostname):
        address = "udp6:" + address

    if version != 3:
        s = netsnmp.Session(Version = version, DestHost = address,
            Timeout    = timeout,
            Retries    = retries,
            Community  = credentials,
        )
    else:
        s = netsnmp.Session(Version = version, DestHost = address,
            Timeout   = timeout,
            Retries   = retries,
            SecLevel  = sec_level,
            AuthProto = auth_proto.upper(),
            AuthPass  = auth_pass,
            SecName   = sec_name,
            PrivProto = priv_proto.upper(),
            PrivPass  = priv_pass,
            Context   = context_name or '',
        )

    s.UseNumeric = 1
    # LM: Only way to enable parsing of OPAQUE FLOAT values is to
    # enable this value, this is a special data type, which is for example
    # used by values of hr_cpu check. If this datatype is not parsable,
    # some checks might not work or raise random exceptions caused by
    # parsing issues.
    # When there are other parsing issues, deal with it inside our code.
    # Seams easier than patching net-snmp python bindings to make them
    # understand special types like OPAQUE FLOAT
    # Never ending story... if you enable this, values like interface states
    # or interface types are not handed over to Check_MK via their numeric
    # values. Instead they are translated via it's MIB. This breaks all checks.
    # Again: Don't enable this.
    s.UseSprintValue = 0

    g_snmp_sessions[(hostname, context_name)] = s
    return s

def inline_snmp_get_oid(hostname, oid, get_varbind = False, context_name = None):
    s = init_snmp_host(hostname, context_name)

    if oid[-2:] == ".*":
        count_snmp_stats('getnext')
        oid_prefix = oid[:-2]
        func       = s.getnext
        what       = 'GETNEXT'
    else:
        count_snmp_stats('get')
        oid_prefix = oid
        func       = s.get
        what       = 'GET'

    vverbose("Executing SNMP %s of %s on %s\n" % (what, oid_prefix, hostname))

    var_list = netsnmp.VarList(netsnmp.Varbind(oid_prefix))

    deactivate_check_mk_timeout()
    res = func(var_list)
    reactivate_check_mk_timeout()

    for var in var_list:
        value = var.val

        if what == "GETNEXT" and not var.tag.startswith(oid_prefix):
            count_snmp_stats('getnext-wrong-answer')
            # In case of .*, check if prefix is the one we are looking for
            value = None

        elif what == "GET" and (not var.tag.startswith(oid_prefix)
                                and ("%s.%s" % (var.tag, var.iid)) != oid_prefix):
            value = None

        elif value == 'NULL' or var.type in [ 'NULL', 'ENDOFMIBVIEW', 'NOSUCHINSTANCE', 'NOSUCHOBJECT' ]:
            # Equalize SNMPv1 and SNMPv2 behaviour: SNMPv1 answers with NULL type value and s.ErrorNum set,
            # SNMPv2 answers with a NOSUCHOBJECT value as result.
            # This should better be fixed within net-snmp, but we workaround it here.
            if s.Version == 1 and var.type == 'NULL' and s.ErrorNum == 2:
                s.ErrorNum = 0
                s.ErrorInd = 0
                s.ErrorStr = ""
                var.type = 'NOSUCHOBJECT'

            count_snmp_stats('getnext-null-answer')
            value = None

        elif get_varbind:
            count_snmp_stats('getnext-answered')
            value = var

        elif value is not None:
            count_snmp_stats('getnext-answered')

        vverbose("=> [%r] %s\n" % (value, var.type))

        return value


def inline_bulkwalk(s, start_oid, start_index=0, limit=None):
    # Start from index 0, and use getbulk operations
    cur_index = start_index
    cur_oid = start_oid
    results = []
    count_snmp_stats('bulkwalks')

    while cur_oid == start_oid or cur_oid.startswith(start_oid):
        if cur_oid == start_oid and cur_index == 0:
            var_list = netsnmp.VarList(netsnmp.Varbind(cur_oid))
        else:
            var_list = netsnmp.VarList(netsnmp.Varbind(cur_oid, cur_index))

        count_snmp_stats('bulk-requests')

        # Always ask for 10 OID bulks, but when less OIDs are needed, change
        # the bulk size to less OIDs
        bulk_size = 10
        if limit != None:
            num_needed = limit - len(results)
            if num_needed < 10:
                bulk_size = num_needed

        deactivate_check_mk_timeout()
        result = s.getbulk(0, bulk_size, var_list)
        reactivate_check_mk_timeout()

        # Put all the OIDs returned by this bulkwalk to the list of answers
        # (when matching the request and valid answer)
        finished = False
        for i in var_list:
            count_snmp_stats('bulk-answered')
            if i.type == None:
                break
            elif i.tag == start_oid or (i.tag != None and i.tag.startswith(start_oid + ".")):
                # exact match or at least a subtree of the requested one. Be careful
                # with oids like e.g. 1.2.3.1 and 1.2.3.12
                results.append(i)
                count_snmp_stats('bulk-used')
                if limit != None and len(results) >= limit:
                    finished = True
                    break
            else:
                count_snmp_stats('bulk-dropped')

        if finished:
            break

        # Proceed with another bulkget request or finished?

        # If cur_index is null
        if not var_list[-1].iid:
            break

        # Refresh cur_oid name and increment cur_index
        old_tree, old_index = cur_oid, cur_index
        cur_oid = var_list[-1].tag
        cur_index = int(var_list[-1].iid)

        # If fetch infos did not change, terminate
        if old_tree == cur_oid and old_index == cur_index:
            break

    return results


def inline_bulkwalk_with_limits(s, oid, check_type, table_base_oid, range_limits):
    num_from_start, mid_specs, num_from_end = sanitize_ranges(range_limits)

    if table_base_oid == None or (num_from_start == None and num_from_end == None and mid_specs == []):
        # When no limit configured at all, simply fetch the whole table
        return inline_bulkwalk(s, oid)

    elif num_from_start != None and num_from_end == None and mid_specs == []:
        # When only limit from start is configured, simply fetch from start while using a limit
        return inline_bulkwalk(s, oid, limit=num_from_start)

    elif not has_table_indexes_cached(check_type, table_base_oid):
        # When having a limit from mid or end, but not knowing the indexes yet, then fetch the
        # first column of the table completely, cache the indexes and limit the rows to be
        # fetched on the other columns of that table
        results = inline_bulkwalk(s, oid)

        save_table_index_cache(check_type, table_base_oid, [ int(r.iid) for r in results ])

        # The bulkwalk had to gather the table indexes, but the expected result of this
        # inline_bulkwalk_with_limits() calls is just a subset. Calculate the correct result.

        needed_ranges = get_needed_ranges(check_type, table_base_oid, num_from_start,
                                          num_from_end, mid_specs)

        # We already have some entries fetched. use them, maybe we already got
        # all the needed entries
        cleaned_results = []
        needed_ranges_dict = dict(needed_ranges)
        for num, entry in enumerate(results):
            if int(entry.iid) in needed_ranges_dict:
                num_needed = needed_ranges_dict[int(entry.iid)]
                if len(results) >= num + num_needed:
                    cleaned_results += results[num:num+num_needed]
                    needed_ranges.pop(0)
                    if not needed_ranges:
                        break # Stop when all needed ranges have been gathered

        if needed_ranges: # FIXME: Can be removed when this code has been tested
            #print [ int(r.iid) for r in cleaned_results ], needed_ranges
            raise Exception("Failed to correctly calculate SNMP result")

        return cleaned_results

    else:
        # Having mid or end limits and knowing the table indexes -> feth only the needed ranges
        needed_ranges = get_needed_ranges(check_type, table_base_oid, num_from_start,
                                          num_from_end, mid_specs)
        return inline_bulkwalk_for_needed_ranges(s, oid, needed_ranges,
                                                 g_snmp_table_cache[(check_type, table_base_oid)])


def sanitize_ranges(range_limits):
    start_limits, end_limits, mid_specs = [], [], []
    for what, spec in range_limits:
        if what == "first":
            start_limits.append(spec)
        elif what == "last":
            end_limits.append(spec)
        elif what == "mid":
            mid_specs.append(spec)

    num_from_start = start_limits and max(start_limits) or None
    num_from_end = end_limits and max(end_limits) or None

    return num_from_start, mid_specs, num_from_end


def get_needed_ranges(check_type, table_base_oid, num_from_start, num_from_end, mid_specs):
    indexes = g_snmp_table_cache[(check_type, table_base_oid)]

    # Convert start/end definitions to range definitions
    ranges = []
    if num_from_start:
        ranges.append((1, num_from_start))
    ranges += mid_specs
    if num_from_end:
        ranges.append((len(indexes)-num_from_end+1, num_from_end))

    # now convert from table row number to index number
    indexed_ranges = []
    for n, m in ranges:
        try:
            indexed_ranges.append((indexes[n-1], m))
        except IndexError:
            pass # silently ignore ranges exceeding the existing indexes

    # FIXME: Whan if ranges overlap?

    return sorted(indexed_ranges)


def has_table_indexes_cached(check_type, table_base_oid):
    return (check_type, table_base_oid) in g_snmp_table_cache


def save_table_index_cache(check_type, table_base_oid, indexes):
    g_snmp_table_cache[(check_type, table_base_oid)] = indexes


def inline_bulkwalk_for_needed_ranges(s, oid, needed_ranges, table_indexes):
    results = []
    for start_index, limit in needed_ranges:
        # bulkwalks don't fetch the OID which is used to start the bulkwalk
        # from. But our logic needs exactly this OID. Lookup the previous
        # index from the table index cache which is always available when
        # start_index is used
        if start_index != 0:
            nth_index = table_indexes.index(start_index)
            if nth_index == 0:
                start_index = 0
            else:
                start_index = table_indexes[nth_index-1]
        results += inline_bulkwalk(s, oid, start_index, limit)
    return results


def inline_walk(s, oid):
    var_list = netsnmp.VarList(netsnmp.Varbind(oid))
    count_snmp_stats('walks')
    count_snmp_stats('walk-requests')

    deactivate_check_mk_timeout()
    res = s.walk(var_list)
    reactivate_check_mk_timeout()

    results = []
    for var in var_list:
        count_snmp_stats('walk-answered')
        if var.iid is not None:
            count_snmp_stats('walk-used')
            results.append(var)
        else:
            count_snmp_stats('walk-dropped')

    # If a GETNEXT request is given the very last valid instance (or an OID later
    # than this), then the agent needs to indicate that there are no further values
    # to retrieve. With SNMPv1, this uses the same error as a GET request for a
    # non-existant OID: (noSuchName) There is no such variable name in this MIB.
    # This should better be handled by the netsnmp module transparently. But for the
    # moment we try to workaround this issue by resetting this issue.
    if s.Version == 1 and s.ErrorNum == 2:
        s.ErrorNum = 0
        s.ErrorInd = 0
        s.ErrorStr = ""

    return results

def inline_snmpwalk_on_suboid(hostname, check_type, oid, table_base_oid=None, context_name=None):
    s = init_snmp_host(hostname, context_name)

    bulkwalk = is_snmpv3_host(hostname) or is_bulkwalk_host(hostname)
    if bulkwalk:
        what = 'BULKWALK'
    else:
        what = 'WALK'

    # Remove trailing .0 for walks
    # .1.3.6.1.2.1.1.5.0 but receive the value for 1.3.6.1.2.1.1.6.0.
    if oid[-2:] == '.0':
        oid = oid[:-2]

    vverbose("Executing %s of \"%s\" on %s\n" % (what, oid, hostname))

    start_time = time.time()

    if bulkwalk:
        if check_type:
            range_limits = oid_range_limits(hostname, check_type)
        else:
            range_limits = []
        res = inline_bulkwalk_with_limits(s, oid, check_type, table_base_oid, range_limits)
    else:
        res = inline_walk(s, oid)

    # Preserve the walks error states (don't let it be overwritten by the eventual
    # following SNMP GET)
    err_num, err_ind, err_str = s.ErrorNum, s.ErrorInd, s.ErrorStr

    # Adapt handling of snmpbulkwalk command line tool which tries a snmpget
    # after a snmpbulkwalk returned no matching result, but also no error.
    if not res and err_num == 0 and err_ind == 0:
        value = inline_snmp_get_oid(hostname, oid, get_varbind=True, context_name=context_name)
        if value == None or value.val in ['NOSUCHINSTANCE', 'ENDOFMIBVIEW', 'NOSUCHOBJECT']:
            res = []
        else:
            res = [ value ]

    if err_num != 0 or err_ind != 0:
        verbose(tty_red + tty_bold + "ERROR: " + tty_normal + "SNMP error %d/%d (%s)\n" %
                                                              (err_num, err_ind, err_str))

        # Do not raise exception in case of some special error.
        # Behave like the classic snmpwalk_on_suboid()
        # ErrorNum:
        # 2: (noSuchName) There is no such variable name in this MIB.
        # ErrorInd:
        # -
        if err_num in [ 2, ]:
            return []
        elif err_ind == -24:
            # -24: Timeout - snmpwalk_on_suboid() in classic SNMP code raises a
            # MKSNMPError when the snmp command did exit with exit code != 0. Adapt
            # this behaviour here.
            timing  = snmp_timing_of(hostname)
            tries = timing.get('retries', 3) + 1 # add first try to retries
            raise MKSNMPError('SNMP Error on %s: SNMP query timed out after %d tries within %0.2f seconds' %
                                         (hostname, tries, time.time() - start_time))
        elif err_ind == -4:
            # SNMP error 0/-4 (Unknown session ([host])). Seems to occur e.g. in case of name
            # resolution issues
            raise MKSNMPError("Failed to initiate SNMP (Maybe name resolution issue)")
        else:
            raise MKSNMPError('SNMP Error on %s while walking %s: %s (%d/%d).' %
                (hostname, oid, err_str, err_num, err_ind))

    results = []
    for var in res:
        if var.iid is None:
            continue

        this_oid = var.tag + '.' + var.iid
        vverbose("%s => [%s] %r\n" % (this_oid, var.val, var.type))

        # In Check_MK code values of SNMP walk tables are never expected to be None. Change the
        # netsnmp provided value to be a empty string instead to prevent issues with later code.
        value = var.val
        if value == None:
            value = ""

        results.append((this_oid, value))

    return results


def inline_convert_rows_for_stored_walk(rows):
    def should_be_encoded(v):
        for c in v:
            if ord(c) < 32 or ord(c) > 127:
                return True
        return False

    def hex_encode_value(v):
        encoded = ""
        for c in v:
            encoded += "%02X " % ord(c)
        return "\"%s\"" % encoded

    new_rows = []
    for index, (oid, value) in enumerate(rows):
        if value == "ENDOFMIBVIEW":
            continue

        if should_be_encoded(value):
            new_rows.append((oid, hex_encode_value(value)))
        else:
            new_rows.append((oid, value))
    return new_rows
