#!/usr/bin/python
# -*- encoding: utf-8; py-indent-offset: 4 -*-
# .------------------------------------------------------------------------.
# |                ____ _               _        __  __ _  __              |
# |               / ___| |__   ___  ___| | __   |  \/  | |/ /              |
# |              | |   | '_ \ / _ \/ __| |/ /   | |\/| | ' /               |
# |              | |___| | | |  __/ (__|   <    | |  | | . \               |
# |               \____|_| |_|\___|\___|_|\_\___|_|  |_|_|\_\              |
# |                                        |_____|                         |
# |             _____       _                       _                      |
# |            | ____|_ __ | |_ ___ _ __ _ __  _ __(_)___  ___             |
# |            |  _| | '_ \| __/ _ \ '__| '_ \| '__| / __|/ _ \            |
# |            | |___| | | | ||  __/ |  | |_) | |  | \__ \  __/            |
# |            |_____|_| |_|\__\___|_|  | .__/|_|  |_|___/\___|            |
# |                                     |_|                                |
# |                     _____    _ _ _   _                                 |
# |                    | ____|__| (_) |_(_) ___  _ __                      |
# |                    |  _| / _` | | __| |/ _ \| '_ \                     |
# |                    | |__| (_| | | |_| | (_) | | | |                    |
# |                    |_____\__,_|_|\__|_|\___/|_| |_|                    |
# |                                                                        |
# | mathias-kettner.com                                 mathias-kettner.de |
# '------------------------------------------------------------------------'
#  This file is part of the Check_MK Enterprise Edition (CEE).
#  Copyright by Mathias Kettner and Mathias Kettner GmbH.  All rights reserved.
#
#  Distributed under the Check_MK Enterprise License.
#
#  You should have  received  a copy of the Check_MK Enterprise License
#  along with Check_MK. If not, email to mk@mathias-kettner.de
#  or write to the postal address provided at www.mathias-kettner.de

#.
#   .--Global Vars---------------------------------------------------------.
#   |          ____ _       _           _  __     __                       |
#   |         / ___| | ___ | |__   __ _| | \ \   / /_ _ _ __ ___           |
#   |        | |  _| |/ _ \| '_ \ / _` | |  \ \ / / _` | '__/ __|          |
#   |        | |_| | | (_) | |_) | (_| | |   \ V / (_| | |  \__ \          |
#   |         \____|_|\___/|_.__/ \__,_|_|    \_/ \__,_|_|  |___/          |
#   |                                                                      |
#   +----------------------------------------------------------------------+
#   | Global variables which are specific to the keepalive mode            |
#   '----------------------------------------------------------------------'

g_total_check_output = ""
g_keepalive_fd       = None
g_num_checks         = 0      # count total number of check cycles
g_keepalive_timeout  = 60
g_initial_times      = None   # saving execution time, user time etc.
g_ipaddress_cache    = {}     # helper local ipaddress cache
g_orig_global_vars   = {}     # saves global variables in verbose mode for
                              # checking for changed variables during runtime

ignore_changed_global_variables += [
    "g_num_checks",
    "g_keepalive_timeout",
    "g_orig_global_vars",
]

#.
#   .--Keepalive-----------------------------------------------------------.
#   |               _  __                     _ _                          |
#   |              | |/ /___  ___ _ __   __ _| (_)_   _____                |
#   |              | ' // _ \/ _ \ '_ \ / _` | | \ \ / / _ \               |
#   |              | . \  __/  __/ |_) | (_| | | |\ V /  __/               |
#   |              |_|\_\___|\___| .__/ \__,_|_|_| \_/ \___|               |
#   |                            |_|                                       |
#   +----------------------------------------------------------------------+
#   | Generic keepalive handling code. Not specific to a keepalive mode.   |
#   '----------------------------------------------------------------------'

# The command_function is handling the command sent by the core. It gets the cmdline
# string sent by the core and needs to parse it on its own. The function must
# return a nagios plugin exit code for telling the core the result state.
def do_keepalive(command_function, parse_cmdline_function):
    ignore_timeout_signal()
    prepare_keepalive_fds()
    register_keepalive_sigint_handler()
    read_packed_config()
    save_global_vars()

    while True:
        cleanup_globals()
        cmdline = read_keepalive_line()
        set_initial_times()

        cmdline = cmdline.strip()
        if cmdline == "*":
            read_packed_config()
            cleanup_globals()
            reset_global_caches()
            save_global_vars()
            continue

        elif not cmdline:
            break

        hostname = None
        try:
            try:
                command_tuple = parse_cmdline_function(cmdline)
                hostname = command_tuple[0]
            except MKSkipCheck:
                # The parse cmdline function skips this line and tries to recover the protocol
                # with the next received line.
                continue

            except Exception, e:
                # The helper <> CMC protocol may be out of sync. Terminate the helper!
                sys.stdout.write("ERROR: Failed to process helper command: \"%s\": %s\n" %
                                                                                (cmdline, e))
                sys.stdout.write("The helper <> CMC protocol may be out of sync. Terminating!\n")
                sys.stdout.flush()

                write_result(3, "UNKNOWN - Failed to process command \"%s\": %s, "
                                "see cmc.log for details\n" % (cmdline, e))
                break

            clear_keepalive_result()
            count_check_cycle()

            try:
                activate_keepalive_timeout(int(read_keepalive_line()))

                status = command_function(command_tuple)

                ignore_timeout_signal()
                reset_timeout_counter()

                write_result(status, g_total_check_output)

            except MKTimeout:
                ignore_timeout_signal()
                reset_timeout_counter()
                status = exit_code_spec(hostname).get("timeout", 2)
                write_active_check_result(hostname, status,
                    "%s - Check_MK timed out after %d seconds\n" %
                         (core_state_names[status], g_keepalive_timeout))

            clear_keepalive_result()

        except Exception, e:
            ignore_timeout_signal()
            reset_timeout_counter()
            if opt_debug:
                raise
            else:
                if isinstance(e,  MKGeneralException):
                    sys.stdout.write("%s\n" % e)
                else:
                    import traceback # Always log details to the cmc.log
                    sys.stdout.write("An exception occured while "
                                     "processing host \"%s\"\n" % hostname)
                    sys.stdout.write(traceback.format_exc())
            write_active_check_result(hostname, 3, "UNKNOWN - %s\n" % e)

        # Flush file descriptors of stdout and stderr, so that diagnostic
        # messages arrive in time in cmc.log
        sys.stdout.flush()
        sys.stderr.flush()

        cleanup_globals() # Prepare for next check
        restore_use_cachefile()
        restore_original_agent_caching_usage()

        # Check if all global variables are clean, but only in verbose logging mode
        if opt_verbose:
            keepalive_check_changed_global_vars()

        keepalive_check_memory()

        # In case of profiling do just this one cycle and end afterwards
        if g_profile:
            output_profile()
            sys.exit(0)


def save_global_vars():
    global g_orig_global_vars
    if opt_verbose: # only checked in verbose mode
        g_orig_global_vars = copy_globals()


# FIXME: Is this still needed? Is the keepalive specific IP address cache needed?
# This is never resetting when an address is not resolvable anymore.
def resolve_ip_address_for_keepalive(hostname):
    if hostname in g_ipaddress_cache:
        return g_ipaddress_cache[hostname]

    if is_cluster(hostname):
        ipaddress = None
    else:
        try:
            ipaddress = lookup_ip_address(hostname)
        except:
            raise MKGeneralException("Cannot resolve hostname %s into IP address" % hostname)
    g_ipaddress_cache[hostname] = ipaddress
    return ipaddress


def set_timeout_counter():
    signal.alarm(g_keepalive_timeout)


def reset_timeout_counter():
    signal.alarm(0)


def ignore_timeout_signal():
    signal.signal(signal.SIGALRM, signal.SIG_IGN) # Prevent ALRM from CheckHelper.cc


def activate_keepalive_timeout(timeout):
    global g_keepalive_timeout
    g_keepalive_timeout = timeout # save for later error reporting

    signal.signal(signal.SIGALRM, handle_keepalive_timeout)
    signal.alarm(timeout)


def handle_keepalive_timeout(signum, frame):
    raise MKTimeout()


def get_keepalive_times():
    return map(lambda a: a[0]-a[1], zip(os.times(), g_initial_times))


def set_initial_times():
    global g_initial_times
    g_initial_times = os.times()


def count_check_cycle():
    global g_num_checks
    g_num_checks += 1


def prepare_keepalive_fds():
    global g_keepalive_fd
    # Prevent against plugins that output debug information (but shouldn't).
    # Their stdout will interfer with communication with the Micro Core.
    # So we simply redirect stdout to stderr, which will appear in the cmc.log,
    # with the following trick:
    # 1. move the filedescriptor 1 to a parking position
    # 2. dup the stderr channel to stdout (2 to 1)
    # 3. Send our answers to the Micro Core with the parked FD.
    # BEWARE: this must not happen after we have execve'd ourselves!
    if opt_keepalive_fd:
        g_keepalive_fd = opt_keepalive_fd
    else:
        g_keepalive_fd = os.dup(1)
        os.dup2(2, 1)  # Send stuff that is written to stdout instead to stderr


def write_result(status, results):
    results_utf8 = make_utf8(results)
    output = "%03d\n%08d\n%s" % (status, len(results_utf8), results_utf8)

    while output:
        bytes_written = os.write(g_keepalive_fd, output)
        output = output[bytes_written:]


def write_active_check_result(hostname, status, result):
    write_result(status, format_active_check_result(hostname, result))


def keepalive_check_changed_global_vars():
    new_global_vars = copy_globals()
    for varname, value in g_orig_global_vars.items():
        if varname not in ignore_changed_global_variables and value != new_global_vars[varname]:
            if opt_verbose > 1:
                sys.stderr.write("WARNING: global variable %s has changed: %r ==> %r\n"
                        % (varname, value, new_global_vars[varname]))
            else:
                sys.stderr.write("WARNING: global variable %s has changed: %s ==> %s\n"
                        % (varname, repr(value)[:50], repr(new_global_vars[varname])[:50]))
    new_vars = set(new_global_vars.keys()).difference(set(g_orig_global_vars.keys()))
    if new_vars:
        sys.stderr.write("WARNING: new variable appeared: %s\n" % ", ".join(new_vars))
    sys.stderr.flush()


# Diagnostic function for detecting global variables that have
# changed during checking. This is slow and canno be used
# in production mode.
def copy_globals():
    import copy
    global_saved = {}
    for varname, value in globals().items():
        # Some global caches are allowed to change.
        if varname not in [ "g_service_description", "g_checked_item", "g_check_type",
                            "g_multihost_checks", "g_singlehost_checks",
                            "g_nodesof_cache", "g_compiled_regexes", "vars_before_config",
                            "g_initial_times", "g_keepalive_initial_memusage",
                            "g_global_caches" ] + g_global_caches \
            and type(value).__name__ not in [ "function", "module", "SRE_Pattern" ]:
            global_saved[varname] = copy.copy(value)
    return global_saved


# Determine currently (VmSize, VmRSS) in Bytes
def current_memory_usage():
    parts = file('/proc/self/stat').read().split()
    vsize = int(parts[22])        # in Bytes
    rss   = int(parts[23]) * 4096 # in Pages
    return (vsize, rss)


keepalive_memcheck_cycle = 20
g_keepalive_initial_memusage = None
def keepalive_check_memory():
    if g_num_checks % keepalive_memcheck_cycle != 0: # Only do this after every 10 checks
        return

    global g_keepalive_initial_memusage
    if not g_keepalive_initial_memusage:
        g_keepalive_initial_memusage = current_memory_usage()
    else:
        usage = current_memory_usage()
        # Allow VM size to grow by at most 50%
        if usage[0] > 1.5 * g_keepalive_initial_memusage[0]:
            sys.stderr.write("memory usage increased from %s to %s after %d check cycles. Restarting.\n" % (
                    get_bytes_human_readable(g_keepalive_initial_memusage[0]),
                    get_bytes_human_readable(usage[0]), g_num_checks))
            restart_myself()


def restart_myself():
    sys.argv = [ x for x in sys.argv if not x.startswith('--keepalive-fd=') ]
    os.execvp("cmk", sys.argv + [ "--keepalive-fd=%d" % g_keepalive_fd ])


def add_keepalive_result_line(result):
    global g_total_check_output
    g_total_check_output += result


def add_keepalive_check_result(hostname, service, state, output, cached_at=None, cache_interval=None):
    # Replace \n to enable multiline ouput. remove binary 0, CMC does not like it.
    output = output.replace("\n", "\x01", 1).replace("\n","\\n").replace("\0", "")
    result = "\t%d\t%s\t%s\t%s\n" % (state, hostname, service, output)
    if cached_at:
        result = "\tcached_at=%d\tcache_interval=%d%s" % (cached_at, cache_interval, result)
    add_keepalive_result_line(result)


def add_keepalive_active_check_result(hostname, result):
    add_keepalive_result_line(format_active_check_result(hostname, result))


def format_active_check_result(hostname, result):
    return "%s\t%s" % (hostname, result)


def clear_keepalive_result():
    global g_total_check_output
    g_total_check_output = ""


# Just one lines from stdin. But: make sure that
# nothing more is read - not even into some internal
# buffer of sys.stdin! We do this by reading every
# single byte. I know that this is not performant,
# but we just read hostnames - not much data.
def read_keepalive_line():
    line = ""
    while True:
        byte = os.read(0, 1)
        if byte == '\n':
            return line
        elif not byte: # EOF
            return ''
        else:
            line += byte


#.
#   .--Checking------------------------------------------------------------.
#   |               ____ _               _    _                            |
#   |              / ___| |__   ___  ___| | _(_)_ __   __ _                |
#   |             | |   | '_ \ / _ \/ __| |/ / | '_ \ / _` |               |
#   |             | |___| | | |  __/ (__|   <| | | | | (_| |               |
#   |              \____|_| |_|\___|\___|_|\_\_|_| |_|\__, |               |
#   |                                                 |___/                |
#   +----------------------------------------------------------------------+
#   | Keepalive service check handling code                                |
#   '----------------------------------------------------------------------'

def do_check_keepalive():
    do_keepalive(command_function=execute_keepalive_command,
                 parse_cmdline_function=parse_keepalive_check_command)


def parse_keepalive_check_command(cmdline):
    args = cmdline.split()

    if '--cache' in args:
        args.remove('--cache')
        cache = True
    else:
        cache = False

    # FIXME: remove obsolete check-inventory
    if '--check-inventory' in args:
        args.remove('--check-inventory')
        mode_function = check_discovery
    elif '--check-discovery' in args:
        args.remove('--check-discovery')
        mode_function = check_discovery
    else:
        mode_function = do_check

    num_args = len(args)
    if num_args == 2:
        hostname, ipaddress = args[:2]

    elif num_args == 1:
        hostname = args[0]
        ipaddress = resolve_ip_address_for_keepalive(hostname)

    else:
        raise MKGeneralException("Received invalid command line")

    return hostname, ipaddress, cache, mode_function


# The CMC always provides arguments. The last two arguments are the hostname
# and the ipaddress of the host to be asked for. The other arguments might
# be different parameters to configure the actions to be done
def execute_keepalive_command((hostname, ipaddress, cache, mode_function)):
    if cache:
        set_use_cachefile()
        enforce_using_agent_cache()

    return mode_function(hostname, ipaddress)
