#!/usr/bin/python
# -*- encoding: utf-8; py-indent-offset: 4 -*-
# .------------------------------------------------------------------------.
# |                ____ _               _        __  __ _  __              |
# |               / ___| |__   ___  ___| | __   |  \/  | |/ /              |
# |              | |   | '_ \ / _ \/ __| |/ /   | |\/| | ' /               |
# |              | |___| | | |  __/ (__|   <    | |  | | . \               |
# |               \____|_| |_|\___|\___|_|\_\___|_|  |_|_|\_\              |
# |                                        |_____|                         |
# |             _____       _                       _                      |
# |            | ____|_ __ | |_ ___ _ __ _ __  _ __(_)___  ___             |
# |            |  _| | '_ \| __/ _ \ '__| '_ \| '__| / __|/ _ \            |
# |            | |___| | | | ||  __/ |  | |_) | |  | \__ \  __/            |
# |            |_____|_| |_|\__\___|_|  | .__/|_|  |_|___/\___|            |
# |                                     |_|                                |
# |                     _____    _ _ _   _                                 |
# |                    | ____|__| (_) |_(_) ___  _ __                      |
# |                    |  _| / _` | | __| |/ _ \| '_ \                     |
# |                    | |__| (_| | | |_| | (_) | | | |                    |
# |                    |_____\__,_|_|\__|_|\___/|_| |_|                    |
# |                                                                        |
# | mathias-kettner.com                                 mathias-kettner.de |
# '------------------------------------------------------------------------'
#  This file is part of the Check_MK Enterprise Edition (CEE).
#  Copyright by Mathias Kettner and Mathias Kettner GmbH.  All rights reserved.
#
#  Distributed under the Check_MK Enterprise License.
#
#  You should have  received  a copy of the Check_MK Enterprise License
#  along with Check_MK. If not, email to mk@mathias-kettner.de
#  or write to the postal address provided at www.mathias-kettner.de

try:
    from Crypto.Cipher import AES
    from hashlib import md5
    rtc_import_error = None
except ImportError, e:
    rtc_import_error = "%s" % e


def do_real_time_checks(args):
    do_keepalive(command_function=execute_keepalive_rtc,
                 parse_cmdline_function=parse_rtc_command)

# CMC sends these commands:
#
# [pkg size] [hostname] [ipaddress]\n
# [timeout]\n
# [pkg]
#
# The first line is available in cmdline (argument of this function)
# The timeout is read and handled by the generic do_keepalive() code
# The pkg must be read by this function
def execute_keepalive_rtc((hostname, ipaddress, pkg_size)):
    if rtc_import_error:
        raise MKGeneralException("Skipping real-time check result "
                                 "(Missing dependency: %s)" % rtc_import_error)

    pkg = fetch_rtc_package(pkg_size)
    # Each package must start with an agent section header. If it is not,
    # the decryption might have failed. Skip the package in this case.
    if pkg[0] != "<":
        raise MKGeneralException("Skipping real-time check result of %s "
                                 "(Invalid data, secret might be wrong)" % hostname)


    # Save raw package for eventual later crash reporting
    save_rtc_package(pkg)

    # Here the whole command has been read

    info = parse_info(pkg.split("\n"), hostname)[0]

    check_types = all_check_types_for_rtc_pkg(info)

    # Storing the info as cached hostinfo makes it possible to hook
    # data into get_realhost_info(). This is the lowlevel function to
    # get data from the agent. Storing the cached hostinfo prevents
    # this function from contacting the agent.
    store_cached_hostinfo(hostname, info)

    exit_spec = exit_code_spec(hostname)

    load_item_state(hostname)
    # Totally ignore all data returned by do_all_checks_on_host.
    # These vars are meant for the active checks "Check_MK" etc.
    # but not relevant for the real time check processing
    agent_version, num_success, error_sections, problems = \
        do_all_checks_on_host(hostname, ipaddress, only_check_types=check_types,
                              fetch_agent_version=False)
    save_item_state(hostname)
    delete_rtc_package()

    return 0


def all_check_types_for_rtc_pkg(info):
    check_types = []
    for section_name in info.keys():
        check_types += all_check_types_for_agent_section(section_name)
    return check_types


def all_check_types_for_agent_section(section_name):
    check_types = []
    for check_type in check_info.keys():
        basename = check_type.split(".", 1)[0]
        if basename == section_name:
            check_types.append(check_type)
    return check_types


def save_rtc_package(pkg):
    global g_last_real_time_check_package
    g_last_real_time_check_package = pkg


def delete_rtc_package():
    save_rtc_package(None)


def get_rtc_package():
    return g_last_real_time_check_package


def fetch_rtc_package(pkg_size):
    protocol_version, timestamp, encrypted_pkg = read_rtc_package(pkg_size)

    if protocol_version != 0:
        raise MKGeneralException("Skipping real time check result (Unsupported protocol version)")

    if timestamp < time.time() - 5:
        raise MKGeneralException("Skipping real time check result (Older than 5 seconds)")

    return decrypt_rtc_package(encrypted_pkg)


def decrypt_rtc_package(encrypted_pkg):
    # Adapt OpenSSL handling of key and iv
    def derive_key_and_iv(password, key_length, iv_length):
        d = d_i = ''
        while len(d) < key_length + iv_length:
            d_i = md5(d_i + password).digest()
            d += d_i
        return d[:key_length], d[key_length:key_length+iv_length]

    key, iv = derive_key_and_iv(get_rtc_secret(), 32, AES.block_size)

    decryption_suite = AES.new(key, AES.MODE_CBC, iv)
    decrypted_pkg = decryption_suite.decrypt(encrypted_pkg)

    # Strip of fill bytes of openssl
    return decrypted_pkg.strip("\x10")


def get_rtc_secret():
    if cmc_real_time_checks and cmc_real_time_checks.get("secret"):
        return cmc_real_time_checks["secret"]
    else:
        return ""


# FIXME: Error handling, timeout, ...?
def read_rtc_package(pkg_size):
    # unencrypted header, 12 bytes: "001449832241"
    # 2 bytes: protocol version
    # 10 bytes: unix timestamp
    header_size = 12

    header           = os.read(0, header_size)
    protocol_version = int(header[:2])
    timestamp        = int(header[2:])

    encrypted_pkg = os.read(0, pkg_size-header_size)

    return protocol_version, timestamp, encrypted_pkg



# FIXME: The command line should be cleaned up to make this handable in
# generic code. For example put the hostname always at the first/last pos.
def get_hostname_from_rtc_command(cmdline):
    args = cmdline.split()
    if len(args) > 1:
        return args[1]


# This function can raise MKSkipCheck() to silently skip the received line.
# Other exceptions will result in an error message printed to the cmc log
# but the command is also skipped.
def parse_rtc_command(cmdline):
    args = cmdline.split()

    try:
        pkg_size = int(args[0])
    except ValueError:
        raise MKSkipCheck()

    if len(args) == 3:
        hostname, ipaddress = args[1:]
    elif len(args) == 2:
        hostname = args[1]
        ipaddress = resolve_ip_address_for_keepalive(hostname)
    else:
        raise MKSkipCheck()

    return hostname, ipaddress, pkg_size
