#!/usr/bin/python
# -*- encoding: utf-8; py-indent-offset: 4 -*-
# .------------------------------------------------------------------------.
# |                ____ _               _        __  __ _  __              |
# |               / ___| |__   ___  ___| | __   |  \/  | |/ /              |
# |              | |   | '_ \ / _ \/ __| |/ /   | |\/| | ' /               |
# |              | |___| | | |  __/ (__|   <    | |  | | . \               |
# |               \____|_| |_|\___|\___|_|\_\___|_|  |_|_|\_\              |
# |                                        |_____|                         |
# |             _____       _                       _                      |
# |            | ____|_ __ | |_ ___ _ __ _ __  _ __(_)___  ___             |
# |            |  _| | '_ \| __/ _ \ '__| '_ \| '__| / __|/ _ \            |
# |            | |___| | | | ||  __/ |  | |_) | |  | \__ \  __/            |
# |            |_____|_| |_|\__\___|_|  | .__/|_|  |_|___/\___|            |
# |                                     |_|                                |
# |                     _____    _ _ _   _                                 |
# |                    | ____|__| (_) |_(_) ___  _ __                      |
# |                    |  _| / _` | | __| |/ _ \| '_ \                     |
# |                    | |__| (_| | | |_| | (_) | | | |                    |
# |                    |_____\__,_|_|\__|_|\___/|_| |_|                    |
# |                                                                        |
# | mathias-kettner.com                                 mathias-kettner.de |
# '------------------------------------------------------------------------'
#  This file is part of the Check_MK Enterprise Edition (CEE).
#  Copyright by Mathias Kettner and Mathias Kettner GmbH.  All rights reserved.
#
#  Distributed under the Check_MK Enterprise License.
#
#  You should have  received  a copy of the Check_MK Enterprise License
#  along with Check_MK. If not, email to mk@mathias-kettner.de
#  or write to the postal address provided at www.mathias-kettner.de


import defaults, config
import re, os
from lib import *

agent_dir = defaults.var_dir + "/agents/"


agent_package_types = {
    "linux_rpm"   : ("application/x-rpm", "-%(version)s-%(release)s.noarch.rpm"),
    "linux_deb"   : ("application/x-deb", "_%(version)s-%(release)s_all.deb"),
    "linux_tgz"   : ("application/x-tgz", "_%(version)s-%(release)s.tar.gz"),
    "aix_tgz"     : ("application/x-tgz", "_%(version)s-%(release)s.tar.gz"),
    "windows_msi" : ("application/x-msi", "-%(version)s-%(release)s.msi"),
}

#   .--Helpers-------------------------------------------------------------.
#   |                  _   _      _                                        |
#   |                 | | | | ___| |_ __   ___ _ __ ___                    |
#   |                 | |_| |/ _ \ | '_ \ / _ \ '__/ __|                   |
#   |                 |  _  |  __/ | |_) |  __/ |  \__ \                   |
#   |                 |_| |_|\___|_| .__/ \___|_|  |___/                   |
#   |                              |_|                                     |
#   +----------------------------------------------------------------------+
#   |  Some generic helper functions                                       |
#   '----------------------------------------------------------------------'

def validate_host_name(host_name):
    if host_name and not re.match("^[-a-zA-Z0-9_.]+$", host_name):
        raise MKGeneralException(_("Invalid host name"))


def validate_host_permission(host_name):
    if not config.may("wato.download_all_agents"):
        config.need_permission("wato.download_agents")
        if not host_name:
            raise MKAuthException(_("Sorry, you are not allowed to download agents without specifying a host name."))

        result = html.live.query("GET hosts\nColumns: host_name\nFilter: host_name = %s\n" % host_name)
        if not result:
            raise MKAuthException(_("Sorry, you have no permission for this host."))


#.
#   .--Agents--------------------------------------------------------------.
#   |                      _                    _                          |
#   |                     / \   __ _  ___ _ __ | |_ ___                    |
#   |                    / _ \ / _` |/ _ \ '_ \| __/ __|                   |
#   |                   / ___ \ (_| |  __/ | | | |_\__ \                   |
#   |                  /_/   \_\__, |\___|_| |_|\__|___/                   |
#   |                          |___/                                       |
#   +----------------------------------------------------------------------+
#   |  Helper functions for loading / querying existing agents.            |
#   '----------------------------------------------------------------------'

def agent_path(aghash, opsys):
    return defaults.var_dir + "/agents/" + opsys + "/_PACKAGES/" + aghash


def read_agent(aghash, opsys):
    return file(agent_path(aghash, opsys)).read()


def agent_file_size(aghash, package_type):
    return os.stat(agent_dir + package_type + "/_PACKAGES/" + aghash).st_size


def agent_exists(opsys, aghash):
    return os.path.exists(agent_dir + opsys + "/_PACKAGES/" + aghash)


# Create two-tier dict of all agent configurations
# top level: agent configuration hash
# second level: operating system
# value: list of hostnames
# Example result:
#
# {'1903d12d26a5fcca': {'aix_tgz': ['myserver456'],
#                       'linux_deb': ['myserver456'],
#                       'linux_rpm': ['myserver456'],
#                       'linux_tgz': ['myserver456'],
#                       'windows_msi': ['myserver456']},
#  '42dde486974e3378': {'aix_tgz': ['VANILLA'],
#                       'linux_deb': ['VANILLA'],
#                       'linux_rpm': ['VANILLA'],
#                       'linux_tgz': ['VANILLA'],
#                       'windows_msi': ['VANILLA']},
#  'f2c20b00acfce354': {'aix_tgz': ['myserver123', 'GENERIC'],
#                       'linux_deb': ['myserver123', 'GENERIC'],
#                       'linux_rpm': ['myserver123', 'GENERIC'],
#                       'linux_tgz': ['myserver123', 'GENERIC'],
#                       'windows_msi': ['myserver123', 'GENERIC']}}
def find_agents():
    agconfs = {}
    if os.path.exists(agent_dir):
        for opsys in os.listdir(agent_dir):
            if opsys.startswith("."):
                continue
            for hostname in os.listdir(agent_dir + opsys):
                if hostname.startswith(".") or hostname == "_PACKAGES":
                    continue

                linkpath = agent_dir + opsys + "/" + hostname
                if not os.path.exists(linkpath):
                    # Host had an agent once ago, but now does not have.
                    # This can result in dangling links.
                    os.remove(linkpath)
                    continue
                aghash = os.readlink(linkpath).split("/")[1]
                hostnames = agconfs.setdefault(aghash, {}).setdefault(opsys, [])
                if hostname == "_GENERIC":
                    hostnames.append("GENERIC")
                elif hostname == "_VANILLA":
                    hostnames.append("VANILLA")
                else:
                    hostnames.append(hostname)

    return agconfs


def find_agents_of_host(host_name):
    package_types = []
    host_aghash = None
    agconfs = find_agents()
    for aghash, entry in agconfs.items():
        for package_type, host_names in entry.items():
            if host_name in host_names and agent_exists(package_type, aghash):
                package_types.append(package_type)
                host_aghash = aghash
    return host_aghash, package_types


def download_agent(path, basepath, aghash, opsys):
    if not os.path.exists(path):
        raise MKGeneralException(_("Unknown agent configuration"))

    conf = eval(file(basepath + "_PACKAGES/" + aghash + ".conf").read())
    version = conf["check_mk_version"]
    release = aghash

    mime_type, filename_template = agent_package_types[opsys]
    filename = "check-mk-agent" + filename_template % vars()
    html.req.content_type = mime_type
    html.req.headers_out['Content-Disposition'] = "attachment; filename=\"%s\"" % filename
    content = file(path).read()
    html.write(content)


#.
#   .--Deployment----------------------------------------------------------.
#   |       ____             _                                  _          |
#   |      |  _ \  ___ _ __ | | ___  _   _ _ __ ___   ___ _ __ | |_        |
#   |      | | | |/ _ \ '_ \| |/ _ \| | | | '_ ` _ \ / _ \ '_ \| __|       |
#   |      | |_| |  __/ |_) | | (_) | |_| | | | | | |  __/ | | | |_        |
#   |      |____/ \___| .__/|_|\___/ \__, |_| |_| |_|\___|_| |_|\__|       |
#   |                 |_|            |___/                                 |
#   +----------------------------------------------------------------------+
#   |  Functions dealing wiht automatic agent updates                      |
#   '----------------------------------------------------------------------'

# This page is secured by authentication. We are logged in as
# some administrative user if this code is reached.
def page_register_agent():
    host_name = html.var("host")
    if not host_name:
        raise MKGeneralException("Missing host name")

    validate_host_name(host_name)
    validate_host_permission(host_name)

    host_secret = create_auth_secret()
    update_agent_deployment_status(host_name, {
        "last_registered"  : time.time(),
        "host_secret"      : host_secret,
        })
    html.write("%s\n" % host_secret)


# This page has two modes:
# mode=status -> Query target agent state
# mode=agent  -> Download actual agent
# It is called *unauthenticated*! Authentication is done on
# a per-host-key base. The key is established during host registry.
def page_deploy_agent():
    config.login_super_user()
    try:
        do_deploy_agent()
    except Exception, e:
        if config.debug:
            raise
        html.write("ERROR: %s\n" % e)


def do_deploy_agent():
    # Note: this page is being called unauthenticated. Authentication
    # is done via host and host_secret
    host_name = html.var("host")
    validate_deploy_authentication(host_name)

    opsys = html.var("os")
    if opsys not in agent_package_types:
        raise MKGeneralException(_("Invalid or no package type specified"))

    update_agent_deployment_status(host_name, {
        "last_contact"     : time.time(),
        "installed_aghash" : html.var("installed_aghash") or None,
        "last_error"       : html.var("last_error") or None,
        "opsys"            : opsys,
    })

    mode = html.var("mode")
    if mode not in ( "agent", "status" ):
        raise MKGeneralException(_("Missing or invalid mode. Allowed are 'agent' and 'status'"))

    status = load_agent_deployment_status(host_name)
    aghash = target_agent_hash(host_name)
    if aghash:
        status["agent_available"] = True
        status["aghash"] = aghash
        status["signatures"] = load_agent_signatures(aghash, opsys)
    else:
        status["agent_available"] = False

    if mode == "agent":
        deployment_send_agent(host_name, status, aghash, opsys)
    else:
        deployment_send_status(status)


def deployment_send_agent(host_name, status, aghash, opsys):
    if status["agent_available"]:
        target_hash = html.var("aghash")
        if not target_hash:
            raise MKGeneralException("Missing target aghash")

        if not re.match("^[0-9a-f]{16}$", target_hash):
            raise MKGeneralException("Invalid target aghash")

        if target_hash != aghash: # Important for security also! Host must not download foreign agent
            raise MKGeneralException("Existing agent hash '%s' does not match target agent hash")

        # We just disable the download, not the whole feedback and diagnostic stuff
        if not config.agent_deployment_enabled:
            raise MKGeneralException(_("Agent deployment is currently globally disabled on the deployment server."))

        try:
            check_deployment_enabling_for_host( host_name) # raises Exceptions
        except Exception, e:
            raise MKGeneralException(_("Agent updates are disabled for this host: %s") % e)

        basepath = defaults.var_dir + "/agents/" + opsys + "/"
        path = basepath + "_PACKAGES/%s" % aghash
        download_agent(path, basepath, aghash, opsys)
        update_agent_deployment_status(host_name, {
            "last_download"     : time.time(),
            "downloaded_aghash" : aghash,
        })


def deployment_send_status(status):
    if not status["agent_available"]:
        html.write("AgentAvailable: False\n")
        return

    html.write("AgentAvailable: True\n")
    html.write("TargetHash: %s\n" % status["aghash"])
    for certificate, signature in status["signatures"].items():
        html.write("Signature: %s %s\n" % (hexdump(certificate), hexdump(signature)))


def hexdump(s):
    return "".join(["%02x" % ord(x) for x in s])


def validate_deploy_authentication(host_name):
    if not host_name:
        raise MKGeneralException("Missing host name")

    validate_host_name(host_name)
    status = load_agent_deployment_status(host_name)

    if "host_secret" not in status:
        raise MKGeneralException("This host is not registered for deployment")

    if not html.var("host_secret"):
        raise MKGeneralException("Missing host secret")

    if status["host_secret"] != html.var("host_secret"):
        raise MKAuthException("Invalid host secret '%s' for host %s" % (html.var("host_secret"), host_name))


def find_agent_signed_certificates(aghash):
    fully_signed = []
    partially_signed = []
    by_key = {}

    for opsys in agent_package_types:
        signatures = load_agent_signatures(aghash, opsys)
        for certificate in signatures:
            for key_id, key in config.agent_signature_keys.items():
                if key["certificate"] == certificate:
                    by_key.setdefault(key_id, []).append(opsys)

    for key_id, opsyses in by_key.items():
        if len(opsyses) == len(agent_package_types):
            fully_signed.append(key_id)
        else:
            partially_signed.append(key_id)

    return fully_signed, partially_signed


def have_signed_agents():
    for aghash in find_agents():
        fully_signed, partially_signed = find_agent_signed_certificates(aghash)
        if fully_signed or partially_signed:
            return True
    return False


def target_agent_hash(host_name):
    basepath = defaults.var_dir + "/agents/linux_rpm/"
    path = basepath + host_name
    if os.path.exists(path):
        aghash = os.readlink(path).split("/")[1]
        return aghash
    else:
        return None


def load_agent_signatures(aghash, opsys):
    path = agent_path(aghash, opsys) + ".sig"
    if os.path.exists(path):
        return eval(file(path).read())
    else:
        return {} # dict from certificate (PEM) to signature


def save_agent_signatures(aghash, opsys, signatures):
    path = agent_path(aghash, opsys) + ".sig"
    file(path, "w").write("%s\n" % pprint.pformat(signatures))


def load_agent_deployment_status(host_name):
    make_deployment_dir()
    if os.path.exists(deployment_status_path(host_name)):
        # There were cases in the past where the file could end up as empty
        # file. Handle this case as if the file was not existant.
        content = file(deployment_status_path(host_name)).read()
        if content:
            return eval(content)

    return {}


def save_agent_deployment_status(host_name, status):
    make_deployment_dir()
    file(deployment_status_path(host_name) + ".@NEW", "w").write(repr(status) + "\n")
    os.rename(deployment_status_path(host_name) + ".@NEW", deployment_status_path(host_name))


def update_agent_deployment_status(host_name, update):
    status = load_agent_deployment_status(host_name)
    status.update(update)
    save_agent_deployment_status(host_name, status)


def deployment_status_path(host_name):
    return defaults.var_dir + "/agent_deployment/" + host_name


def make_deployment_dir():
    path = defaults.var_dir + "/agent_deployment"
    if not os.path.exists(path):
        os.makedirs(path)


def create_auth_secret():
    return get_random_string(64, ord('a'), ord('z'))


def get_cached_deployment_status(host_name):
    if not html.is_cached("agent_deployment_status"):
        html.set_cache("agent_deployment_status", {})
    all_hosts = html.get_cached("agent_deployment_status")
    if host_name not in all_hosts:
        status = load_agent_deployment_status(host_name)
        status["target_aghash"] = target_agent_hash(host_name)
        all_hosts[host_name] = status
        return status
    else:
        return all_hosts[host_name]


def agent_needs_update(status):
    return status.get("target_aghash") and status["target_aghash"] != status.get("installed_aghash")


def has_baked_agent(status):
    return status.get("target_aghash") != None


def has_downloaded_agent(status):
    return status.get("last_download") != None


def host_has_error(status):
    return not not status.get("last_error")


def host_is_registered(status):
    return "host_secret" in status


def check_deployment_enabling_for_host(host_name):
    # Example for config.agent_deployment_host_selection
    # {'match_exclude_hosts': ['srv01234'],
    #  'match_folder': 'heimat',
    #  'match_hostgroups': ['linux'],
    #  'match_hosts': ['srv12', 'srv34'],
    #  'match_hosttags': ['prod', '!dmz', 'ip-v4', '!ip-v6'],
    # }
    selection = config.agent_deployment_host_selection
    if 'match_hosts' in selection and host_name not in selection['match_hosts']:
        raise MKGeneralException(_("This host is not listed in the explicit list of host names."))

    if 'match_exclude_hosts' in selection and host_name in selection['match_exclude_hosts']:
        raise MKGeneralException(_("This host is on the exclusion list."))

    # Get information about this host via Livestatus
    response = html.live.query("GET hosts\nColumns: custom_variables groups\nFilter: host_name = %s\n" % host_name)
    if len(response) < 1:
        raise MKGeneralException(_("This host is not known in the activated monitoring configuration"))

    if 'match_hostgroups' in selection:
        deployment_check_host_groups(response[0][1], selection["match_hostgroups"])

    if 'match_hosttags' in selection:
        deployment_check_host_tags(response[0][0].get("TAGS"), selection["match_hosttags"])

    if 'match_folder' in selection:
        host_folder = response[0][0].get("FILENAME", "")
        selection_folder = "/wato/" + selection["match_folder"]
        if not host_folder.startswith(selection_folder):
            raise MKGeneralException(_("This host in not in the selected WATO folder."))


def deployment_check_host_tags(tags_string, taglist):
    tags_of_host = tags_string.split()
    for tag_condition in taglist:
        if tag_condition[0] == '!':
            required_tag = tag_condition[1:]
            negate = True
        else:
            required_tag = tag_condition
            negate = False
        if (required_tag in tags_of_host) == negate:
            raise MKGeneralException(_("This host failed host tag condition '%s' (it has the tags %s)") % (
                tag_condition, ", ".join(tags_of_host)))


def deployment_check_host_groups(host_groups, sufficient_host_groups):
    for group in host_groups:
        if group in sufficient_host_groups:
            return
    raise MKGeneralException(_("None of the groups of this host (%s) is listed in the selected groups (%s)") % (
        ", ".join(host_groups), ", ".join(sufficient_host_groups)))


#.
#   .--Download------------------------------------------------------------.
#   |            ____                      _                 _             |
#   |           |  _ \  _____      ___ __ | | ___   __ _  __| |            |
#   |           | | | |/ _ \ \ /\ / / '_ \| |/ _ \ / _` |/ _` |            |
#   |           | |_| | (_) \ V  V /| | | | | (_) | (_| | (_| |            |
#   |           |____/ \___/ \_/\_/ |_| |_|_|\___/ \__,_|\__,_|            |
#   |                                                                      |
#   +----------------------------------------------------------------------+
#   |  Page for manual download of an agent (need normal login)            |
#   '----------------------------------------------------------------------'

def page_download_agent():
    aghash = html.var("hash")
    if aghash and not re.match("^[a-f0-9]{16}$", aghash):
        raise MKGeneralException(_("Invalid configuration hash"))

    # Alternative: host name
    hostname = html.var("host")
    validate_host_name(hostname)

    if (not aghash and not hostname) or (aghash and hostname):
        raise MKGeneralException(_("Please specify hostname or hash."))

    validate_host_permission(hostname)

    opsys = html.var("os")
    if opsys not in agent_package_types:
        raise MKGeneralException(_("Invalid or no package type specified"))

    basepath = defaults.var_dir + "/agents/" + opsys + "/"
    if aghash:
        path = basepath + "_PACKAGES/" + aghash
    else:
        path = basepath + hostname
        if not os.path.exists(path):
            raise MKGeneralException(_("No baked agent for this host available"))
        aghash = os.readlink(path).split("/")[1]

    download_agent(path, basepath, aghash, opsys)
