#!/usr/bin/python
# -*- encoding: utf-8; py-indent-offset: 4 -*-
# .------------------------------------------------------------------------.
# |                ____ _               _        __  __ _  __              |
# |               / ___| |__   ___  ___| | __   |  \/  | |/ /              |
# |              | |   | '_ \ / _ \/ __| |/ /   | |\/| | ' /               |
# |              | |___| | | |  __/ (__|   <    | |  | | . \               |
# |               \____|_| |_|\___|\___|_|\_\___|_|  |_|_|\_\              |
# |                                        |_____|                         |
# |             _____       _                       _                      |
# |            | ____|_ __ | |_ ___ _ __ _ __  _ __(_)___  ___             |
# |            |  _| | '_ \| __/ _ \ '__| '_ \| '__| / __|/ _ \            |
# |            | |___| | | | ||  __/ |  | |_) | |  | \__ \  __/            |
# |            |_____|_| |_|\__\___|_|  | .__/|_|  |_|___/\___|            |
# |                                     |_|                                |
# |                     _____    _ _ _   _                                 |
# |                    | ____|__| (_) |_(_) ___  _ __                      |
# |                    |  _| / _` | | __| |/ _ \| '_ \                     |
# |                    | |__| (_| | | |_| | (_) | | | |                    |
# |                    |_____\__,_|_|\__|_|\___/|_| |_|                    |
# |                                                                        |
# | mathias-kettner.com                                 mathias-kettner.de |
# '------------------------------------------------------------------------'
#  This file is part of the Check_MK Enterprise Edition (CEE).
#  Copyright by Mathias Kettner and Mathias Kettner GmbH.  All rights reserved.
#
#  Distributed under the Check_MK Enterprise License.
#
#  You should have  received  a copy of the Check_MK Enterprise License
#  along with Check_MK. If not, email to mk@mathias-kettner.de
#  or write to the postal address provided at www.mathias-kettner.de

# Choords:
# 0,0 is at the *bottom* left of the page. When you specify
# left and top, then a larger top is nearer to the top of the
# page.

import defaults, config
import tempfile
import subprocess
import os, sys, StringIO, random

# mod_python does not include all neccessary paths on Ubuntu
sys.path.append('/usr/lib/python2.7/dist-packages')
sys.path.append('/usr/lib/python2.6/dist-packages')

# Import software from reportlab (thanks to them!)
from reportlab.pdfgen import canvas
from reportlab.lib.units import mm
from reportlab.lib.utils import ImageReader
from PIL import PngImagePlugin

def from_mm(dim):
    if type(dim) in (int, float):
        return dim * mm
    else:
        return map(lambda x: x * mm, dim)

# Constants used for conveniance for internal use
white       = (1.0, 1.0, 1.0)
black       = (0.0, 0.0, 0.0)
green       = (0.4, 1.0, 0.4)
yellow      = (1.0, 1.0, 0.0)
orange      = (1.0, 0.6, 0.3)
red         = (1.0, 0.4, 0.4)
blue        = (0.4, 0.6, 1.0)
gray        = (0.5, 0.5, 0.5)
lightgray   = (0.8, 0.8, 0.8)

# Note: these are monitoring specific colors. They do really not belong
# here.
css_class_colors = {
    "up"          : green,
    "down"        : red,
    "unreach"     : orange,
    "pending"     : lightgray,
    "ok"          : green,
    "warn"        : yellow,
    "crit"        : red,
    "unknown"     : orange,
    "hstate0"     : green,
    "hstate1"     : yellow,
    "hstate2"     : red,
    "hstatep"     : lightgray,
    "state0"      : green,
    "state1"      : yellow,
    "state2"      : red,
    "state3"      : orange,
    "statep"      : lightgray,
    "downtime"    : blue,
    "unmonitored" : lightgray,
    "hostdown"    : (0.2, 0.3, 0.8),
    "flapping"    : (1.0, 0.0, 1.0),
    "ooservice"   : (0.8, 0.8, 0.8),
    "chaos"       : (0.5, 0.3, 1.0),
}

# Make a color darker. v ranges from 0 (not darker) to 1 (black)
def darken_color(rgb, v):
    def darken(x, v):
        return x * (1.0 - v)
    return tuple([ darken(x, v) for x in rgb ])

# Make a color lighter. v ranges from 0 (not lighter) to 1 (white)
def lighten_color(rgb, v):
    def lighten(x, v):
        return 1.0 - ((1.0 - x) * (1.0 - v))
    return tuple([ lighten(x, v) for x in rgb ])



# All dimensions that the user (caller) uses are in mm. All our internal
# state variables are in the native PDF dimension, which is convered by
# reportlab.lib.units.mm. The function from_mm convert from user-style into
# internal-style. In a later version we could have the user himself decide
# about the unit he wants to use.

class Document:
    def __init__(self, **args):
        # Static paper settings for this document
        self._pagesize            = from_mm(args["pagesize"])
        self._margins             = from_mm(args["margins"])
        self._mirror_margins      = args["margins"]
        self._pagebreak_function  = args.get("pagebreak_function")
        self._pagebreak_arguments = args.get("pagebreak_arguments", [])

        # set derived helper variables (all in pt)
        self._width, self._height = self._pagesize
        self._height = self._pagesize[1]

        self._margin_top, self._margin_right, \
            self._margin_bottom, self._margin_left = self._margins

        self._inner_width  = self._width -  self._margin_left - self._margin_right
        self._inner_height = self._height - self._margin_top  - self._margin_bottom
        self._left         = self._margin_left
        self._right        = self._width -  self._margin_right
        self._top          = self._height - self._margin_top
        self._bottom       = self._margin_bottom

        # create PDF document
        self._output_buffer = StringIO.StringIO()
        self._canvas = canvas.Canvas(self._output_buffer, pagesize=self._pagesize)

        # initialize our page state
        self._page_number      = 1
        self._tabstops         = []
        self._heading_numbers  = {}
        self._linepos          = self._top # current vertical cursor position

        # Create initial graphics state, i.e. all style settings,
        # that can change while the document is being rendered. We keep a stack
        # a graphics states to that can be pushed and from that can be pulled.
        self._gfx_state = {
            "font_family"      : args["font_family"],
            "font_size"        : args["font_size"],  # pt
            "font_zoom_factor" : 1.0,
            "line_width"       : 0.05 * mm,
            "line_height"      : args["lineheight"], # in relation to font_size
            "fill_color"       : black,
            "dashes"           : None,
            "line_color"       : black,
            "bold"             : False,
            "tt"               : False,
        }
        self._gfx_state_stack = []
        self.set_gfx_state()


    def end(self, sendas=None, do_send=True):
        self._canvas.showPage()
        self._canvas.save()
        if do_send:
            html.set_output_format("pdf")

            if type(sendas) == unicode:
                sendas = sendas.encode("utf-8")
            html.req.headers_out["Content-Disposition"] = "attachment; filename=" + sendas
            html.write(self._output_buffer.getvalue())
            self._output_buffer.close()
        else:
            pdf_source = self._output_buffer.getvalue()
            self._output_buffer.close()
            return pdf_source


    # Methods dealing with manipulating the graphics state (font size, etc.)

    def save_state(self):
        self._gfx_state_stack.append(self._gfx_state.copy())
        self._canvas.saveState() # needed for clip rect

    def restore_state(self):
        # Problem here: After a page break the internal state stack
        # of the canvas object is cleared. But page breaks can happen
        # at many places implicitely! Calling restoreState() with a
        # cleared state stack will first write a state pop command into
        # the PDF and *then* raise an exception. So even if we catch
        # the exception the PDF will be inconsistent.
        if self._canvas.state_stack:
            self._canvas.restoreState() # needed for clip rect
        self._gfx_state = self._gfx_state_stack.pop()
        self.set_gfx_state()

    # Private function for making all graphics settings active. We could
    # optimizie this by only changing the attributes actually being modified,
    # but the use of that is probably neglectable.
    def set_gfx_state(self):
        s = self._gfx_state
        self._canvas.setFillColorRGB(*s["fill_color"])
        self._canvas.setStrokeColorRGB(*s["line_color"])
        self._canvas.setLineWidth(s["line_width"])
        if s["dashes"]:
            self._canvas.setDash(*s["dashes"])
        else:
            self._canvas.setDash([])
        if s["tt"]:
            family = "Courier"
        else:
            family = s["font_family"]
        if s["bold"]:
            family += "-Bold"
        if family == "Times":
            family += "-Roman"
        self._canvas.setFont(family, s["font_size"] * s["font_zoom_factor"])

    def set_font_bold(self, b = True):
        self._gfx_state["bold"] = b
        self.set_gfx_state()

    def set_font_tt(self, tt = True):
        self._gfx_state["tt"] = tt
        self.set_gfx_state()

    def set_font_zoom(self, zoom):
        self._gfx_state["font_zoom_factor"] = zoom
        self.set_gfx_state()

    def set_font_size(self, s):
        self._gfx_state["font_size"] = s
        self.set_gfx_state()

    def set_line_width(self, w):
        self._gfx_state["line_width"] = w * mm
        self.set_gfx_state()

    def set_dashes(self, dashes):
        self._gfx_state["dashes"] = [ d * mm for d in dashes ]
        self.set_gfx_state()

    def reset_colors(self):
        self.set_fill_color(black)
        self.set_line_color(black)

    def set_fill_color(self, color):
        self._gfx_state["fill_color"] = color
        self.set_gfx_state()

    def set_line_color(self, color):
        self._gfx_state["line_color"] = color
        self.set_gfx_state()

    def set_font_color(self, color):
        self._gfx_state["fill_color"] = color
        self._gfx_state["line_color"] = color
        self.set_gfx_state()

    def get_available_fonts(self):
        return self._canvas.getAvailableFonts()


    # Page handling
    def page_number(self):
        return self._page_number

    def next_page(self):
        self._canvas.showPage()
        self._linepos = self._top
        self._page_number += 1
        if self._mirror_margins:
            self._margin_left, self._margin_right = self._margin_right, self._margin_left
            self._left         = self._margin_left
            self._right        = self._width -  self._margin_right

    def need_pagebreak(self, needed_space = 0.0):
        break_at = self._bottom
        return self._linepos - needed_space * mm <= break_at

    def check_pagebreak(self, needed_space = 0.0):
        if self.need_pagebreak(needed_space):
            self.do_pagebreak()

    def do_pagebreak(self):
        self.next_page()
        if self._pagebreak_function:
            self._pagebreak_function(*self._pagebreak_arguments)


    # Functions for direct rendering at absolut places of the page. All positions
    # are specified in a top level way and consider page borders:

    # Examples for positions:
    # ( "w", 10.0 )     10 mm from left,   vertically centered
    # ( "s", 10.0 )     10 mm from bottom, horizontally centered
    # ( "ne", (5, 7) )  5 mm from top, 7 mm from right
    # "c"               At the center
    # Returns the choords of the top left corner of the element (left, top)
    def convert_position(self, position, el_width, el_height):
        h_center = self._left + (self._right - self._left) / 2 - el_width / 2
        v_center = self._bottom + (self._top - self._bottom) / 2 - el_height /2

        if position == "c":
    	    return h_center, v_center
        else:
            anchor = position[0]
            if len(anchor) == 1: # s, n, w, e
                offset = position[1] * mm
                if anchor == "n":
                    return h_center, self._top - offset
                elif anchor == "s":
                    return h_center, self._bottom + offset + el_height
                elif anchor == "w":
                    return self._left + offset, v_center
                elif anchor == "e":
                    return self._right - offset - el_width, v_center

            else: # nw, sw, ne, se
                h_offset = position[1][0] * mm
                v_offset = position[1][1] * mm
                if anchor[0] == "n":
                    y = self._top - v_offset - el_height
                else:
                    y = self._bottom + v_offset
                if anchor[1] == "w":
                    x = self._left + h_offset
                else:
                    x = self._right - h_offset - el_width
                return x, y

    def place_hrule(self, position, width=0.05, color=None):
        el_width = self._right - self._left
        el_height = width * mm
        l, t = self.convert_position(position, el_width, el_height)
        self.save_state()
        if color:
            self.set_fill_color(color)
        self._canvas.rect(l, t, el_width, el_height, fill=1, stroke=0)
        self.restore_state()

    def place_text(self, position, text):
        el_height = self._gfx_state["font_size"] * self._gfx_state["font_zoom_factor"]
        el_width = self._canvas.stringWidth(text)
        l, t = self.convert_position(position, el_width, el_height)
        self._canvas.drawString(l, t + el_height * 0.22, text) # Try to move to correct Y position
        # Debug the text bounding box
        # self._canvas.rect(l, t, el_width, el_height, fill=0)

    def place_page_box(self):
        self.save_state()
        self.set_line_width(0.1)
        self.set_line_color(black)
        self._canvas.rect(self._left, self._bottom, self._inner_width, self._inner_height, fill=0)
        self.restore_state()

    def place_pil_image(self, position, pil, width_mm, height_mm, resolution):
        width, height = self.get_image_dimensions(pil, width_mm, height_mm, resolution)
        x, y = self.convert_position(position, width, height)
        ir = ImageReader(pil)
        self._canvas.drawImage(ir, x, y-height, width, height, mask = 'auto')

    # Functions for adding floating text

    def add_paragraph(self, txt):
        lines = self.wrap_text(txt, (self._right - self._left))
        for line in lines:
            self.add_text_line(line)

    def debug(self, *args):
        for arg in args:
            self.add_paragraph(repr(arg))

    def add_heading(self, level, text, numbers):
        next_number = self._heading_numbers.get(level, 0) + 1
        self._heading_numbers[level] = next_number

        for lev, num in self._heading_numbers.items():
            if lev > level:
                self._heading_numbers[lev] = 0

        if numbers:
            numparts = [ self._heading_numbers.get(l, 1) for l in range(1, level+1) ]
            heading = ".".join(map(str, numparts))
            if level == 1:
                heading += "."
            heading += " " + text
        else:
            heading = text

        self.add_margin(7)
        if level == 1:
            self.add_margin(5)

        zoom = {
            1 : 1.8,
            2 : 1.5,
            3 : 1.2,
        }.get(level, 1.0)

        self.save_state()
        self.set_font_zoom(zoom)
        self.set_font_bold()
        self.advance(self.lineskip())
        self._canvas.drawString(self._left, self._linepos, heading)
        self.advance(self.lineskip() / 2)
        self.restore_state()

    # Add vertical white space, skip. If that does not fit onto the current
    # page, then make a page break and *do not* skip!
    def add_margin(self, height = None, force=False):
        if height != None:
            marg = height * mm
        else:
            marg = self.lineskip()

        if self.need_pagebreak(marg):
            self.do_pagebreak()
        else:
            self.margin(marg, force)

    def add_hrule(self, margin=0.1, width=0.05, color=None):
        self._linepos -= margin * mm
        self.save_state()
        self.set_line_width(width)
        if color:
            self.set_line_color(color)
        else:
            self.set_line_color(black)
        self._canvas.line(self._left, self._linepos, self._right, self._linepos)
        self._linepos -= margin * mm
        self.restore_state()

    def add_pil_image(self, pil, width, height, resolution=None, border=True):
        width, height = self.get_image_dimensions(pil, width, height, resolution)

        self.advance(height)
        x = self._left + (self._inner_width - width) / 2; # center
        y = self._linepos
        ir = ImageReader(pil)
        self._canvas.drawImage(ir, x, y, width, height, mask = 'auto')
        if border:
            self._canvas.rect(x, y, width, height, fill = 0)


    # Add space for a rectangular drawing area where the user can draw
    # himself using render_...() functions. Makes sure that this area
    # is completely contained on one page. Skips to the next page if
    # the current page is too full. Returns left, top, width and height
    # of the actually allocated canvas in mm.
    def add_canvas(self, width_mm, height_mm, border_width=0, left_mm = None):
        self.advance(height_mm * mm)

        if left_mm == None:
            left = self._margin_left
        else:
            left = left_mm * mm

        right  = left + width_mm * mm
        bottom = self._linepos
        top    = bottom + height_mm * mm

        if border_width:
            self.save_state()
            self.set_line_width(border_width * mm)
            self._canvas.rect(left, bottom, right-left, top-bottom, fill=0, stroke=1)
            self.restore_state()

        return left/mm, top/mm, (right-left)/mm, (top-bottom)/mm


    # Add one line of text. This line may include horizontal tabulators ('\t').
    # You can set the width of the tabulators with set_tabstops()
    def add_text_line(self, l, bold=False):
        self.check_pagebreak()

        def aligned_string(x, y, t, alignment):
            if alignment == 'l':
                self._canvas.drawString(x, y, t)
            elif alignment == 'r':
                self._canvas.drawRightString(x, y, t)
            else:
                self._canvas.drawCentredString(x, y, t)

        # TODO: das hier ist überflüssig, war nur für die Tabelle
        def aligned_image(x, y, image_path, alignment):
            # FIXME: (1) alignment is being ignored
            # FIXME: (2) the y-padding bug is in add_table. fix it there
            width = self.lineskip()
            height = width
            pil = PngImagePlugin.PngImageFile(fp = file(image_path))
            ir = ImageReader(pil)
            try:
                self._canvas.drawImage(ir, x, y-1.5, width, height, mask = 'auto')
            except:
                raise Exception("Cannot render image %s" % image_path)


        l = l.strip(' ')
        self._linepos -= self.lineskip()
        tab = -1

        for part in l.split('\t'):
            self.save_state()

            if tab >= 0:
                format_chars, x_position = self._tabstops[tab] # added extra tab stop every 20 mm
                if 'b' in format_chars:
                    self.set_font_bold(True)
                if 't' in format_chars:
                    self.set_font_tt(True)
                if 'g' in format_chars:
                    self.set_font_color(gray)
                if 'r' in format_chars:
                    alignment = 'r'
                elif 'c' in format_chars:
                    alignment = 'c'
                else:
                    alignment = 'l'

                # Negative values are interpreted as offset from the right border
                if x_position < 0:
                    x_position = self._right - x_position
            else:
                x_position = 0
                alignment = 'l'
            tab += 1
            abs_x = self._left + x_position
            abs_y = self._linepos + (self._gfx_state["font_size"] * self._gfx_state["font_zoom_factor"]) * 0.2
            aligned_string(abs_x, abs_y, part, alignment)
            self.restore_state()


    def set_tabstops(self, tabstops):
        # t is a list of tab stops. Each entry is either an int
        # or float -> tabstop in mm. Or it is a string that has
        # prefix of characters followed by a number (mm). The
        # characters specify alignment and font style. We convert
        # all this here to a pair ( "bc", 17.2 ) of the alignment
        # characters and the tabstop in *internal* dimensions.
        def convert_tabstop(t):
            if type(t) in (str, unicode):
                formatchars = ""
                while not t[0].isdigit():
                    formatchars += t[0]
                    t = t[1:]
                return formatchars, float(t) * mm
            else:
                return "", float(t) * mm

        self._tabstops = map(convert_tabstop, tabstops)


    def clear_tabstops(self):
        self._tabstops = []


    def wrap_text(self, text, width): # width in internal dimension
        lines = []
        words = text.split()
        line = ""
        while words:
            if len(words) == 1: # No chance for further wrapping
                lines.append(words[0])
                break

            while words and self._canvas.stringWidth(line + " " + words[0]) < width:
                if line:
                    line += " "
                line += words[0]
                words = words[1:]
            if line:
                lines.append(line)
                line = ""
            else:
                break
        return lines


    # Intelligent table renderding with word wrapping and pagination
    def add_table(self, header_texts, raw_rows, font_size, show_headings,
                  padding, spacing, hrules, vrules, rule_width, row_shading):
        self.save_state()
        self.set_font_size(font_size)

        rule_width *= mm
        x_padding, y_padding = from_mm(padding)
        x_spacing, y_spacing = from_mm(spacing)
        if not show_headings:
            header_texts = []

        # Horizontal padding only makes sense if vertical rules are being rendered.
        # if vrules:
        #     x_spacing = 0
        # else:
        #     x_padding = 0

        # if hrules or row_shading["enabled"]:
        #     y_spacing = 0
        # else:
        #     y_padding = 0

        if header_texts:
            num_cols = len(header_texts)
        elif raw_rows:
            num_cols = len(raw_rows[0])
        else:
            return # No headers, empty table. Nothing to show

        # Convert the header and the rows into special renderable objects. Such
        # an object defines functions for handling size and rendering itself.
        # Currently there are three types of entries allowed:
        # 1. ( "icon", "/omd/.../path/to/icon.png" ) --> an image
        # 2. ( "object", ObjectThing               ) --> aleady render object
        # 3. "Some text"
        # Note: Regardless of the type, everything is embedded in a pair of
        # css and the thing, e.g.
        # ( "number", "0.75" ), or ("", ("icon", "/bar/foo.png") )
        # The headers come *without* the css field and are always texts.
        headers = []
        for header_text in header_texts:
            headers.append(TextCell("heading", header_text))

        rows = []
        for raw_row in raw_rows:
            row = []
            rows.append(row)
            for css, entry in raw_row:
                if type(entry) == tuple:
                    if entry[0] == "icon":
                        row.append(IconCell(entry[1]))
                    elif entry[0] == "object":
                        row.append(entry[1])
                    else:
                        raise Exception("Invalid table entry %r in add_table()" % entry)
                else:
                    row.append(TextCell(css, entry))

        # Now we balance the widths of the columns. Each render object has an
        # absolute minimum width (e.g. the width of the longest word) and
        # a maximum width (e.g. the length of the unwrapped text). All dimensions
        # are in internal units (not mm).
        stats = [ [ 0, 0.0, 0.0, 0.0, None, True ] for c in range(num_cols) ]

        for row in (headers and [headers] or []) + rows:
            for col, render_object in enumerate(row):
                max_width = render_object.maximal_width(self) * mm
                min_width = render_object.minimal_width(self) * mm
                is_dynamic = render_object.can_add_dynamic_width()

                stats[col][0] += 1
                stats[col][1] += max_width
                stats[col][2] = max(min_width, stats[col][2])
                stats[col][3] = max(max_width, stats[col][3])
                if not is_dynamic:
                    stats[col][5] = False

        # Compute required total width
        sum_min = sum([s[2] for s in stats])
        sum_max = sum([s[3] for s in stats])

        # Now compute the available width, i.e. take the usable page width
        # and substract spacing and padding.
        available_width = self._inner_width - ((num_cols-1) * x_spacing) - (num_cols * 2 * x_padding)

        # If there is space enough for not breaking single words, then
        # we begin with giving each column the width of their maximal
        # word. The rest is then distributed such that each column gets
        # the space from the rest that is related to its non-wrapping-width.
        # Columns with can_add_dynamic_width() == False will not take
        # part in the remaining-space-distribution

        if sum_min <= available_width:
            remaining = available_width - sum_min
            sum_weight = 0
            for s in stats:
                row_count, total_width, min_width, max_width, _weight, is_dynamic = s
                if is_dynamic:
                    weight = max_width - min_width + 1 * mm # add 1mm in order to avoid zero weights
                    s[4] = weight
                    sum_weight += weight
                else:
                    s[4] = 0

            column_widths = []
            for row_count, total_width, min_width, max_width, weight, is_dynamic in stats:
                if sum_weight > 0:
                    width = min_width + (weight / sum_weight * remaining)
                else:
                    width = min_width
                column_widths.append(width)

        # Not enough space for even printing the table without breaking
        # words in half. Divide space according to sum_max.
        else:
            sum_weight = 0
            for s in stats:
                if s[0]:
                    weight = s[1] / s[0]
                else:
                    weight = 0
                sum_weight += weight
                s[4] = weight

            column_widths = []
            for s in stats:
                width = available_width * s[4] / sum_weight
                column_widths.append(width)

        brutto_column_widths = []
        for nr, netto_width in enumerate(column_widths):
            w = netto_width + 2*x_padding
            if nr == 0:
                w += x_spacing / 2
            elif nr == len(column_widths)-1:
                w += x_spacing / 2
            else:
                w += x_spacing
            brutto_column_widths.append(w)


        def paint_hrule():
            if hrules:
                self.add_hrule(width = rule_width/mm, margin=0)

        def paint_row(row, add_headers_after_pagebreak, row_oddeven):
            # Give each cell information about its final width so it can reorganize internally.
            # This is used for text cells that do the wrapping.
            for column_width, render_object in zip(column_widths, row):
                render_object.set_width(self, column_width)

            # Now - after the text-wrapping - we know the maximum height of all cells
            # a in row and can decide whether it fits on the current page.
            if row:
                row_height = max([ render_object.height(self) * mm for render_object in row ])
            else:
                row_height = self.lineskip()

            needed_vspace = row_height + 2*y_padding + y_spacing
            if self.need_pagebreak(needed_space = needed_vspace / mm):
                self.do_pagebreak()
                even = "odd"
                if add_headers_after_pagebreak:
                    paint_headers()

            def paint_vrule(left):
                if vrules:
                    self._canvas.setLineWidth(rule_width)
                    self._canvas.setStrokeColorRGB(*black)
                    self._canvas.line(left, self._linepos, left, self._linepos - row_height - 2*y_padding)

            # Apply row shading
            if row_shading["enabled"]:
                h = (row_height + 2*y_padding)/mm
                self.render_rect(self._left/mm, self._linepos/mm - h, self._inner_width/mm, h, fill_color=row_shading[row_oddeven])

            # Finally paint
            left = self._left
            for nr, (column_width, render_object) in enumerate(zip(column_widths, row)):
                height_mm = render_object.height(self)

                brutto_width_mm = column_width/mm + 2*x_padding/mm
                brutto_height_mm = height_mm + 2*y_padding/mm
                render_object.render(self,
                        left/mm, self._linepos/mm,
                        column_width/mm + 2*x_padding/mm, (row_height+2*y_padding)/mm,
                        x_padding/mm, y_padding/mm,
                        row_oddeven)

                paint_vrule(left)
                left += column_width + 2*x_padding + x_spacing
            paint_vrule(left)
            self.advance(needed_vspace - y_spacing/2)
            paint_hrule()
            self.advance(y_spacing/2)

        def paint_headers():
            if headers:
                paint_hrule()
                paint_row(headers, add_headers_after_pagebreak=False, row_oddeven="heading")
                if not hrules:
                    paint_hrule()

        paint_headers()
        row_oddeven = "even"
        for row in rows:
            row_oddeven = row_oddeven == "even" and "odd" or "even"
            paint_row(row, add_headers_after_pagebreak=True, row_oddeven=row_oddeven)

        self.restore_state()

    # Lowlevel functions for direct rendering into the page. Dimensions are in
    # mm. Positions are from the top left of the physical page. These functions
    # are e.g. being used in the render functions of Perf-O-Meters.

    def render_text(self, left_mm, top_mm, text, align="left", bold=False, color=black):
        self.save_state()
        self.set_font_bold(bold)
        self.set_font_color(color)
        if align == "left":
            self._canvas.drawString(left_mm * mm, top_mm * mm, text)
        elif align == "right":
            self._canvas.drawRightString(left_mm * mm, top_mm * mm, text)
        else:
            self._canvas.drawCentredString(left_mm * mm, top_mm * mm, text)
        self.restore_state()

    def render_image(self, left_mm, top_mm, width_mm, height_mm, path):
        pil = PngImagePlugin.PngImageFile(fp = file(path))
        ir = ImageReader(pil)
        try:
            self._canvas.drawImage(ir, left_mm*mm, top_mm*mm, width_mm*mm, height_mm*mm, mask = 'auto')
        except Exception, e:
            raise Exception("Cannot render image %s: %s" % (path, e))


    def get_line_skip(self):
        return self.lineskip() / mm

    def text_width(self, text):
        return self._canvas.stringWidth(text) / mm

    # TODO: unify with render_text()
    def render_aligned_text(self, left_mm, top_mm, width_mm, height_mm, text, align="center", valign="bottom", bold=False):
        if bold:
            self.save_state()
            self.set_font_bold()
        ex_height = self.font_height()
        top = top_mm * mm + (height_mm * mm - ex_height) / 2
        if valign == "middle":
            top -= ex_height * 0.5 # estimate
        if align == "center":
            self._canvas.drawCentredString((left_mm + width_mm/2.0) * mm, top, text)
        elif align == "left":
            self._canvas.drawString(left_mm * mm, top, text)
        elif align == "right":
            self._canvas.drawRightString((left_mm + width_mm) * mm, top, text)
        if bold:
            self.restore_state()

    def render_rect(self, left_mm, top_mm, width_mm, height_mm, line_width = None, line_color=None, fill_color=None):
        self.save_state()

        # Default to unfilled rect with fine black outline
        if line_color == None and fill_color == None and line_width == None:
            line_width = 0.05

        if line_width != None and line_color == None:
            line_color = black

        if line_width:
            self.set_line_width(line_width)
        if line_color:
            self.set_line_color(line_color)
        if fill_color:
            self.set_fill_color(fill_color)

        self._canvas.rect(left_mm * mm, top_mm * mm, width_mm * mm, height_mm * mm,
                          fill = fill_color and 1 or 0,
                          stroke = (line_color or line_width) and 1 or 0)

        self.restore_state()


    def render_line(self, left1_mm, top1_mm, left2_mm, top2_mm, width = 0.05, color = black, dashes=None):
        self.save_state()
        self.set_line_width(width)
        self.set_line_color(color)
        if dashes:
            self.set_dashes(dashes)
        self._canvas.line(left1_mm * mm, top1_mm * mm, left2_mm * mm, top2_mm * mm)
        self.restore_state()

    # Access to paths
    def begin_path(self):
        self._path = self._canvas.beginPath()

    def move_to(self, left, top):
        self._path.moveTo(left * mm, top * mm)

    def line_to(self, left, top):
        self._path.lineTo(left * mm, top * mm)

    def close_path(self):
        self._path.close()

    def fill_path(self, color, gradient=None):
        self.save_state()
        # Older versions of reportlab do not support gradients
        try:
            self._canvas.linearGradient
        except:
            gradient = None

        if gradient:
            grad_left, grad_top, grad_width, grad_height, color_range, switch_points = gradient
            self._canvas.saveState()
            self._canvas.clipPath(self._path, stroke=0)
            self._canvas.linearGradient(grad_left * mm, grad_top * mm, grad_width * mm, grad_height * mm,
                                        color_range, switch_points, extend=False)
            self._canvas.restoreState()
        else:
            self.set_fill_color(color)
            self._canvas.drawPath(self._path, stroke=0, fill=1)

        self.restore_state()

    def stroke_path(self, color=black, width=0.05):
        self.save_state()
        self.set_line_color(color)
        self.set_line_width(width * mm)
        self._canvas.drawPath(self._path, stroke=1, fill=0)
        self.restore_state()


    def add_clip_rect(self, left, top, width, height):
        clip_path = self._canvas.beginPath()
        clip_path.moveTo(left * mm, top * mm)
        clip_path.lineTo(left * mm + width * mm, top * mm)
        clip_path.lineTo(left * mm + width * mm, top * mm + height * mm)
        clip_path.lineTo(left * mm, top * mm + height * mm)
        clip_path.lineTo(left * mm, top * mm)
        clip_path.close()
        self._canvas.clipPath(clip_path, stroke=0)


    # Internal functions

    # Compute the absolute distance between two lines
    def lineskip(self):
        return self._gfx_state["line_height"] * self._gfx_state["font_zoom_factor"] * self._gfx_state["font_size"]

    # Estimate the height of a one-line text
    def font_height(self):
        return self._canvas.stringWidth('M') * 1

    def advance(self, l):
        if self._linepos - l < self._bottom:
            self.do_pagebreak()
            self._linepos = max(self._bottom, self._linepos - l)
        else:
            self._linepos -= l

    def hilite_linepos(self):
        self.save_state()
        self.set_line_color(green)
        self.set_line_width(0.5)
        self._canvas.line(self._left, self._linepos, self._left + .5*mm, self._linepos)
        self._canvas.drawString(self._left, self._linepos, str(int(self._linepos)))
        self.restore_state()

    # Current vertical position of cursor
    def line_pos(self):
        return self._linepos / mm

    def left(self):
        return self._left / mm

    def right(self):
        return self._right / mm

    def width(self):
        return self.right() - self.left()

    # Insert a vertical margin of m. If this leads to a page
    # break then do *not* insert that margin at the top of the new
    # page. Also do not insert that margin if we are already at the
    # top of the page. But: you can force the margin
    def margin(self, m, force=False):
        if self._linepos != self._top or force:
            self._linepos -= m
            if self._linepos < self._bottom:
                self.do_pagebreak()

    def rect(self, x, y, width, height, color):
        self.save_state()
        self.set_fill_color(color)
        self._canvas.rect(x, y, width, height, fill=1, stroke=0)
        self.restore_state()

    # Get and compute image dimensions, convert from mm. If resolution
    # is set, then width and height are being ignored.
    def get_image_dimensions(self, pil, width_mm, height_mm, resolution_dpi = None):
        # Get bounding box of image in order to get aspect (width / height)
        bbox = pil.getbbox()
        pix_width, pix_height = bbox[2], bbox[3]
        if resolution_dpi != None:
            resolution_mm = resolution_dpi / 2.45
            resolution_pt = resolution_mm / mm # now we have pixels / pt
            width = pix_width / resolution_pt
            height = pix_height / resolution_pt
        else:
            aspect = float(pix_width) / float(pix_height)
            # Both are unset, and no resolution: scale to inner width
            if width_mm == None and height_mm == None:
                width = (self._right - self._left);
                height = width / aspect;
            else: # At least one known
                if width_mm != None:
                    width = width_mm * mm
                if height_mm != None:
                    height = height_mm * mm
                if width_mm == None:
                    width = height * aspect;
                elif height_mm == None:
                    height = width / aspect;
        return width, height


# Note: all dimensions this objects handles with are in mm! This is due
# to the fact that this API is also available externally
class TextCell:
    def __init__(self, csses, text):
        self._text = text
        self._bold = False
        self._color = black
        self._bg_color = white
        self._alignment = "left"

        if csses == None:
            csses = ""

        # TODO: Sollte das nicht lieber raus aus dem allgemeinen pdf.py? Ist eigentlich
        # Spezifisch für Views, etc.
        if "heading" in csses or "state" in csses:
            self._bold = True

        if "number" in csses:
            self._alignment = "right"

        if "unused" in csses:
            self._color = (0.6, 0.6, 0.6)

        elif "leftheading" in csses:
            self._bg_color = lightgray

        for css, color in css_class_colors.items():
            if css in csses:
                self._bg_color = color
                self._alignment = "center"

        self._narrow = "narrow" in csses

    def minimal_width(self, pdfdoc): # without padding
        # TODO: consider bold here!
        max_width = 0
        for word in self._text.split():
            max_width = max(max_width, pdfdoc.text_width(word))
        return max_width

    def maximal_width(self, pdfdoc): # without padding
        return pdfdoc.text_width(self._text)

    def can_add_dynamic_width(self):
        return not self._narrow

    def width(self, pdfdoc):
        return self._width

    # Do wrapping of text to actual width. width() and height()
    # can be called only after this has run.
    def set_width(self, pdfdoc, width):
        self._width = width
        self._lines = pdfdoc.wrap_text(self._text, width)

    def height(self, pdfdoc):
        return max(1, len(self._lines)) * pdfdoc.get_line_skip()

    # Render itself at left/top into direction right/down by width/height
    def render(self, pdfdoc, left, top, width, height, x_padding, y_padding, row_oddeven):
        if self._bg_color != white:
            color = self._bg_color
            if row_oddeven == "odd":
                color = lighten_color(color, 0.2)
            pdfdoc.render_rect(left, top-height, width, height, fill_color = color)

        for line in self._lines:
            top -= pdfdoc.get_line_skip()
            y = top - y_padding + (pdfdoc.get_line_skip() - pdfdoc.font_height()/mm) / 2
            if self._alignment == "right":
                x = left + width - x_padding
            elif self._alignment == "left":
                x = left + x_padding
            else:
                x = left + (width/2)
            pdfdoc.render_text(x, y, line, align=self._alignment, bold=self._bold, color=self._color)


# Rendering *one* Multisite-Icon
class IconCell:
    def __init__(self, path):
        self._image_path = path

    def minimal_width(self, pdfdoc):
        return self.height(pdfdoc)

    def maximal_width(self, pdfdoc):
        return self.height(pdfdoc)

    def can_add_dynamic_width(self):
        return False

    def set_width(self, pdfdoc, width):
        pass

    def width(self, pdfdoc):
        return self.height(pdfdoc)

    def height(self, pdfdoc):
        return pdfdoc.get_line_skip()

    def render(self, pdfdoc, left, top, width, height, x_padding, y_padding, row_oddeven):
        w = self.width(pdfdoc)
        pdfdoc.render_image(left+x_padding, top-w-y_padding, w, w, self._image_path)


def random_background_color():
    return tuple([ 0.7 + random.random() * 0.3 for x in range(3)])

def random_foreground_color():
    return tuple([ random.random() * 0.8 for x in range(3)])


#.
#   .--PDF2PNG-------------------------------------------------------------.
#   |              ____  ____  _____ ____  ____  _   _  ____               |
#   |             |  _ \|  _ \|  ___|___ \|  _ \| \ | |/ ___|              |
#   |             | |_) | | | | |_    __) | |_) |  \| | |  _               |
#   |             |  __/| |_| |  _|  / __/|  __/| |\  | |_| |              |
#   |             |_|   |____/|_|   |_____|_|   |_| \_|\____|              |
#   |                                                                      |
#   +----------------------------------------------------------------------+
#   | Generic functions to perform PDF to PNG conversion, e.g. for report  |
#   | thumbnail creation.                                                  |
#   '----------------------------------------------------------------------'

# On RedHat 5 the tool that we need is not available.
# Better check this and do not break the layout
def is_pdf2png_possible():
    return os.path.exists("/usr/bin/pdftoppm")


def pdf2png(pdf_source):
    # Older version of pdftoppm cannot read pipes. The need to seek around
    # in the file. Therefore we need to save the PDF source into a temporary file.
    temp_file = tempfile.NamedTemporaryFile(dir = defaults.tmp_dir)
    temp_file.write(pdf_source)
    temp_file.flush()

    command = ["pdftoppm", "-png", "-f", "1", "-l", "1", "-scale-to", "1000", temp_file.name]
    p = subprocess.Popen(command, shell = False, close_fds = True,
                         stdout = subprocess.PIPE, stderr = subprocess.PIPE)
    stdout, stderr = p.communicate()
    exitcode = p.returncode
    if exitcode != 0:
        raise MKInternalError(_('Cannot create PNG from PDF: %s, Exit code is %d, '
                                'command was "%s", PDF source code was "%s..."')
                 % (stderr, exitcode, " ".join(command), pdf_source[:500]))

    return stdout
