#!/usr/bin/python
# -*- encoding: utf-8; py-indent-offset: 4 -*-
# .------------------------------------------------------------------------.
# |                ____ _               _        __  __ _  __              |
# |               / ___| |__   ___  ___| | __   |  \/  | |/ /              |
# |              | |   | '_ \ / _ \/ __| |/ /   | |\/| | ' /               |
# |              | |___| | | |  __/ (__|   <    | |  | | . \               |
# |               \____|_| |_|\___|\___|_|\_\___|_|  |_|_|\_\              |
# |                                        |_____|                         |
# |             _____       _                       _                      |
# |            | ____|_ __ | |_ ___ _ __ _ __  _ __(_)___  ___             |
# |            |  _| | '_ \| __/ _ \ '__| '_ \| '__| / __|/ _ \            |
# |            | |___| | | | ||  __/ |  | |_) | |  | \__ \  __/            |
# |            |_____|_| |_|\__\___|_|  | .__/|_|  |_|___/\___|            |
# |                                     |_|                                |
# |                     _____    _ _ _   _                                 |
# |                    | ____|__| (_) |_(_) ___  _ __                      |
# |                    |  _| / _` | | __| |/ _ \| '_ \                     |
# |                    | |__| (_| | | |_| | (_) | | | |                    |
# |                    |_____\__,_|_|\__|_|\___/|_| |_|                    |
# |                                                                        |
# | mathias-kettner.com                                 mathias-kettner.de |
# '------------------------------------------------------------------------'
#  This file is part of the Check_MK Enterprise Edition (CEE).
#  Copyright by Mathias Kettner and Mathias Kettner GmbH.  All rights reserved.
#
#  Distributed under the Check_MK Enterprise License.
#
#  You should have  received  a copy of the Check_MK Enterprise License
#  along with Check_MK. If not, email to mk@mathias-kettner.de
#  or write to the postal address provided at www.mathias-kettner.de

try:
    import simplejson as json
except ImportError:
    import json

import pdf
import cStringIO
import textwrap
import base64

# Frequently used variable names:
# translated_metrics: Completely parsed and translated into metrics, e.g. { "foo" : { "value" : 17.0, "unit" : { "render" : ... }, ... } }
# graph_template:     Template for a graph. Essentially a dict with the key "metrics"
# graph:              Dict describing a graph, contains data points, legend labels, abstract sizing, etc.
# start_time:         Unix timestamp of the beginning of the queried range
# end_time:           Unix timestamp of the end of the queried range
# step:               Distance of two consequtive datapoints in seconds

html_size_per_ex = 11.0
min_resize_width = 50
min_resize_height = 6


# Processing of graphs happens in several steps

#   GraphSpecification     Graph definition     Abstract graph         Rendered Graph
# ------------------------------------------------------------------------------------
#   Graph template -----+-> Graph definition --> Abstract graph -+---> HTML5/Canvas
#                       |                                        |
#   Custom graph -------+                                        '---> PDF
#                       |
#   Multi graph --------'

# Graph specification: Some configuration or data that specifies how a graph
#   should be constructed. There are various possible forms of graph specifications.
#   Once specification is a template for host/service graphs plus the specification
#   of one certain host + services.
#
# Graph templates: are defined in plugins/metrics/check_mk.py
#
# Graph definition: contains explicit references to site/host/service/perfvar
#    for each metric. Also information which curves should be rendered in
#    which color, with which title and wether line/area/stack should be done.
#
# Abstract graph: contains actual data points and line/area macros. Also
#    contains V/Y-Achsis with explicit labels. Rendering can be done solely based
#    on that data.
#
# Note: The dict of the abstract graph contains its definition in the key
#    "definition". The definition contains the original specification in the
#    key "specification". That allows to update the graph or even the definition
#    when you just have the graph.


#   .--Default Render Options----------------------------------------------.
#   |                   ____                _                              |
#   |                  |  _ \ ___ _ __   __| | ___ _ __                    |
#   |                  | |_) / _ \ '_ \ / _` |/ _ \ '__|                   |
#   |                  |  _ <  __/ | | | (_| |  __/ |                      |
#   |                  |_| \_\___|_| |_|\__,_|\___|_|                      |
#   |                                                                      |
#   |                   ___        _   _                                   |
#   |                  / _ \ _ __ | |_(_) ___  _ __  ___                   |
#   |                 | | | | '_ \| __| |/ _ \| '_ \/ __|                  |
#   |                 | |_| | |_) | |_| | (_) | | | \__ \                  |
#   |                  \___/| .__/ \__|_|\___/|_| |_|___/                  |
#   |                       |_|                                            |
#   '----------------------------------------------------------------------'

default_graph_render_options = {
    "font_size"       : 11,
    "resizable"       : True,
    "show_controls"   : True,
    "show_legend"     : True,
    "show_service"    : False,
    "preview"         : False,
    "interaction"     : True,
    "editing"         : False,
    "fixed_timerange" : False,
}

def add_default_render_options(graph_render_options):
    options = default_graph_render_options.copy()
    options.update(graph_render_options)
    options.setdefault("size", config.load_user_file("graph_size", (70, 16)))
    return options


#.
#   .--Graph Specification-------------------------------------------------.
#   |      ____                  _  __ _           _   _                   |
#   |     / ___| _ __   ___  ___(_)/ _(_) ___ __ _| |_(_) ___  _ __        |
#   |     \___ \| '_ \ / _ \/ __| | |_| |/ __/ _` | __| |/ _ \| '_ \       |
#   |      ___) | |_) |  __/ (__| |  _| | (_| (_| | |_| | (_) | | | |      |
#   |     |____/| .__/ \___|\___|_|_| |_|\___\__,_|\__|_|\___/|_| |_|      |
#   |           |_|                                                        |
#   +----------------------------------------------------------------------+
#   |  A graph specification is some collection of information from that   |
#   |  one or more graph definitions can be computed. Currently there      |
#   |  exis two types of specifications: "template" and "custom".          |
#   |  Athird one will                                                     |
#   |  follow, which implements graphs derived from templates, showing     |
#   |  data from several services of the same type.                        |
#   |                                                                      |
#   |  A graph specification is a pair of (type, graph_spec..info),        |
#   |  where type is one of "template", "custom"                           |
#   '----------------------------------------------------------------------'

graph_specification_types = {}

def create_graph_definitions_from_specification(graph_specification):
    spec_type, spec_info = graph_specification
    func = graph_specification_types[spec_type]
    return func(spec_info)

def create_graph_definitions_from_explicit(spec_info):
    graph_definition = spec_info.copy()
    graph_definition["specification"] = ( "explicit", spec_info )
    return [ graph_definition ]

graph_specification_types["explicit"] = create_graph_definitions_from_explicit

#.
#   .--Graph Templates-----------------------------------------------------.
#   |            _____                    _       _                        |
#   |           |_   _|__ _ __ ___  _ __ | | __ _| |_ ___  ___             |
#   |             | |/ _ \ '_ ` _ \| '_ \| |/ _` | __/ _ \/ __|            |
#   |             | |  __/ | | | | | |_) | | (_| | ||  __/\__ \            |
#   |             |_|\___|_| |_| |_| .__/|_|\__,_|\__\___||___/            |
#   |                              |_|                                     |
#   +----------------------------------------------------------------------+
#   |  Convert a graph template plus its context (host/service) into an    |
#   |  actual graph definition.                                            |
#   '----------------------------------------------------------------------'

# Example for graph_specification_info:
# {
#     "site"                : None, # site hint, can be a site or None
#     "host_name"           : "server123",
#     "service_description" : "CPU load",
#     "graph_index"         : 0, # first graph, also allows: missing or None -> all graphs
# }

def create_graph_definitions_from_template(graph_specification_info):
    site =  graph_specification_info["site"]
    host_name = graph_specification_info["host_name"]
    service_description = graph_specification_info["service_description"]
    graph_index = graph_specification_info.get("graph_index") # can be None -> show all graphs

    row = get_graph_data_from_livestatus(site, host_name, service_description)

    perf_data_string, translated_metrics, check_command, graph_templates = find_possible_graphs(row)
    site = row["site"]

    graph_definitions = []
    for index, graph_template in enumerate(graph_templates):
        if graph_index == None or index == graph_index:
            graph_definition = create_graph_definition_from_template(graph_template, translated_metrics, site, host_name, service_description)
            # Put the specification of this graph into the graph_definition
            spec_info = graph_specification_info.copy()
            spec_info["graph_index"] = index
            graph_definition["specification"] = ("template", spec_info)
            graph_definitions.append(graph_definition)
    return graph_definitions


def create_graph_definition_from_template(graph_template, translated_metrics, site, host_name, service_description):
    graph_definition = {
        "metrics"                 : [],
        "explicit_vertical_range" : get_graph_range(graph_template, translated_metrics),
        "horizontal_rules"        : horizontal_rules_from_template(graph_template, translated_metrics), # e.g. lines for WARN and CRIT
        "title"                   : None,
        "omit_zero_metrics"       : graph_template.get("omit_zero_metrics", False),
        "consolidation_function"  : graph_template.get("consolidation_function", "max"),
    }

    if graph_template.get("title"):
        graph_definition["title"] = replace_expressions(graph_template["title"], translated_metrics)

    for metric_definition in graph_template["metrics"]:
        metric = parse_template_metric(site, host_name, service_description, metric_definition, translated_metrics, graph_definition)
        if "unit" not in graph_definition:
             graph_definition["unit"] = metric["unit"]
        elif metric["unit"] != graph_definition["unit"]:
            raise MKGeneralException(_("Cannot create graph with metrics with "
                                   "the two different units '%s' and '%s'") %
                                   (metric["unit"]["title"], graph_definition["unit"]["title"]))
        graph_definition["metrics"].append(metric)

        # Use title and label of this metrics as default for the graph
        if metric["title"] and not graph_definition["title"]:
            graph_definition["title"] = metric["title"]

    graph_definition.setdefault("unit", "")
    return graph_definition


def horizontal_rules_from_template(graph_template, translated_metrics):
    horizontal_rules = []
    for entry in graph_template.get("scalars", []):
        if len(entry) == 2:
            expression, title = entry
        else:
            expression = entry
            if expression.endswith(":warn"):
                title = _("Warning")
            elif expression.endswith(":crit"):
                title = _("Critical")
            else:
                title = expression

        try:
            value, unit, color = evaluate(expression, translated_metrics)
            if value:
                horizontal_rules.append((
                    value,
                    unit["render"](value),
                    color,
                    title,
                ))
        # Scalar value like min and max are always optional. This makes configuration
        # of graphs easier.
        except:
            pass

    return horizontal_rules


def parse_template_metric(site, host_name, service_description, metric_definition, translated_metrics, graph_definition):
    # Extract explicit color definition, if one exists
    if "#" in metric_definition[0]:
        expression, custom_color = metric_definition[0].split("#")
    else:
        expression = metric_definition[0]
        custom_color = None

    # Extract explicit unit definition, if one exists
    if "@" in expression:
        expression_candidate, custom_unit_name = expression.split("@")
        if custom_unit_name in unit_info.keys():  # Keep "@" in name if unit string doesn't exist
            expression = expression_candidate
        else:
            custom_unit_name = None
    else:
        custom_unit_name = None

    # Extract explicit title (third tuple argument), if one exists
    if len(metric_definition) >= 3:
        title = metric_definition[2]
    else:
        title = None

    # Determine title and color by trying to evaluate the expression
    # with the current (or dummy) metrics.
    if "," in expression:
        # TODO: beware of division by zero. All metrics are set to 1 here.
        value, unit, color = evaluate(expression, translated_metrics)

    # No RPN expression, just the name of a metric, maybe with suffix .min, .max or .average
    else:
        # expression is fs_growth or fs_growth.min or +1.8V_bus_voltage_voltage (old nagios style)
        if expression.endswith(".min") or expression.endswith(".max") or expression.endswith(".average"):
            metric_name = expression.rsplit(".", 1)[0]
        else:
            metric_name = expression

        if metric_name not in translated_metrics:
            raise MKGeneralException(_("Graph definition uses undefined metric %s, available are %s") %
                (metric_name, ", ".join(sorted(translated_metrics.keys()))))

        mi = translated_metrics[metric_name]
        if not title:
            title = mi["title"]
        color = mi["color"]
        unit = mi["unit"]

    if custom_unit_name:
        unit = unit_info[custom_unit_name]

    if custom_color:
        color = "#" + custom_color

    # Create the RPN needed for RRDTool. Here we need to apply a scaling information.
    # Also we need to query the original variable name, not the translated one.

    # Break the RPN into parts and translate each part separately
    parts = expression.split(",")
    atoms = []
    for part in parts: # var names, operators
        if part.endswith(".max") or part.endswith(".min") or part.endswith(".average"):
            part, consolidation_function = part.rsplit(".", 1)
            if graph_definition["consolidation_function"] != None \
               and consolidation_function != graph_definition["consolidation_function"]:
                raise MKGeneralException(_("The expression \"%s\" uses a different consolidation "
                                           "function as the graph (%s). This is not allowed.") %
                                          (expression, graph_definition["consolidation_function"]))
        else:
            consolidation_function = None

        # Some parts are operators. We leave them. We are just interested in
        # names of metrics.
        if part in translated_metrics: # name of a variable that we know
            tme = translated_metrics[part]
            metric_name = tme.get("orig_name", part) # original name before translation
            # We do the replacement of special characters with _ right here.
            # Normally it should be a task of the core. But: We have variables
            # named "/" - which is very silly, but is due to the bogus perf names
            # of the df check. So the CMC could not really distinguish this from
            # the RPN operator /.
            atoms.append(("rrd", site, host_name, service_description, pnp_cleanup(metric_name), consolidation_function, tme["scale"]))
        else:
            try:
                atoms.append(("constant", float(part)))
            except:
                atoms.append(("operator", part))

    expression = convert_parsed_rpn_to_expression(atoms)

    return {
        "title"      : title,
        "color"      : color,
        "unit"       : unit["id"],
        "expression" : expression,
        "line_type"  : metric_definition[1],
    }

# Convert this:
#
# [('rrd', 'heute', u'heute', u'CPU utilization', u'user', 'min', 1.0),
#  ('rrd', 'heute', u'heute', u'CPU utilization', u'system', 'max', 1.0),
#  ('operator', '+'),
#  ('constant', 2.0),
#  ('operator', '*')]
#
# ... into this:
# ('operator',
#      '*',
#      [('operator',
#        '+',
#        [('rrd',
#          'heute',
#          u'heute',
#          u'CPU utilization',
# ....
def convert_parsed_rpn_to_expression(atoms):
    stack = []
    for atom in atoms:
        if atom[0] == "operator":
            op2 = stack.pop()
            op1 = stack.pop()
            stack.append((atom + ([op1, op2],)))
        else:
            stack.append(atom)

    if len(stack) != 1:
        raise MKGeneralException(_("Invalid RPN expression: %d elements on stack (expecting 1)") % len(stack))

    return stack[0]

graph_specification_types["template"] = create_graph_definitions_from_template

#.
#   .--Layouting the graph-------------------------------------------------.
#   |                  _                            _                      |
#   |                 | |    __ _ _   _  ___  _   _| |_                    |
#   |                 | |   / _` | | | |/ _ \| | | | __|                   |
#   |                 | |__| (_| | |_| | (_) | |_| | |_                    |
#   |                 |_____\__,_|\__, |\___/ \__,_|\__|                   |
#   |                             |___/                                    |
#   +----------------------------------------------------------------------+
#   |  Compute the abstract graph from its definitions by fetching RRD     |
#   |  data, computing time series data and taking layout decisions. The   |
#   |  result - the abstract graph - is fully layouted but still inde-     |
#   |  pendent of the output device (HTML Canvas or PDF).                  |
#   '----------------------------------------------------------------------'

def compute_abstract_graph(graph_definition, graph_data_range, graph_render_options):
    graph_render_options = add_default_render_options(graph_render_options)

    # Fetch all raw RRD data
    rrd_data = fetch_rrd_data_for_graph(graph_definition, graph_data_range)

    # RRDTool aligns start/end/step to its internal precision. This is returned
    # as first 3 values in each RRD data row.
    start_time, end_time, step = pop_actual_range_of_rrd_data(rrd_data)
    if start_time == None: # Empty graph
        start_time, end_time = graph_data_range["time_range"]
        step = 60
    elif chop_last_empty_step(graph_data_range, step, rrd_data):
        end_time -= step

    curves = compute_graph_curves(graph_definition["metrics"], rrd_data)
    if graph_definition.get("omit_zero_metrics"):
        curves = only_nonzero_curves(curves)

    pin_time = load_graph_pin()

    unit = unit_info[graph_definition["unit"]]
    compute_scalars(unit, curves, start_time, end_time, step, pin_time)
    layouted_curves, mirrored = layout_graph_curves(curves)  # do stacking, mirroring

    width, height = graph_render_options["size"]

    return {
        # Labelling, size, layout
        "title"                   : graph_definition.get("title"),
        "width"                   : width, # in widths of lower case 'x'
        "height"                  : height,
        "mirrored"                : mirrored,

        # Actual data and axes
        "curves"                  : layouted_curves,
        "horizontal_rules"        : graph_definition["horizontal_rules"],
        "vertical_axis"           : compute_graph_v_axis(graph_definition, graph_data_range, height, layouted_curves, mirrored),
        "time_axis"               : compute_graph_t_axis(start_time, end_time, width, step),

        # Displayed range
        "start_time"              : start_time,
        "end_time"                : end_time,
        "step"                    : step,
        "explicit_vertical_range" : graph_definition["explicit_vertical_range"],
        "requested_vrange"        : graph_data_range.get("vertical_range"),
        "requested_start_time"    : graph_data_range["time_range"][0],
        "requested_end_time"      : graph_data_range["time_range"][1],
        "requested_step"          : graph_data_range["step"],
        "pin_time"                : pin_time,

        # Definition itself, for reproducing the graph
        "definition"              : graph_definition,
    }


#.
#   .--Fetch RRDData-------------------------------------------------------.
#   |   _____    _       _       ____  ____  ____  ____        _           |
#   |  |  ___|__| |_ ___| |__   |  _ \|  _ \|  _ \|  _ \  __ _| |_ __ _    |
#   |  | |_ / _ \ __/ __| '_ \  | |_) | |_) | | | | | | |/ _` | __/ _` |   |
#   |  |  _|  __/ || (__| | | | |  _ <|  _ <| |_| | |_| | (_| | || (_| |   |
#   |  |_|  \___|\__\___|_| |_| |_| \_\_| \_\____/|____/ \__,_|\__\__,_|   |
#   |                                                                      |
#   +----------------------------------------------------------------------+
#   |  Core for getting the acutal raw data points via Livestatus from RRD |
#   '----------------------------------------------------------------------'

def fetch_rrd_data_for_graph(graph_definition, graph_data_range):
    needed_rrd_data = get_needed_rrd_data(graph_definition["metrics"])
    by_service = group_needed_rrd_data_by_service(needed_rrd_data)
    rrd_data = {}
    for (site, host_name, service_description), entries in by_service.items():
        try:
            for (perfvar, cf, scale), data in \
                fetch_rrd_data(site, host_name, service_description, entries, graph_definition, graph_data_range):
                    rrd_data[(site, host_name, service_description, perfvar, cf, scale)] = data
        except livestatus.MKLivestatusNotFoundError:
            pass
    return rrd_data


def pop_actual_range_of_rrd_data(rrd_data):
    start_time = None
    end_time = None
    step = None
    last_spec_title = None

    for spec, rrddata in rrd_data.items():
        spec_title = "%s/%s/%s" % (spec[1], spec[2], spec[3]) # host/service/perfvar
        if rrddata:
            if start_time == None:
                start_time, end_time, step = rrddata[:3]
                last_spec_title = spec_title
            else:
                if (start_time, end_time, step) != tuple(rrddata[:3]):
                    raise MKGeneralException(
                        _("You have configured the RRDs for the services in this graph with different time resolutions. "
                          "We cannot create a graph, sorry. Please reconfigure the RRDs of the affected services and "
                          "use the same configuration. "
                          "Metric %s has %d, %d, %d, Metric %s has %d, %d, %d") % (
                                        last_spec_title, start_time, end_time, step,
                                        spec_title, rrddata[0], rrddata[1], rrddata[2]))
            del rrddata[:3]
        else:
            raise Exception(_("Cannot get RRD data for %s") % spec_title)

    return start_time, end_time, step


def get_needed_rrd_data(metric_definitions):
    needed = set()
    for metric_definition in metric_definitions:
        expression = metric_definition["expression"]
        for rrd_data in find_rrd_expressions_in_tree(expression):
            needed.add(rrd_data)
    return needed


def group_needed_rrd_data_by_service(needed_rrd_data):
    by_service = {}
    for site, host_name, service_description, perfvar, cf, scale in needed_rrd_data:
        by_service.setdefault((site, host_name, service_description), []).append((perfvar, cf, scale))
    return by_service


def find_rrd_expressions_in_tree(expression):
    if expression[0] == "rrd":
        yield tuple(expression[1:]) # could be converted to list due to JSON
    elif expression[0] == "operator":
        for operand in expression[2]:
            for rrd_data in find_rrd_expressions_in_tree(operand):
                yield rrd_data
    else:
        return


def fetch_rrd_data(site, host_name, service_description, entries, graph_definition, graph_data_range):
    query = livestatus_query_for_rrd_data(host_name, service_description, entries, graph_definition, graph_data_range)
    html.live.set_only_sites(site)
    row = html.live.query_row(query)
    html.live.set_only_sites(None)

    for nr, (perfvar, cf, scale) in enumerate(entries):
        yield (perfvar, cf, scale), row[nr]


def livestatus_query_for_rrd_data(host_name, service_description, entries, graph_definition, graph_data_range):
    start_time, end_time = graph_data_range["time_range"]

    step = graph_data_range["step"]

    lql_columns = []
    for nr, (perfvar, cf, scale) in enumerate(entries):
        if graph_definition["consolidation_function"]:
            cf = graph_definition["consolidation_function"]
        else:
            cf = cf or "max"

        rpn = "%s.%s" % (perfvar, cf)
        if scale != 1.0:
            rpn += ",%f,*" % scale
        lql_columns.append("rrddata:m%d:%s:%d:%d:%d" % (nr, rpn, start_time, end_time, max(1, step)))

    if service_description != "_HOST_":
        query = "GET services\n" \
                "Filter: host_name = %s\n" \
                "Filter: service_description = %s\n" \
                "Columns: %s\n" % (host_name, service_description, " ".join(lql_columns))
    else:
        query = "GET hosts\n" \
                "Filter: host_name = %s\n" \
                "Columns: %s\n" % (host_name, " ".join(lql_columns))

    return query


#.
#   .--Curves--------------------------------------------------------------.
#   |                    ____                                              |
#   |                   / ___|   _ _ ____   _____  ___                     |
#   |                  | |  | | | | '__\ \ / / _ \/ __|                    |
#   |                  | |__| |_| | |   \ V /  __/\__ \                    |
#   |                   \____\__,_|_|    \_/ \___||___/                    |
#   |                                                                      |
#   +----------------------------------------------------------------------+
#   |  Compute the curves from the raw RRD data by evaluating expressions. |
#   '----------------------------------------------------------------------'

# The idea is to omit the empty last step of graphs which are showing the
# last data which ends now (at the current time) where there is not yet
# data available for the current RRD step. Showing an empty space on the
# right of the graph seems a bit odd, so strip of the last (empty) step.
#
# This makes only sense for graphs which are ending "now". So disable this
# for the other graphs.
def chop_last_empty_step(graph_data_range, step, rrd_data):
    # Disable graph chop for graphs which do not end within the current step
    if abs(time.time() - graph_data_range["time_range"][1]) > step:
        return False

    # Chop of one step from the end of the graph if that is None
    # for all curves. This is in order to avoid a gap when querying
    # up to the current time.
    for data in rrd_data.values():
        if not data or data[-1] != None:
            return False
    else:
        for data in rrd_data.values():
            del data[-1]
        return True


def compute_graph_curves(metrics, rrd_data):
    curves = []
    for metric_definition in metrics:
        curves.append({
            "line_type" : metric_definition["line_type"],
            "color"     : metric_definition["color"],
            "title"     : metric_definition["title"],
            "rrddata"   : evaluate_time_series_expression(metric_definition["expression"], rrd_data),
        })
    return curves


def evaluate_time_series_expression(expression, rrd_data):
    if rrd_data:
        num_points = len(rrd_data.values()[0])
    else:
        num_points = 1

    if expression[0] == "operator":
        operator, operands = expression[1:]
        operands_evaluated = [
            evaluate_time_series_expression(a, rrd_data)
            for a in operands
        ]
        return time_series_math(operator, operands_evaluated)
    elif expression[0] == "rrd":
        key = tuple(expression[1:])
        if key in rrd_data:
            return rrd_data[key]
        else:
            return [ None ] * num_points
    else: # constant
        return [ expression[1] ] * num_points


def time_series_math(operator, operands_evaluated):
    if operator not in time_series_operators:
        raise MKGeneralException(_("Undefined operator '%s' in graph expression") % html.attrencode(operator))
    op_title, op_func = time_series_operators[operator]
    def op_func_dealing_with_none(os):
        if None in os:
            return None
        else:
            try:
                return op_func(os)
            except ZeroDivisionError:
                return None


    return map(op_func_dealing_with_none, zip(*operands_evaluated))


time_series_operators = {
    "+" :   ( _("Sum"), sum),
    "*" :   ( _("Product"), lambda os: reduce(operator.mul, os, 1)),
    "-" :   ( _("Difference"), lambda os: os[0] - os[1]),
    "/" :   ( _("Fraction"), lambda os: os[0] / os[1]),
    "MAX" : ( _("Maximum"), max),
    "MIN" : ( _("Minimum"), min),
}


def only_nonzero_curves(curves):
    nonzero_curves = []
    for curve in curves:
        for value in curve["rrddata"]:
            if value != None and value != 0.0:
                nonzero_curves.append(curve)
                break

    return nonzero_curves


#.
#   .--Layout Curves-------------------------------------------------------.
#   |  _                            _      ____                            |
#   | | |    __ _ _   _  ___  _   _| |_   / ___|   _ _ ____   _____  ___   |
#   | | |   / _` | | | |/ _ \| | | | __| | |  | | | | '__\ \ / / _ \/ __|  |
#   | | |__| (_| | |_| | (_) | |_| | |_  | |__| |_| | |   \ V /  __/\__ \  |
#   | |_____\__,_|\__, |\___/ \__,_|\__|  \____\__,_|_|    \_/ \___||___/  |
#   |             |___/                                                    |
#   +----------------------------------------------------------------------+
#   |  Translate mathematical values into points in grid to paint          |
#   '----------------------------------------------------------------------'


# Compute the location of the curves of the graph, implement
# stacking and mirroring (displaying positive values in negative
# direction).
def layout_graph_curves(curves):
    mirrored = False # True if negative area shows positive values

    # Build positive and optional negative stack. The points itself
    # in the stack are always positive
    stacks = [None, None]

    # Compute the logical position (i.e. measured in the original unit)
    # of the data points, where stacking and Y-mirroring is being applied.
    # For areas we put (lower, higher) as point into the list of points.
    # For lines simply the values. For mirrored values from is >= to.
    def add_points(a, b):
        if a == None and b == None:
            return None
        else:
            return (a or 0.0) + (b or 0.0)

    def mult_point(a, m):
        if a == None:
            return None
        else:
            return a * m

    # Make sure that first entry in pair is not greater than second
    def fix_swap(pp):
        lower, upper = pp
        if lower <= upper:
            return pp
        else:
            return upper, lower

    def mirror_point(p):
        if p == None:
            return p
        else:
            return -p

    layouted_curves = []
    for curve in curves:
        if curve.get("dont_paint"):
            continue

        layouted_curve = {
            "color" : curve["color"],
            "title" : curve["title"],
            "scalars" : curve["scalars"],
        }
        layouted_curves.append(layouted_curve)

        line_type = curve["line_type"]
        raw_points = halfstep_interpolation(curve["rrddata"])

        # Handles lines, they cannot stack
        if line_type.endswith("line"):
            if line_type == "line":
                layouted_curve["points"] = raw_points
                layouted_curve["type"] = "line"

            elif line_type == "-line":
                mirrored = True
                layouted_curve["points"] = map(mirror_point, raw_points)
                layouted_curve["type"] = "line"

        # Handle area and stack.
        else:
            layouted_curve["type"] = "area"
            if line_type[0] == '-':
                stack_nr = 1
                line_type = line_type[1]
                mirror = -1
                mirrored = True
            else:
                stack_nr = 0
                mirror = 1

            if line_type == "area" or stacks[stack_nr] == None:
                points = []
                for p in raw_points:
                    if p == None:
                        points.append((None, None))
                    elif mirror == -1:
                        points.append(fix_swap((mult_point(p, mirror), 0.0)))
                    else:
                        points.append(fix_swap((0.0, p)))
                stacks[stack_nr] = [point or 0.0 for point in raw_points]
                layouted_curve["points"] = points

            # real stack
            else:
                new_stack = [ add_points(s, p) for (s, p) in zip(stacks[stack_nr], raw_points) ]
                if mirror == -1:
                    layouted_curve["points"] = [
                        fix_swap((mult_point(e, mirror), mult_point(a, mirror)))
                        for (a, e)
                        in zip(stacks[stack_nr], new_stack) ]
                else:
                    layouted_curve["points"] = map(fix_swap, zip(stacks[stack_nr], new_stack))
                stacks[stack_nr] = new_stack

    return layouted_curves, mirrored


# Result is a list with len(rrddata)*2 + 1 vertical values
def halfstep_interpolation(rrddata):
    if not rrddata:
        return []

    points = [ rrddata[0], rrddata[0] ]
    last_point = rrddata[0]
    for nr, point in enumerate(rrddata[1:]):
        if last_point == None and point == None:
            points += [ None, None ]
        elif last_point == None:
            points += [ point, point ]
        elif point == None:
            points += [ last_point, None ]
        else:
            points += [ (point + last_point) / 2, point ]

        last_point = point

    points += [ last_point ]
    return points


#.
#   .--Scalars-------------------------------------------------------------.
#   |                  ____            _                                   |
#   |                 / ___|  ___ __ _| | __ _ _ __ ___                    |
#   |                 \___ \ / __/ _` | |/ _` | '__/ __|                   |
#   |                  ___) | (_| (_| | | (_| | |  \__ \                   |
#   |                 |____/ \___\__,_|_|\__,_|_|  |___/                   |
#   |                                                                      |
#   +----------------------------------------------------------------------+
#   |  For each curve compute the scalar values min, max, average, last    |
#   |  and value at position o pin.                                        |
#   '----------------------------------------------------------------------'

def compute_scalars(unit, curves, start_time, end_time, step, pin_time):
    for curve in curves:
        if curve.get("dont_paint"):
            continue

        rrddata = curve["rrddata"]
        minimum, maximum, summed, pin, first, last = None, None, 0.0, None, None, None
        count = 0
        for value in rrddata:
            if value != None:
                count += 1
                summed += value
                maximum = max(value, maximum)
                if minimum == None:
                    minimum = value
                else:
                    minimum = min(value, minimum)
                last = value
                if first == None:
                    first = value

        if count:
            average = summed / count
        else:
            average = None

        if pin_time != None:
            pin = get_value_at_timestamp(start_time, end_time, step, pin_time, rrddata)

        scalars = {
            "pin"     : pin,
            "first"   : first,
            "last"    : last,
            "max"     : maximum,
            "min"     : minimum,
            "average" : average,
        }
        curve["scalars"] = {}
        for key, value in scalars.items():
            if value == None:
                curve["scalars"][key] = (None, _("n/a"))
            else:
                curve["scalars"][key] = (value, unit["render"](value))


def get_value_at_timestamp(start_time, end_time, step, pin_time, rrddata):
    nth_value = (pin_time - start_time) / step
    if nth_value >= 0 and nth_value < len(rrddata):
        return rrddata[nth_value]


#.
#   .--Vertical Axis-------------------------------------------------------.
#   |      __     __        _   _           _      _          _            |
#   |      \ \   / /__ _ __| |_(_) ___ __ _| |    / \   __  _(_)___        |
#   |       \ \ / / _ \ '__| __| |/ __/ _` | |   / _ \  \ \/ / / __|       |
#   |        \ V /  __/ |  | |_| | (_| (_| | |  / ___ \  >  <| \__ \       |
#   |         \_/ \___|_|   \__|_|\___\__,_|_| /_/   \_\/_/\_\_|___/       |
#   |                                                                      |
#   +----------------------------------------------------------------------+
#   |  Computation of vertical axix, including labels and range            |
#   '----------------------------------------------------------------------'

# Compute the displayed vertical range and the labelling
# and scale of the vertical axis.
# If mirrored == True, then the graph uses the negative
# v-region for displaying positive values - so show the labels
# without a - sign.
def compute_graph_v_axis(graph_definition, graph_data_range, height, layouted_curves, mirrored):
    real_range, vrange, min_value, max_value = compute_v_axis_min_max(graph_definition,
                                                                      graph_data_range, height,
                                                                      layouted_curves, mirrored)

    # Compute vertical axis
    # Guestimate a useful number of vertical labels
    if height < 10:
        num_v_labels = (height - 4) / 2
    else:
        num_v_labels = (height - 4) / math.log(height)
    label_distance_at_least = float(vrange) / max(num_v_labels, 1)

    # The stepping of the labels is not always decimal, where
    # we choose distances like 10, 20, 50. It can also be "binary", where
    # we have 512, 1024, etc. or "time", where we have seconds, minutes,
    # days
    stepping = unit_info[graph_definition["unit"]].get("stepping", "decimal")

    if stepping == "integer":
        label_distance_at_least = max(label_distance_at_least, 1) # e.g. for unit type "count"

    divide_by = 1

    if stepping == "binary":
        base = 16
        steps = [ (2, 0.5), (4, 1), (8, 2), (16, 4) ]

    elif stepping == "time":
        if max_value > 3600 * 24:
            divide_by = 86400
            base = 10
            steps = [ (2, 0.5), (5, 1), (10, 2) ]
        elif max_value >= 10:
            base = 60
            steps = [ (2, 0.5), (3, 0.5), (5, 1), (10, 2), (20, 5), (30, 5), (60, 10) ]
        else: # ms
            base = 10
            steps = [ (2, 0.5), (5, 1), (10, 2) ]

    else: # "decimal" or "integer"
        base = 10
        steps = [ (2, 0.5), (5, 1), (10, 2) ]

    mantissa, exponent = frexpb(label_distance_at_least / divide_by, base)

    # We draw a label at either 1, 2, or 5 of the choosen
    # exponent
    for step, substep in steps:
        if mantissa <= step:
            mantissa = step
            submantissa = substep
            break

    label_distance = mantissa * (base ** exponent) * divide_by
    sub_distance = submantissa * (base ** exponent) * divide_by

    # We need to round the position of the labels. Otherwise some
    # strange things can happen due to internal precision limitation.
    # Here we compute the number of decimal digits we need

    unit = unit_info[graph_definition["unit"]]

    v_axis = {
        "range"               : (min_value, max_value),
        "real_range"          : real_range,
        "label_distance"      : label_distance,
        "sub_distance"        : sub_distance,
        "axis_label"          : None,
    }

    # Adds "labels", "max_label_length" and updates "axis_label" in case
    # of units which use a graph global unit
    v_axis.update(create_vertical_axis_labels(min_value, max_value, unit,
                                              label_distance, sub_distance,
                                              mirrored))

    return v_axis


def compute_v_axis_min_max(graph_definition, graph_data_range, height, layouted_curves, mirrored):
    min_value, max_value = None, None

    # Now make sure that all points are within the range.
    # Enlarge a given range if neccessary.
    for curve in layouted_curves:
        for point in curve["points"]:

            # Line points
            if type(point) == float:
                max_value = max(max_value, point)
                if min_value == None:
                    min_value = point
                elif point != None:
                    min_value = min(min_value, point)

            # Area points
            elif type(point) == tuple:
                lower, higher = point
                max_value = max(max_value, higher)
                if min_value == None:
                    min_value = lower
                elif lower != None:
                    min_value = min(min_value, lower)

    # If all of our data points are None, then we have no
    # min/max values. In this case we assume 0 as a minimum
    # value and 1 as a maximum. Otherwise we will run into
    # an exception
    if min_value == None and max_value == None:
        min_value = mirrored and -1.0 or 0.0
        max_value = 1.0
    elif min_value == None and max_value > 0:
        min_value = mirrored and -max_value or 0.0
    elif max_value == None and min_value < 0:
        max_value = 1.0
    elif min_value == None:
        if mirrored:
            min_value = - max_value
        else:
            min_value = max_value - 1.0
    elif max_value == None:
        max_value = min_value + 1.0

    # Apply explicit range if defined in graph
    explicit_min_value, explicit_max_value = graph_definition["explicit_vertical_range"]
    if explicit_min_value != None:
        min_value = explicit_min_value
    if explicit_max_value != None:
        max_value = explicit_max_value

    # physical range, without extra margin or zooming
    real_range = min_value, max_value

    # An explizit range set by user zoom has always
    # precedence!
    if graph_data_range.get("vertical_range"):
        min_value, max_value = graph_data_range["vertical_range"]

    # In case the graph is mirrored, the 0 line is always exactly in the middle
    if mirrored:
        abs_limit = max(abs(min_value), max_value)
        min_value = -abs_limit
        max_value = abs_limit

    # Make sure we have a non-zero range. This avoids math errors for
    # silly graphs.
    if min_value == max_value:
        if mirrored:
            min_value -= 1
            max_value += 1
        else:
            max_value = min_value + 1

    vrange = max_value - min_value

    # Make range a little bit larger, approx by 0.5 ex. But only if no zooming
    # is being done.
    if not graph_data_range.get("vertical_range"):
        vrange_per_ex = vrange / height

        # Let displayed range have a small border
        if min_value != 0:
            min_value -= 0.5 * vrange_per_ex
        if max_value != 0:
            max_value += 0.5 * vrange_per_ex

    return real_range, vrange, min_value, max_value


# Create labels for the neccessary range
def create_vertical_axis_labels(min_value, max_value, unit, label_distance, sub_distance, mirrored):
    # round_to is the precision (number of digits after the decimal point)
    # that we round labels to.
    round_to = max(0, 3 - math.trunc(math.log10(max(abs(min_value), abs(max_value)))))

    frac, full = math.modf(min_value / sub_distance)
    if min_value >= 0:
        pos = full * sub_distance
    else:
        if frac != 0:
            full -= 1.0
        pos = full * sub_distance

    # First determine where to put labels and store the label value
    label_specs = []
    while pos <= max_value:
        pos = round(pos, round_to)

        if pos >= min_value:
            f = math.modf(pos / label_distance)[0]
            if abs(f) <= 0.00000000001 or abs(f) >= 0.99999999999:
                if mirrored:
                    label_value = abs(pos)
                else:
                    label_value = pos

                line_width = 2
            else:
                label_value = None
                line_width = 1

            label_specs.append((pos, label_value, line_width))
            if len(label_specs) > 1000:
                break # avoid memory exhaustion in case of error

        # Make sure that we increase position at least that much that it
        # will not fall back to its old value due to rounding! This once created
        # a nice endless loop.
        pos += max(sub_distance, 10**-round_to)

    # Now render the single label values. When the unit has a function to calculate
    # a graph global unit, use it. Otherwise add units to all labels individually.
    if "graph_unit" not in unit:
        return render_labels_with_individual_units(label_specs, unit)
    else:
        return render_labels_with_graph_unit(label_specs, unit)


def render_labels_with_individual_units(label_specs, unit):
    return render_labels(label_specs, unit["render"])


def render_labels_with_graph_unit(label_specs, unit):
    # Build list of real values (not 0 or None) for the graph_unit function
    # which is then calculating the graph global unit
    ignored_values = [0, None]

    values = [ l[1] for l in label_specs if l[1] not in ignored_values ]

    graph_unit, scaled_labels = unit["graph_unit"](values)

    # Merge the scaled labels into the label specifications
    for index, label_spec in enumerate(label_specs):
        if label_spec[1] not in ignored_values:
            label_specs[index] = (label_spec[0], scaled_labels.pop(0), label_spec[2])

    v_axis_attributes = render_labels(label_specs)
    v_axis_attributes["axis_label"] = graph_unit
    return v_axis_attributes


def render_labels(label_specs, render_func=None):
    max_label_length, rendered_labels = 0, []

    for pos, label_value, line_width in label_specs:
        if label_value != None:
            if label_value == 0:
                label = "0"
            else:
                if render_func:
                    label = render_func(label_value)
                else:
                    label = "%s" % label_value

                # Generally remove useless zeroes in fixed point numbers.
                # This is a bit hacky. Got no idea how to make this better...
                label = remove_useless_zeroes(label)
            max_label_length = max(max_label_length, len(label))
        else:
            label = None

        rendered_labels.append((pos, label, line_width))

    return {
        "labels"           : rendered_labels,
        "max_label_length" : max_label_length,
    }


def remove_useless_zeroes(label):
    if '.' not in label:
        return label

    return label.replace(".00 ", " ").replace(".0 ", " ")


#.
#   .--Time Axis-----------------------------------------------------------.
#   |            _____ _                     _          _                  |
#   |           |_   _(_)_ __ ___   ___     / \   __  _(_)___              |
#   |             | | | | '_ ` _ \ / _ \   / _ \  \ \/ / / __|             |
#   |             | | | | | | | | |  __/  / ___ \  >  <| \__ \             |
#   |             |_| |_|_| |_| |_|\___| /_/   \_\/_/\_\_|___/             |
#   |                                                                      |
#   +----------------------------------------------------------------------+
#   |  Computation of time axix including labels                           |
#   '----------------------------------------------------------------------'

def compute_graph_t_axis(start_time, end_time, width, step):
    # Depending on which time range is being shown we have different
    # steps of granularity

    start_time_local = time.localtime(start_time)
    start_date = start_time_local[:3] # y, m, d
    start_month = start_time_local[:2]
    start_year = start_time_local[0]

    end_time_local = time.localtime(end_time)
    end_date = end_time_local[:3]
    end_month = end_time_local[:2]
    end_year = end_time_local[0]

    time_range = end_time - start_time
    time_range_days = time_range / 86400

    label_shift = 0 # shift seconds to future in order to center it
    title_label = None
    label_distance_at_least = 0

    # TODO: Monatsname und Wochenname lokalisierbar machen
    if start_date == end_date:
        labelling = "%H:%M"
        title_label = date_human_readable(start_time)
        label_size = 5

    # Less than one week
    elif time_range_days < 7:
        labelling = "%a %H:%M"
        label_size = 9
        title_label = date_range_human_readable(start_time, end_time)

    elif time_range_days < 32 and start_month == end_month:
        labelling = "%d."
        label_size = 2.5
        title_label = date_month_human_readable(start_time)
        label_shift = 86400 / 2
        label_distance_at_least = 86400

    elif time_range_days < 370:
        if start_year == end_year:
            title_label = str(start_year)
        else:
            title_label = "%d - %d" % (start_year, end_year)

        if time_range_days < 200:
            labelling = "%d.%m."
            label_size = 5
        else:
            labelling = "%b"
            label_size = 3
            label_shift = 86400 * 15

    else:
        labelling = date_human_readable
        title_label = None
        label_size = 8

    # Now guess a nice number of labels. This is similar to the
    # vertical axis, but here the division is not done by 1, 2 and
    # 5 but we need to stick to user friendly time sections - that
    # might even not be equal in size (like months!)
    num_t_labels = int((width - 7) / label_size)
    label_distance_at_least = max(label_distance_at_least, time_range / num_t_labels)

    # Get a distribution function. The function is called with start_time end
    # end_time and outputs an iteration of label positions - tuples of the
    # form (timepos, line_width, has_label).

    # If the distance of the lables is less than one day, we have a distance aligned
    # at minutes.
    for dist_minutes, subdivision in [
          (1, 0.25),   (2, 0.5),    (5, 1),      (10, 2),       (20, 5),   (30, 5),
          (60, 10),    (120, 20),   (240, 30),   (360, 60),     (480, 60), (720, 120),
          (1440, 360), (2880, 480), (4320, 720), (5760, 720) ]:
        if label_distance_at_least <= dist_minutes * 60:
            dist_function = lambda start_time, end_time: \
                            dist_equal(dist_minutes * 60, subdivision * 60, start_time, end_time)
            break

    else:
        # Label distance less than one week? Align lables at days of week
        if label_distance_at_least <= 86400 * 7:
            dist_function = dist_week

        else:
            # Label distance less that two years?
            for months in 1, 2, 3, 4, 6, 12, 18, 24, 36, 48:
                if label_distance_at_least <= 86400 * 31 * months:
                    dist_function = lambda start_time, end_time: dist_month(months, start_time, end_time)
                    break
            else:
                # Label distance is more than 8 years. Bogus, but we must not crash
                dist_function = lambda start_time, end_time: dist_month(96, start_time, end_time)

    # Now iterate over all label points and compute the labels.
    # TODO: could we run into any problems with daylight saving time here?
    labels = []
    seconds_per_char = (end_time - start_time) / (width - 7)
    for pos, line_width, has_label in dist_function(start_time, end_time):
        if has_label:
            if type(labelling) in [ str, unicode ]:
                label = time.strftime(labelling, time.localtime(pos))
            else:
                label = labelling(pos)
        else:
            label = None

        # Should the label be centered within a range? Then add just
        # the line and shift the label with "no line" into the future
        if label_shift:
            labels.append((pos, None, line_width))
            line_width = 0
            pos += label_shift

        # Do not display label if it would not fit onto the page
        if has_label and len(label) / 3.5 * seconds_per_char > end_time - pos:
            label = None
        labels.append((pos, label, line_width))

    return {
        "labels" : labels,
        "range"  : (start_time, end_time),
        "title"  : add_step_to_title(title_label, step),
    }


def add_step_to_title(title_label, step):
    step_label = get_step_label(step)
    if title_label == None:
        return step_label
    else:
        return "%s @ %s" % (title_label, step_label)


def get_step_label(step):
    if step < 3600:
        return "%dm" % (step / 60)
    elif step < 86400:
        return "%dh" % (step / 3600)
    else:
        return "%dd" % (step / 86400)


def dist_week(start_time, end_time):
    # Shift time back to monday
    wday_of_start_time = time.localtime(start_time).tm_wday
    start_time -= 86400 * wday_of_start_time

    # Set time to 00:00:00
    align_broken = list(time.localtime(start_time))
    align_broken[3:6] = (0, 0, 0)
    align = time.mktime(align_broken)

    # Now we have an equal distance of 7 days
    for pos, line_width, has_label in dist_equal(7 * 86400, 86400, start_time, end_time):
        if pos >= start_time:
            yield (pos, line_width, has_label)


def dist_month(months, start_time, end_time):
    # Jump to beginning of month
    broken = list(time.localtime(start_time))
    broken[2:6] = (1, 0, 0, 0)
    broken[-1] = 0 # isdsts

    while True:
        pos = time.mktime(broken)
        if pos > end_time:
            break

        elif pos >= start_time:
            yield pos, 2, True

        for m in range(months):
            broken[1] += 1
            if broken[1] > 12:
                broken[1] -= 12
                broken[0] += 1
            if months > 2:
                pos = time.mktime(broken)
                yield pos, 1, False

    # Add weeks if months are low
    if months <= 2:
        for pos, line_width, has_label in dist_week(start_time, end_time):
            if line_width == 2:
                yield pos, 1, False


# These distance functions yield a sequence of useful
# points for labels plus intermediate points for lines without
# labels. These points have the label (None). The function
# returns tuples of the form (time, line), where
# line is one of 0, 1, 2 (none, thin, thick)
def dist_equal(distance, subdivision, start_time, end_time):
    # First align start_time to the next time that can be divided
    # distance, but align this at 00:00 localtime!
    align_broken = list(time.localtime(start_time))
    align_broken[3:6] = (0, 0, 0)
    align = time.mktime(align_broken)
    fract, wholes = math.modf((start_time - align) / subdivision)

    pos = align + wholes * subdivision
    if fract > 0:
        pos += subdivision

    while pos <= end_time:
        f = math.modf((pos-align)/ distance)[0]
        if f <= 0.0000001 or f >= 0.9999999:
            yield (pos, 2, True)
        else:
            yield (pos, 1, False)
        pos += subdivision


#.
#   .--HTML-Graphs---------------------------------------------------------.
#   |                      _   _ _____ __  __ _                            |
#   |                     | | | |_   _|  \/  | |                           |
#   |                     | |_| | | | | |\/| | |                           |
#   |                     |  _  | | | | |  | | |___                        |
#   |                     |_| |_| |_| |_|  |_|_____|                       |
#   |                                                                      |
#   +----------------------------------------------------------------------+
#   |  Code for rendering a graph in HTML.                                 |
#   '----------------------------------------------------------------------'

# Code paths for rendering graphs in HTML:

# View -> Painter
# View -> Icons -> Hover
# Dashboard -> Embedded Graph
# Javascript Interactive Graph
# Graph Collection
# Graph Overview
# Custom graph



def host_service_graph_popup_cmk(site, host_name, service_description):
    graph_render_options = {
        "size"          : (30, 10),
        "font_size"     : 8,
        "resizable"     : False,
        "show_controls" : False,
        "show_legend"   : False,
        "interaction"   : False,
    }

    end_time = time.time()
    start_time = end_time - 8 * 3600
    graph_data_range = {
        "time_range" : (start_time, end_time),
    }

    graph_specification = (
        "template", {
            "site"                : site,
            "host_name"           : host_name,
            "service_description" : service_description,
    })

    html.write(render_graphs_from_specification_html(graph_specification, graph_data_range, graph_render_options))



def render_graph_or_error_html(abstract_graph, graph_data_range, graph_render_options):
    try:
        return render_graph_html(abstract_graph, graph_data_range, graph_render_options)
    except Exception, e:
        return render_graph_error_html(e)


def render_graph_error_html(msg_or_exc, title=None):
    if isinstance(msg_or_exc, Exception):
        if config.debug:
            raise msg_or_exc
        else:
            import traceback
            msg = traceback.format_exc()
    else:
        msg = msg_or_exc

    if title == None:
        title = _("Cannot display graph")

    return "<div class=\"graph brokengraph\"><div class=\"title\">%s</div>" \
           "<pre>%s</pre></div>" % (html.attrencode(title), html.attrencode(msg))


# Render the complete HTML code of a graph - including its <div> container.
# Later updates will just replace the content of that container.
def render_graph_html(abstract_graph, graph_data_range, graph_render_options):
    html_code = render_graph_html_content(abstract_graph, graph_data_range, graph_render_options)

    # The ajax context will be passed back to us to the page handler ajax_graph() whenever
    # an update of the graph should be done. It must contain everything that we need to
    # create the HTML code of the graph. The entry "graph_id" will be set by the javascript
    # code since it is not known to us.
    ajax_context = {
        "definition"     : abstract_graph["definition"],
        "data_range"     : graph_data_range,
        "render_options" : add_default_render_options(graph_render_options),
    }

    return '<script language="javascript">\n' \
        'create_graph(%s, %s, %s, %s);\n' \
        '</script>\n' % (
            json.dumps(html_code),
            json.dumps(abstract_graph),
            json.dumps(add_default_render_options(graph_render_options)),
            json.dumps(ajax_context))


# Render the HTML code of a graph without its container. That is a canvas object
# for drawing the actual graph and also legend, buttons, resize handle, etc.
def render_graph_html_content(abstract_graph, graph_data_range, graph_render_options):
    graph_render_options = add_default_render_options(graph_render_options)

    css = graph_render_options["preview"] and " preview" or ""
    output = '<div class="graph%s" style="font-size: %.1fpx">' % \
                            (css, graph_render_options["font_size"])

    if graph_render_options["show_controls"]:
        output += render_graph_add_to_icon_for_popup(abstract_graph)

    v_axis_label = abstract_graph["vertical_axis"]["axis_label"]
    if v_axis_label:
        output += '<div class=v_axis_label>%s</div>' % v_axis_label

    # Add the floating elements
    output += '<div class=time>%s</div>' % (abstract_graph["time_axis"]["title"] or "")
    if graph_render_options["show_controls"] and graph_render_options["resizable"]:
        output += '<img class=resize src="images/resize_graph.png">'

    output += render_graph_title(abstract_graph, graph_render_options)
    output += render_graph_canvas(graph_render_options)

    # Note: due to "omit_zero_metrics" the graph might not have any curves
    if show_graph_legend(graph_render_options, abstract_graph):
        output += render_graph_legend(abstract_graph, graph_render_options)

    output += '</div>'
    return output


def render_graph_add_to_icon_for_popup(abstract_graph):
    icon_html = html.render_icon('menu', _('Add this graph to...'))
    element_type_name = "pnpgraph"

    # Data will be transferred via URL and Javascript magic eventually
    # to our function popup_add_element (htdocs/reporting.py)
    # argument report_name --> provided by popup system
    # further arguments:
    return html.render_popup_trigger(
        content = icon_html,
        ident = 'add_visual',
        what = 'add_visual', # URL for content will be 'ajax_popup_' + what + '.py'
        data = [element_type_name, None, abstract_graph["definition"]["specification"]],
        url_vars=[("add_type", "pnpgraph")])


def render_graph_title(abstract_graph, graph_render_options):
    # Title of the graph
    if "title" in graph_render_options:
        title = graph_render_options["title"]

    else:
        title = abstract_graph["title"]
        if graph_render_options["show_service"]:
            spec_type, spec_info = abstract_graph["definition"]["specification"]
            if spec_type == "template":
                service_description = spec_info["service_description"]
                if service_description != "_HOST_":
                    service_url = html.makeuri_contextless([
                        ("view_name", "service"),
                        ("host", spec_info["host_name"]),
                        ("service", service_description)],
                        filename="view.py")
                    title = '<a href="%s">%s</a> / ' % (service_url, service_description) + title
                host_url = html.makeuri_contextless([
                    ("view_name", "hoststatus"),
                    ("host", spec_info["host_name"])],
                    filename="view.py")
                title = '<a href="%s">%s</a> / ' % (host_url, spec_info["host_name"]) + title
    return '<div class=title>%s</div>' % title


def render_graph_canvas(graph_render_options):
    # Create canvas where actual graph will be rendered
    size = graph_render_options["size"]
    graph_width  = size[0] * html_size_per_ex
    graph_height = size[1] * html_size_per_ex
    return '<canvas style="position: relative; width: %dpx; height: %dpx;"' \
           ' width=%d height=%d></canvas>' % (graph_width, graph_height, graph_width*2, graph_height*2)


# Render legend that describe the metrics
def render_graph_legend(abstract_graph, graph_render_options):
    graph_width  = graph_render_options["size"][0] * html_size_per_ex
    font_size_style = "font-size: %dpx;" % graph_render_options["font_size"]

    scalars = get_scalars(abstract_graph)

    output = '<table class=legend style="width:%dpx;">' % (graph_width - 49 - 5) # minus margin

    # Render the title row
    output += '<tr><th></th>'
    for scalar, title, inactive in scalars:
        classes = ["scalar", scalar]
        if inactive and abstract_graph["step"] != 60:
            descr = _("This graph is based on data consolidated with the function \"%s\". The "
                      "values in this column are the \"%s\" values of the \"%s\" values "
                      "aggregated in %s steps. Assuming a check interval of 1 minute, the %s "
                      "values here are based on the %s value out of %d raw values.") % \
                                  (abstract_graph["definition"]["consolidation_function"],
                                   scalar,
                                   abstract_graph["definition"]["consolidation_function"],
                                   get_step_label(abstract_graph["step"]),
                                   scalar,
                                   abstract_graph["definition"]["consolidation_function"],
                                   (abstract_graph["step"]/60))

            descr += "\n\n" + _("Click here to change the graphs "
                       "consolidation function to \"%s\".") % scalar

            classes.append("inactive")
        else:
            descr        = ""

        output += '<th class="%s" style="%s" title=\"%s\">%s</th>' % \
                        (" ".join(classes), font_size_style, html.attrencode(descr), title)
    output += '</tr>'

    # Render the curve related rows
    for curve in graph_curves(abstract_graph):
        output += '<tr>'
        output += '<td style="%s">%s ' % (font_size_style, render_color_icon(curve["color"]))
        output += '%s</td>' % curve["title"]

        for scalar, title, inactive in scalars:
            if scalar == "pin" and not show_pin_time(abstract_graph):
                continue

            if inactive and abstract_graph["step"] != 60:
                inactive_cls = " inactive"
            else:
                inactive_cls = ""

            output += '<td class="scalar%s" style="%s">%s</td>' % \
                        (inactive_cls, font_size_style, curve["scalars"][scalar][1])

        output += '</tr>'

    # Render scalar values
    if abstract_graph["horizontal_rules"]:
        first = True
        for value, readable, color, title in abstract_graph["horizontal_rules"]:
            output += '<tr class="scalar%s">' % (first and " first" or "")
            output += '<td>'
            output += render_color_icon(color)
            output += '%s</td>' % title
            output += '<td colspan=5 class=scalar>%s</td>' % readable
            output += '</tr>'
            first = False

    output += '</table>'
    return output


def get_scalars(abstract_graph):
    scalars = []
    for scalar, title in [
            ("min",     _("Minimum")), ("max",     _("Maximum")),
            ("average", _("Average")),
        ]:

        consolidation_function = abstract_graph["definition"]["consolidation_function"]
        inactive = consolidation_function != None and consolidation_function != scalar

        scalars.append((scalar, title, inactive))

    scalars.append(("last", _("Last"), False))

    if show_pin_time(abstract_graph):
        scalars.append(("pin", render_pin_time_label(abstract_graph), False))

    return scalars


def graph_curves(abstract_graph):
    curves = []
    for curve in abstract_graph["curves"][::-1]:
        if not curve.get("dont_paint"):
            curves.append(curve)
    return curves


def show_graph_legend(graph_render_options, abstract_graph):
    return graph_render_options["show_legend"] and abstract_graph["curves"]


# Estimates the height of the graph legend in pixels
def graph_legend_height_ex(graph_render_options, abstract_graph):
    if not show_graph_legend(graph_render_options, abstract_graph):
        return 0

    num_lines = 3 # header line + spacing
    num_lines += len(graph_curves(abstract_graph))
    num_lines += len(abstract_graph["horizontal_rules"])

    return num_lines


def ajax_graph():
    try:
        context  = json.loads(html.var("context"))
        graph_data_range = context["data_range"]
        graph_render_options = context["render_options"]
        graph_definition = context["definition"]

        modified = False

        if html.has_var("start_time"):
            start_time   = float(html.var("start_time"))
            end_time     = float(html.var("end_time"))
            step         = float(html.var("step"))
            modified     = True
        else:
            start_time, end_time = graph_data_range["time_range"]
            step                 = graph_data_range["step"]

        if html.has_var("resize_x"): # then has always also resize_y
            size_x   = max(min_resize_width, float(html.var("resize_x")) / html_size_per_ex + context["render_options"]["size"][0])
            size_y   = max(min_resize_height, float(html.var("resize_y")) / html_size_per_ex + context["render_options"]["size"][1])
            size     = size_x, size_y
            modified = True
            config.save_user_file("graph_size", size)
        else:
            size = context["render_options"]["size"]

        if html.has_var("range_from"):
            range_from     = float(html.var("range_from"))
            range_to       = float(html.var("range_to"))
            vertical_range = (range_from, range_to)
            modified       = True
        else:
            vertical_range = None

        if html.has_var("pin"):
            save_graph_pin()

        if html.has_var("consolidation_function"):
            graph_definition["consolidation_function"] = html.var("consolidation_function")

        graph_render_options["size"] = size
        graph_data_range["time_range"] = (start_time, end_time)
        graph_data_range["vertical_range"] = vertical_range
        graph_data_range["step"] = step

        # Persist the current data range for the graph editor
        if graph_render_options["editing"]:
            save_user_graph_data_range(graph_data_range)

        abstract_graph = compute_abstract_graph(graph_definition, graph_data_range, graph_render_options)
        html_code = render_graph_html_content(abstract_graph, graph_data_range, graph_render_options)
        json_code = json.dumps({
            "html" : html_code,
            "graph" : abstract_graph,
            "context" : {
                "graph_id"       : context["graph_id"],
                "definition"     : graph_definition,
                "data_range"     : graph_data_range,
                "render_options" : graph_render_options,
            }
        })
        html.write(json_code)

    except Exception, e:
        logger(LOG_ERR, "Ajax call ajax_graph.py failed: %s\n%s" % (e, format_exception()))
        if config.debug:
            raise
        html.write("ERROR: %s" % e)


def save_user_graph_data_range(graph_data_range):
    config.save_user_file("graph_range", graph_data_range)


def load_user_graph_data_range():
    return config.load_user_file("graph_range", {
        "time_range" : (time.time() - 86400, time.time())
    })


def forget_manual_vertical_zoom():
    user_range = load_user_graph_data_range()
    if "vertical_range" in user_range:
        del user_range["vertical_range"]
        save_user_graph_data_range(user_range)


def show_pin_time(abstract_graph):
    timestamp = abstract_graph["pin_time"]
    return timestamp != None and timestamp >= abstract_graph["start_time"] \
           and timestamp <= abstract_graph["end_time"]


def render_pin_time_label(abstract_graph):
    timestamp = abstract_graph["pin_time"]
    return date_human_readable(timestamp) + " " \
           + time.strftime("%H:%M", time.localtime(timestamp))


# TODO: We must not use service_perf_data for determining the
# available list of graphs but need to get to known which metrics
# are available from the RRD files lying around. For that we need
# a new livestatus column. With the current implementation we wont
# get graphs e.g. for network interfaces that are currently down.

# IDEE (1): Damit wir das machen können, brauchen wir eine Benennung
# von RRDs, welche beim Enkodieren keine Daten verlieren. Dabei
# könnte man z.B. auf eine Schema wie bei URL-encode zurückgreifen,
# also %20 für Space.
# IDEE (2): Wir verlangen von Perf-Daten, dass sie - wenn sie das
# Metriksystem verwenden wollen, keine Zeichen enthalten, die von
# PNP in Unterstriche umgewandelt werden. Dann können wir die
# bisherigen Namen verwenden. Wir machen eine Livestatus-Spalte,
# welche die Liste aller Metriken anhand der RRD-Namen zurückgibt:
# host_rrd_metrics, service_rrd_metrics. Damit diese effizient
# berechnet werden kann, macht der CMC einen Cache pro Objekt.
# Immer wenn eine neue RRD angelegt wird, dann wird der Name dann
# an den Cache angehängt, so dass er aktuell bleibt, ohne dass wir
# ständig nachgucken müssen. Sollten wir dann nicht rrddata in
# rrd_data umbenennen? Oder wir lassen es bei host_rrddata und
# machen host_metrics. Datentyp: Liste von Strings.
def find_possible_graphs(row):
    if "service_check_command" in row:
        perf_data_string    = row["service_perf_data"]
        historic_metrics    = row["service_metrics"] # varnames with RRDs
        check_command       = row["service_check_command"]
    else:
        perf_data_string    = row["host_perf_data"]
        historic_metrics    = row["host_metrics"] # varnames with RRDs
        check_command       = row["host_check_command"]

    # If we have no RRD files then we cannot paint any graph :-(
    if not historic_metrics:
        return None, None, None, []

    # We take the information about which metrics are present and about the scalars
    # from both the current perf data (for the scalars) and the knowledge about
    # the existing RRDs (in case the case is currently not being able to output
    # perf data).
    fake_perf_data_string = compute_fake_perf_data(historic_metrics)

    perf_data, check_command = parse_perf_data(perf_data_string, check_command)
    hist_perf_data, check_command = parse_perf_data(fake_perf_data_string, check_command)
    if hist_perf_data == None and perf_data == None:
        return None, None, None, []

    elif perf_data == None:
        perf_data = hist_perf_data
        perf_data_string = fake_perf_data_string

    else:
        current_variables = [ x[0] for x in perf_data ]
        for entry in hist_perf_data:
            if entry[0] not in current_variables:
                perf_data.append(entry)
                perf_data_string += " " + "%s=1" % entry[0]

    translated_metrics = translate_metrics(perf_data, check_command)
    return perf_data_string, translated_metrics, check_command, get_graph_templates(translated_metrics)


def compute_fake_perf_data(historic_metrics):
    parts = []
    for var_name in historic_metrics:
        if " " in var_name:
            parts.append("\"%s\"=1" % var_name)
        else:
            parts.append("%s=1" % var_name)
    return " ".join(parts)


def render_graphs_from_specification_html(graph_specification, graph_data_range, graph_render_options, show_timeranges=False):
    try:
        graph_definitions = create_graph_definitions_from_specification(graph_specification)
    except livestatus.MKLivestatusNotFoundError:
        return render_graph_error_html(_("Cannot fetch data via Livestatus"),
                                       _("Cannot calculate graph definitions"))
    except Exception, e:
        return render_graph_error_html(e, _("Cannot calculate graph definitions"))

    return render_graphs_from_definitions(graph_definitions, graph_data_range, graph_render_options, show_timeranges)


def render_graphs_from_definitions(graph_definitions, graph_data_range, graph_render_options, show_timeranges=False):
    # Estimate step. Step is the number of seconds each fetched data point represents.
    # It does not make sense to fetch the data in *much* greater precision than our
    # display has. A *bit* more precision is useful for better optical zoom.
    graph_data_range.setdefault("step", estimate_graph_step_for_html(graph_data_range["time_range"], graph_render_options))

    output = ""
    for graph_number, graph_definition in enumerate(graph_definitions):
        try:
            abstract_graph = compute_abstract_graph(graph_definition, graph_data_range, graph_render_options)
            main_graph_html = render_graph_or_error_html(abstract_graph, graph_data_range, graph_render_options)
            if show_timeranges:
                output += "<div class=\"graph_with_timeranges\">"
                output += main_graph_html
                output += render_time_range_selection(graph_definition)
                output += "</div>"
            else:
                output += main_graph_html

        except livestatus.MKLivestatusNotFoundError:
            output += render_graph_error_html(_("Cannot fetch data via Livestatus"),
                                              _("Cannot create graph"))

        except Exception, e:
            output += render_graph_error_html(e, _("Cannot create graph"))
    return output


def render_time_range_selection(graph_definition):

    now = int(time.time())
    times = [
        (4*60*60,      _("The last 4 hours")),
        (25*60*60,     _("The last 25 hours")),
        (31*24*60*60,  _("The last 31 days")),
        (365*24*60*60, _("The last 365 days")),
    ]

    output = "<table class=timeranges>"
    for num, (duration, title) in enumerate(times):
        graph_render_options = {
            "size"            : (20, 4),
            "font_size"       : 8,
            "onclick"         : "change_graph_timerange(graph, %d)" % duration,
            "fixed_timerange" : True, # Do not follow timerange changes of other graphs
            "title"           : title,
            "show_legend"     : False,
            "show_controls"   : False,
            "preview"         : True,
            "resizable"       : False,
            "interaction"     : False,
        }

        time_range = now - duration, now
        graph_data_range = {
            "time_range" : time_range,
            "step" : 2 * estimate_graph_step_for_html(time_range, graph_render_options),
        }

        output += "<td title=\"%s\">\n" % (_("Change graph timerange to: %s") % title)
        abstract_graph = compute_abstract_graph(graph_definition, graph_data_range, graph_render_options)
        output += render_graph_html(abstract_graph, graph_data_range, graph_render_options)
        output += "\n</td>"
        output += "</tr><tr>"
    output += "</table>"
    return output


def estimate_graph_step_for_html(time_range, graph_render_options):
    graph_render_options = add_default_render_options(graph_render_options)
    width_in_ex = graph_render_options["size"][1]
    steps_per_ex = html_size_per_ex * 4
    number_of_steps = width_in_ex * steps_per_ex
    return int((time_range[1] - time_range[0]) / number_of_steps)


#.
#   .--Graph Image---------------------------------------------------------.
#   |     ____                 _       ___                                 |
#   |    / ___|_ __ __ _ _ __ | |__   |_ _|_ __ ___   __ _  __ _  ___      |
#   |   | |  _| '__/ _` | '_ \| '_ \   | || '_ ` _ \ / _` |/ _` |/ _ \     |
#   |   | |_| | | | (_| | |_) | | | |  | || | | | | | (_| | (_| |  __/     |
#   |    \____|_|  \__,_| .__/|_| |_| |___|_| |_| |_|\__,_|\__, |\___|     |
#   |                   |_|                                |___/           |
#   +----------------------------------------------------------------------+
#   | It is possible to render a graph as PNG image. This is needed for    |
#   | the graphs sent with mail notifications.                             |
#   '----------------------------------------------------------------------'

# Provides a json list containing base64 encoded PNG images of the current 24h graphs
# of a host or service.
def ajax_graph_images(only_from_local=False):
    graphs = []

    if only_from_local and html.remote_ip() not in [ "127.0.0.1", "::1" ]:
        raise MKUnauthenticatedException(_("You are not allowed to access this page."))

    config.login_super_user()

    try:
        host_name = html.var("host")
        if not host_name:
            raise MKGeneralException(_("Missing mandatory \"host\" parameter"))

        service_description = html.var("service", "_HOST_")

        site = html.var("site")
        # FIXME: We should really enforce site here. But it seems that the notification context
        # has no idea about the site of the host. This could be optimized later.
        #if not site:
        #    raise MKGeneralException("Missing mandatory \"site\" parameter")
        try:
            row = get_graph_data_from_livestatus(site, host_name, service_description)
        except livestatus.MKLivestatusNotFoundError:
            if config.debug:
                raise
            raise Exception(_("Cannot render graph: host %s, service %s not found.") % (host_name, service_description))

        site = row["site"]

        # Always use 25h graph in notifications
        end_time = time.time()
        start_time = end_time - (25 * 3600)
        graph_data_range = {
            "time_range" : (start_time, end_time),
        }

        graph_render_options = {
            "font_size"        : 11,
            "resizable"        : False,
            "show_controls"    : False,
            "show_controls"    : False,
            "show_service"     : False,
            "interaction"      : False,
            "size"             : (80, 30), # ex
            # Specific for PDF rendering.
            "canvas_color"     : (1.0, 1.0, 1.0),
            "background_color" : (0.95, 0.93, 0.91),
            "color_gradient"   : 20.0,
            "show_title"       : True,
            "border_width"     : 0.05,
        }
        graph_render_options = add_default_render_options(graph_render_options)

        font_size = graph_render_options["font_size"]
        mm_per_ex = font_size / 3.0
        width_mm = graph_render_options["size"][0] * mm_per_ex

        graph_specification = (
            "template", {
                "site"                : site,
                "host_name"           : host_name,
                "service_description" : service_description,
                "graph_index"         : None, # all graphs
        })

        graph_definitions = create_graph_definitions_from_specification(graph_specification)

        # TODO Move the pdf graph rendering to metric or another place. The import
        # of the reporting module here should not be necessary.
        import reporting

        graph_data_range = reporting.compute_pdf_graph_data_range(width_mm, start_time, end_time)

        for graph_definition in graph_definitions:
            abstract_graph = compute_abstract_graph(graph_definition,
                                                    graph_data_range, graph_render_options)
            graph_png = render_graph_image(abstract_graph, graph_data_range, graph_render_options)

            graphs.append(base64.b64encode(graph_png))

        html.write(json.dumps(graphs))

    except Exception, e:
        logger(LOG_ERR, "Call to ajax_graph_images.py failed: %s\n%s" % (e, format_exception()))
        if config.debug:
            raise


def render_graph_image(abstract_graph, graph_data_range, graph_render_options):
    # TODO Maybe move the pdf graph rendering to metric or another place. The import
    # of the reporting module here should not be necessary.
    import reporting

    width_ex, height_ex = graph_render_options["size"]
    mm_per_ex = reporting.get_mm_per_ex(graph_render_options["font_size"])

    html.pdf_document = pdf.Document(
        font_family = "Helvetica",
        font_size   = graph_render_options["font_size"],
        lineheight  = 1.2,
        pagesize    = (width_ex * mm_per_ex, height_ex * mm_per_ex),
        margins     = (0, 0, 0, 0)
    )

    reporting.render_graph_pdf(abstract_graph, graph_data_range, graph_render_options,
                               pos_left=0.0, pos_top=0.0, total_width=(width_ex * mm_per_ex),
                               total_height=(height_ex * mm_per_ex))

    pdf_graph = html.pdf_document.end(do_send=False)
    #file("/tmp/x.pdf", "w").write(pdf_graph)
    return pdf.pdf2png(pdf_graph)


# TODO Might be used for a dedicated graph image web service later
#from PIL import Image, ImageDraw, ImageFont
#def graph_image():
#    html.set_content_type("image/png")
#    html.set_http_header("Content-Disposition", "inline; filename=\"graph_image.png\"")
#
#    try:
#        # ...
#    except Exception, e:
#        logger(LOG_ERR, "Call to graph_image.py failed: %s\n%s" % (e, format_exception()))
#        if config.debug:
#            raise
#        show_error_image(e)
#
#
#def show_error_image(e):
#    try:
#        width  = int(html.var("width", 600))
#        height = int(html.var("height", 150))
#    except ValueError:
#        width, height = 600, 150
#
#    image = Image.new("RGBA", (width, height), (255, 255, 255))
#    draw = ImageDraw.Draw(image)
#    font = ImageFont.load_default().font
#
#    char_width = font.getsize("x")[0]
#    if char_width == 0:
#        char_width = 10
#
#    margin = offset = 10
#    for line in textwrap.wrap(format_exception(), width=width/char_width):
#        draw.text((margin, offset), line, (0, 0, 0), font=font)
#        offset += font.getsize(line)[1]
#
#    buf = cStringIO.StringIO()
#    image.save(buf, "PNG")
#    html.write(buf.getvalue())

#.
#   .--Graph-Pin-----------------------------------------------------------.
#   |            ____                 _           ____  _                  |
#   |           / ___|_ __ __ _ _ __ | |__       |  _ \(_)_ __             |
#   |          | |  _| '__/ _` | '_ \| '_ \ _____| |_) | | '_ \            |
#   |          | |_| | | | (_| | |_) | | | |_____|  __/| | | | |           |
#   |           \____|_|  \__,_| .__/|_| |_|     |_|   |_|_| |_|           |
#   |                          |_|                                         |
#   +----------------------------------------------------------------------+
#   | Users can position a pin on the graph to mark a time to show the     |
#   | shown metrics values in the legend                                   |
#   '----------------------------------------------------------------------'

def load_graph_pin():
    return config.load_user_file("graph_pin", None)

def save_graph_pin():
    try:
        pin_timestamp = int(html.var("pin"))
        if pin_timestamp == -1:
            pin_timestamp = None
    except ValueError:
        pin_timestamp = None
    config.save_user_file("graph_pin", pin_timestamp)

#.
#   .--Graph-Collections---------------------------------------------------.
#   |            ____      _ _           _   _                             |
#   |           / ___|___ | | | ___  ___| |_(_) ___  _ __  ___             |
#   |          | |   / _ \| | |/ _ \/ __| __| |/ _ \| '_ \/ __|            |
#   |          | |__| (_) | | |  __/ (__| |_| | (_) | | | \__ \            |
#   |           \____\___/|_|_|\___|\___|\__|_|\___/|_| |_|___/            |
#   |                                                                      |
#   +----------------------------------------------------------------------+
#   |  Graph collections (also known as baskets)                           |
#   '----------------------------------------------------------------------'

class GraphCollection(pagetypes.PageRenderer, pagetypes.Container, pagetypes.Overridable, pagetypes.Base):
    def __init__(self, d):
        pagetypes.Base.__init__(self, d)

    @classmethod
    def may_contain(self, element_type_name):
        return element_type_name == "pnpgraph"

    # Bei classmethod ist das self das Klassenobjekt!
    @classmethod
    def type_name(self):
        return "graph_collection"

    @classmethod
    def default_topic(self):
        return _("Metrics")

    @classmethod
    def phrase(self, what):
        return {
            "title"          : _("Graph Collection"),
            "title_plural"   : _("Graph Collections"),
            "add_to"         : _("Add to graph collection"),
            "clone"          : _("Clone Graph Collection"),
            "create"         : _("Create Graph Collection"),
            "edit"           : _("Edit Graph Collection"),
            "new"            : _("New Collection"),
        }.get(what, pagetypes.Base.phrase(what))

    @classmethod
    def builtin_pages(self):
        return {
            "default" : {
                "title"       : _("My Graphs"),
                "topic"       : _("Metrics"),
                "description" : _("A place where you can put graphs that you currently work with"),
            }
        }

    # Extra columns for list view
    def render_extra_columns(self):
        table.cell(_("Graphs"), len(self.elements()))


    # Render a graph collection, implement deletion and movement of graphs
    def render(self):
        # If the user changes something and does not own that collection
        # then we simply create a copy on demand. TODO: We could move this
        # into the base class if that concept is useful for other page types
        # as well.
        if html.transaction_valid() and not self.is_mine():
            self.clone().render()
            return

        html.header(self.page_header())
        html.begin_context_buttons()
        self.context_button_list()
        if config.may("general.reporting"):
            html.context_button(_("Export as PDF"), html.makeuri([],
                                filename="report_instant_graph_collection.py"), "report",
                                id="context_pdf_export")
        self.context_button_edit()
        html.end_context_buttons()

        if html.check_transaction():
            if html.has_var("_delete"):
                nr = int(html.var("_delete"))
                del self.elements()[nr]

            else: # move
                nr = int(html.var("_move"))
                whither = int(html.var("_whither"))
                self.move_element(nr, whither)

            self.save_user_instances()

        graph_data_range = self.get_graph_data_range()

        html.write("<table class=graph_collection>")
        max_nr = len(self.elements())-1
        for nr, element in enumerate(self.elements()):
            self.render_graph_row(element, graph_data_range, nr, max_nr)
        html.write("</table>")
        html.footer()


    def get_graph_data_range(self):
        timerange_spec = Timerange().default_value()
        return {
            "time_range" : map(int, Timerange().compute_range(timerange_spec)[0])
        }


    def render_graph_row(self, element, graph_data_range, nr, max_nr):
        html.write("<tr>")
        html.write("<td class=controls>")
        self.render_graph_controls(nr, max_nr)
        html.write("</td>")
        html.write("<td class=graph>")

        try:
            graph_specification = self.graph_specification_from_element(element)
            graph_render_options = {
                "show_service": True,
            }
            output = render_graphs_from_specification_html(graph_specification, graph_data_range,
                                                    graph_render_options, show_timeranges=True)

            if output:
                html.write(output)
            else:
                html.write(render_graph_error_html(_("Found no graph matching this definition.")))
        except KeyError:
            html.write(render_graph_error_html(
                _("This graph definition can not be rendered. Maybe it is a deprecated definition. "
                  "Please try to remove and add the graph again.\n\n%s") %
                                                html.attrencode(repr(element))))


        html.write("</td>")
        html.write("</tr>")


    def graph_specification_from_element(self, element):
        # Convert old style (where only host/service/source was allowed)
        # to new style with real graph specification
        context    = element["context"]
        parameters = element["parameters"]
        if type(parameters) == dict:
            return ("template", {
                "site" : context.get("site"),
                "host_name" : context["host"],
                "service_description" : context["service"],
                "graph_index" : parameters["source"] - 1,
            })
        else:
            return parameters


    def render_graph_controls(self, nr, max_nr):
        html.write("<td class=controls>")
        delete_url = html.makeactionuri([("_delete", nr)])
        html.icon_button(delete_url, _("Remove this graph from the collection"), "delete")
        if nr > 0:
            url = html.makeactionuri([("_move", nr), ("_whither", 0)])
            html.icon_button(url, _("Move this graph to the top"), "top")
            url = html.makeactionuri([("_move", nr), ("_whither", nr-1)])
            html.icon_button(url, _("Move this graph one position up"), "up")
        if nr < max_nr:
            url = html.makeactionuri([("_move", nr), ("_whither", nr+1)])
            html.icon_button(url, _("Move this graph one position down"), "down")
            url = html.makeactionuri([("_move", nr), ("_whither", max_nr)])
            html.icon_button(url, _("Move this graph to the end"), "bottom")
        html.write("</td>")



pagetypes.declare(GraphCollection)

#.
#   .--ValueSpecs----------------------------------------------------------.
#   |        __     __    _            ____                                |
#   |        \ \   / /_ _| |_   _  ___/ ___| _ __   ___  ___ ___           |
#   |         \ \ / / _` | | | | |/ _ \___ \| '_ \ / _ \/ __/ __|          |
#   |          \ V / (_| | | |_| |  __/___) | |_) |  __/ (__\__ \          |
#   |           \_/ \__,_|_|\__,_|\___|____/| .__/ \___|\___|___/          |
#   |                                       |_|                            |
#   +----------------------------------------------------------------------+
#   |  Various ValueSpec for editing graphs                                |
#   '----------------------------------------------------------------------'

class UnitChoice(DropdownChoice):
    def __init__(self, **args):
        DropdownChoice.__init__(self,
            title = _("Unit"),
            choices = self._unit_choices,
            **args)

    def _unit_choices(self):
        return sorted([
            (name, info.get("description", info["title"])) for
             (name, info) in unit_info.items()],
             cmp = lambda a,b: cmp(a[1], b[1]))


class MetricChoice(ValueSpec):
    def __init__(self, **args):
        ValueSpec.__init__(self, **args)


    def render_input(self, varprefix, value):
        if html.has_var(varprefix + "_host"):
            value = self.from_html_vars(varprefix)

        host_name, service_description, metric_name, consolidation_function = value
        html.write("<div id=metricchoice_%s class=metricchoice>" % varprefix)
        self.show_metric_choice(varprefix, host_name, service_description,
                                metric_name, consolidation_function, not_if_empty=False)
        html.write("</div>")


    def from_html_vars(self, varprefix):
        return (html.var(varprefix + "_host"),
                html.var_utf8(varprefix + "_service"),
                html.var(varprefix + "_metric"),
                self.vs_consolidation_function().from_html_vars(varprefix + "_cf"))


    def validate_value(self, value, varprefix):
        host_name, service_description, metric_name, consolidation_function = value
        if not service_description:
            raise MKUserError(varprefix + "_service", _("Please select a service"))
        if not metric_name:
            raise MKUserError(varprefix + "_metric", _("Please select a metric"))


    @staticmethod
    def show_metric_choice(varprefix, host_name, service_description,
                           metric_name, consolidation_function, not_if_empty):

        on_change = "metric_choice_changed('%s', this);" % varprefix

        service_entries = MetricChoice.find_host_services(host_name)
        service_choices = [
             (description, description != "_HOST_" and description or _("Host Metrics"))
             for description, check_command, perfvars in service_entries
             if perfvars ]

        if service_description not in [ c[0] for c in service_choices ]:
            service_description = ""

        if not service_choices:
            service_choices = [ ("", _("(No services with metrics found)")) ]
            if not_if_empty:
                return

        elif not service_description:
            service_choices = [ ("", _("(Please select a service)")) ] + service_choices

        html.text_input(varprefix + "_host", host_name, size=16, attrs={"onkeyup" : on_change})

        html.sorted_select(varprefix + "_service", service_choices, service_description, onchange=on_change)

        perfvars = dict([(e[0], e[2]) for e in service_entries]).get(service_description, None)
        if perfvars == None:
            metrics_choices = [ ("", _("(Please select a service)")) ]
        elif perfvars == []:
            metrics_choices = [ ("", _("(This service has no metrics)")) ]
        else:
            check_command = [ ck for (desc, ck, pv) in service_entries if desc == service_description ][0]
            metrics_choices = MetricChoice.metric_choices(check_command, perfvars)
            if metric_name not in [ c[0] for c in metrics_choices]:
                metric_name = "" # This metric does not exist for this service
            if not metric_name:
                metrics_choices = [ ("", _("(Please select a metric)")) ] + metrics_choices
        html.sorted_select(varprefix + "_metric", metrics_choices, metric_name, onchange=on_change)
        MetricChoice.vs_consolidation_function(on_change = on_change).render_input(varprefix + "_cf", consolidation_function)


    def value_to_text(self, value):
        host_name, service_description, metric_name, consolidation_function = value
        metric_title = metric_info.get(metric_name, {}).get("title", metric_name)
        cf_title = MetricChoice.vs_consolidation_function().value_to_text(consolidation_function)
        return u"%s ▸ %s ▸ %s ▸ %s" % (host_name, service_description, metric_title, cf_title)


    @staticmethod
    def vs_consolidation_function(on_change=""):
        choices = [
            ( "max", _("Maximum"), ),
            ( "min", _("Minimum"), ),
            ( "average", _("Average"), ),
        ]
        return DropdownChoice(choices = choices, on_change = on_change)


    @staticmethod
    def metric_choices(check_command, perfvars):
        choices = []
        for nr, perfvar in enumerate(perfvars):
            translated = perfvar_translation(nr, perfvar, check_command)
            name = translated["name"]
            mi = metric_info.get(name, {})
            choices.append((name, mi.get("title", name.title())))
        return choices


    @staticmethod
    def find_host_services(host_name):
        # TODO: site hint!
        # Also fetch host data with the *same* query. This saves one round trip. And head
        # host has at least one service
        response = []
        host_check_command = None
        for service_description, check_command, metrics, host_check_command, host_metrics in \
               html.live.query(
                "GET services\n"
                "Columns: description check_command metrics host_check_command host_metrics \n"
                "Filter: host_name = %s\n" % host_name.replace("\n", "")):

            response.append((service_description, check_command, metrics))
        if host_check_command:
            response.append(("_HOST_", host_check_command, host_metrics))
        return response


    @staticmethod
    def page_ajax_metric_choice():
        varprefix = html.var("varprefix")
        host_name = html.var("host")
        service_description = html.get_unicode_input("service")
        metric_name = html.var("metric")
        consolidation_function = html.var("cf")
        MetricChoice.show_metric_choice(varprefix, host_name, service_description,
                                        metric_name, consolidation_function, not_if_empty=True)

#.
#   .--Custom Graphs-------------------------------------------------------.
#   |                   ____          _                                    |
#   |                  / ___|   _ ___| |_ ___  _ __ ___                    |
#   |                 | |  | | | / __| __/ _ \| '_ ` _ \                   |
#   |                 | |__| |_| \__ \ || (_) | | | | | |                  |
#   |                  \____\__,_|___/\__\___/|_| |_| |_|                  |
#   |                                                                      |
#   |                    ____                 _                            |
#   |                   / ___|_ __ __ _ _ __ | |__  ___                    |
#   |                  | |  _| '__/ _` | '_ \| '_ \/ __|                   |
#   |                  | |_| | | | (_| | |_) | | | \__ \                   |
#   |                   \____|_|  \__,_| .__/|_| |_|___/                   |
#   |                                  |_|                                 |
#   '----------------------------------------------------------------------'

class CustomGraph(pagetypes.PageRenderer, pagetypes.Overridable, pagetypes.Container, pagetypes.Base):
    def __init__(self, d):
        pagetypes.Base.__init__(self, d)


    # Bei classmethod ist das self das Klassenobjekt!
    @classmethod
    def type_name(self):
        return "custom_graph"


    @classmethod
    def default_topic(self):
        return _("Metrics")


    @classmethod
    def phrase(self, what):
        return {
            "title"          : _("Custom Graph"),
            "title_plural"   : _("Custom Graphs"),
            "clone"          : _("Clone Custom Graph"),
            "create"         : _("Create Custom Graph"),
            "edit"           : _("Edit Custom Graph"),
            "new"            : _("New Graph"),
        }.get(what, pagetypes.Base.phrase(what))


    @classmethod
    def sanitize(self, d):
        d.setdefault("metrics", [])
        d.setdefault("graph_options", {})
        d["graph_options"].setdefault("unit", "")
        d["graph_options"].setdefault("consolidation_function", None)
        d["graph_options"].setdefault("explicit_vertical_range", (None, None))


    # Extra columns for list view
    def render_extra_columns(self):
        table.cell(_("Metrics"))
        for metric in self._["metrics"]:
            html.write(render_color_icon(metric["color"]))
        table.cell(_("Unit"), unit_info[self._["graph_options"]["unit"]]["description"])


    @classmethod
    def custom_list_buttons(self, instance):
        if instance.may_edit():
            html.icon_button(instance.graph_designer_url(), _("Design This Graph"), "custom_graph")


    def graph_designer_url(self):
        return html.makeuri_contextless([("graph", self.name())], filename="custom_graph_design.py")


    @classmethod
    def new_custom_graph(self):
        self.need_overriding_permission("edit")
        owner = config.user_id
        graph_name = self.default_name()
        title = graph_name.replace("_", " ").title()
        new_page_dict = {
            "owner": config.user_id,
            "title" : title,
            "name" : graph_name,
            "topic": self.default_topic(),
        }
        new_page = self(new_page_dict)
        self.add_page(new_page)
        self.save_user_instances()
        return self.find_page(graph_name)


    # .--------------------------------------------------------------------.
    # | ELEMENT ACCESS                                                     |
    # '--------------------------------------------------------------------'

    def graph_options(self):
        return self._["graph_options"]


    def is_empty(self):
        return not self._["metrics"]


    # .--------------------------------------------------------------------.
    # | RENDERING THE GRAPH                                                |
    # '--------------------------------------------------------------------'

    # Render a custom graph
    def render(self):
        html.header(self.page_header())
        html.begin_context_buttons()
        self.context_button_list()
        self.context_button_edit()
        html.context_button(_("Design"), self.graph_designer_url(), "custom_graph")
        html.end_context_buttons()
        self.show_graph(show_timeranges=True)
        html.footer()


    def show_graph(self, show_timeranges):
        # Variant, where the graph it its own explicit specification
        # graph_specification = ( "explicit", self.create_graph_recipe() )
        graph_specification = ( "custom", self.name() )
        graph_render_options = {
            "editing": True
        }

        # Page reloads due to editing operations must keep the displayed data range
        if html.is_transaction():
            graph_data_range = load_user_graph_data_range()
        else:
            graph_data_range = {
                "time_range" : (time.time() - 4 * 3600, time.time()),
            }

        html_code = render_graphs_from_specification_html(graph_specification, graph_data_range,
                 graph_render_options, show_timeranges=show_timeranges)
        html.write(html_code)



    # .--------------------------------------------------------------------.
    # | CREATE GRAPH RECIPE                                                |
    # '--------------------------------------------------------------------'

    # Our instance definition is quite similar to - but not exactly the same as -
    # a graph recipe. We could have put the recipe into a seperate dict key and
    # use it 1:1. But this makes problem with the general property editing. It also
    # would not allow us to create some abstraction between the edited thing and
    # the actual graph recipe.
    def create_graph_recipe(self):
        recipe = {
            'metrics'          : list(self.metrics_for_recipe()),
            'horizontal_rules' : [],
            'title'            : self.title(),
            'specification'    : ( "custom", self.name() ),
        }
        recipe.update(self.graph_options())
        return recipe


    def metrics_for_recipe(self):
        # We need to retranslate into the original perfvar names and scales.
        # E.g. from "fs_used" to ( "/", MB ). For that purpose we need to
        # fetch the check_command and perfvar names of the actual hosts and
        # services. This needs to be done every time the graph is rendered,
        # since these might change due to an update of Check_MK.
        retranslation_map = self.retranslation_map()
        for metric in self._["metrics"]:
            if metric["visible"]:
                new_metric = metric.copy()
                try:
                    new_metric["expression"] = self.expression_for_recipe(metric["expression"], retranslation_map)
                except KeyError:
                    continue # Skip this metric. It is currently not renderable
                yield new_metric


    @staticmethod
    def expression_for_recipe(expression, retranslation_map):
        if expression[0] == "rrd":
            # ('rrd', None, 'server123', u'Filesystem /', 'fs_used', 'max', 1.0)
            host_name, service_description, metric_name, consolidation_function = expression[1:]
            site, trans = retranslation_map[(host_name, service_description)][metric_name]
            return ( "rrd", site, host_name, service_description, trans["orig_name"], consolidation_function, trans["scale"] )
        elif expression[0] == "operator":
            return ( "operator", expression[1], map(lambda exp: CustomGraph.expression_for_recipe(exp, retranslation_map), expression[2]))
        else:
            return expression


    # Example for a retranslation map
    # {(u'server123', u'Filesystem /'): {
    #    'fs_growth' :  ('mysite',  {'orig_name': 'growth',      'scale': 12.136296296296296}),
    #    'fs_size':     ('mysite'   {'orig_name': 'fs_size',     'scale': 1048576}),
    #    'fs_trend':    ('mysite'   {'orig_name': 'trend',       'scale': 12.136296296296296}),
    #    'fs_used':     ('mysite',  {'orig_name': '_',           'scale': 1048576}),
    #    'inodes_used': ('mysite',  {'orig_name': 'inodes_used', 'scale': 1.0})},
    def retranslation_map(self):
        retranslation_map = {}
        for (host_name, service_description), (site, check_command, perfvars) in self.retranslation_info().items():
            per_service_map = {}
            for perfvar_nr, perfvar_name in enumerate(perfvars):
                trans = perfvar_translation(perfvar_nr, perfvar_name, check_command)
                per_service_map[trans["name"]] = site, { "orig_name" : perfvar_name, "scale" : trans["scale"]}
            retranslation_map[(host_name, service_description)] = per_service_map
        return retranslation_map


    def retranslation_info(self):
        info = {}
        html.live.set_prepend_site(True)
        for site, host_name, service_description, check_command, perfvars in html.live.query(self.service_retranslation_lql()):
            info[(host_name, service_description)] = site, check_command, perfvars
        for site, host_name, check_command, perfvars in html.live.query(self.host_retranslation_lql()):
            info[(host_name, "_HOST_")] = site, check_command, perfvars
        html.live.set_prepend_site(False)
        return info


    def host_retranslation_lql(self):
        lql_header = "GET hosts\nColumns: host_name check_command metrics\n"

        entries = []
        for host_name in self.all_needed_hosts():
            lql = "Filter: host_name = %s\n" % host_name
            entries.append(lql)

        return lql_header + "".join(entries) + "Or: %d\n" % len(entries)


    def service_retranslation_lql(self):
        lql_header = "GET services\nColumns: host_name service_description check_command metrics\n"

        entries = []
        for host_name, service_descriptions in self.all_needed_services().items():
            lql = "Filter: host_name = %s\n" % host_name
            for service_description in service_descriptions:
                lql += "Filter: service_description = %s\n" % service_description
            lql += "Or: %d\n" % len(service_descriptions)
            lql += "And: 2\n"
            entries.append(lql)

        return lql_header + "".join(entries) + "Or: %d\n" % len(entries)


    def all_needed_hosts(self):
        needed_hosts = set() # from host to list of services
        for metric in self._["metrics"]:
            if metric["visible"]:
                self.add_needed_hosts_of_expression(needed_hosts, metric["expression"])
        return needed_hosts


    def all_needed_services(self):
        needed_services = {} # from host to list of services
        for metric in self._["metrics"]:
            if metric["visible"]:
                self.add_needed_services_of_expression(needed_services, metric["expression"])
        return needed_services


    @staticmethod
    def add_needed_hosts_of_expression(needed_hosts, expression):
        if expression[0] == "rrd":
            host_name, service_description, metric, consolidation_function = expression[1:]
            needed_hosts.add(host_name)
        elif expression[0] == "operator":
            for operand in expression[2]:
                CustomGraph.add_needed_hosts_of_expression(needed_hosts, operand)

    @staticmethod
    def add_needed_services_of_expression(needed_services, expression):
        if expression[0] == "rrd":
            host_name, service_description, metric, consolidation_function = expression[1:]
            needed_services.setdefault(host_name, []).append(service_description)
        elif expression[0] == "operator":
            for operand in expression[2]:
                CustomGraph.add_needed_services_of_expression(needed_services, operand)


    @staticmethod
    def create_graph_definitions(graph_specification):
        CustomGraph.load()
        # the specification is simply the name here
        instance = CustomGraph.find_page(graph_specification)
        if instance:
            return [ instance.create_graph_recipe() ]
        else:
            raise MKUserError(None, _("No custom graph with the name <tt>%s</tt>") % html.attrencode(graph_specification))



    # .--------------------------------------------------------------------.
    # | POPUPS FOR ADDING GRAPHS / METRICS                                 |
    # '--------------------------------------------------------------------'

    # Code for "Add this metric to custom graph...."
    @staticmethod
    def page_ajax_popup_add_metric():
        CustomGraph.load()
        html.body_start()
        html.write("<ul>")
        host_name = html.var("host")
        service_description = html.var("service")
        metric = html.var("metric")
        html.write("<li><span>%s</span></li>" % _("Add this metric to custom graph..."))

        def add_to_entry(title, graph_name):
            url = html.makeactionuri_contextless([
                ("graph", graph_name), # empty = new graph
                ("_add_metric", "1"),
                ("new_metric_host", host_name),
                ("new_metric_service", service_description),
                ("new_metric_metric", metric),
                ("new_cf", "0"),
                ], filename="custom_graph_design.py")
            html.write("<li><a href=\"%s\">" % url)
            html.write("<img src=\"images/icon_custom_graph.png\"> ")
            html.write("%s</a></li>" % title)

        for graph in CustomGraph.pages():
            add_to_entry(graph.title(), graph.name())
        add_to_entry(_("New custom graph..."), "@NEW@")

        html.write("</ul>")
        html.body_end()


    # Code for "Add this graph to custom graph..."
    @classmethod
    def render_addto_popup(self, added_type):
        if added_type != "pnpgraph":
            return

        self.render_addto_popup_title(_("Add all metrics of this graph to"))
        for page in self.pages():
            self.render_addto_popup_entry(self.type_name(), page.name(), page.title())
        self.render_addto_popup_entry(self.type_name(), "@NEW@", _("New custom graph..."))


    @classmethod
    def add_element_via_popup(self, page_name, element_type, create_info):
        self.load()
        if page_name == "@NEW@":
            page = self.new_custom_graph()
            html.set_var("graph", page.name()) # Fix URLs that we create
            created_new_graph = True
        else:
            page = self.find_page(page_name)
            created_new_graph = False

        page.add_metrics_from_other_graph(create_info, add_options = created_new_graph)
        self.save_user_instances()
        return page.graph_designer_url(), created_new_graph


    def add_metrics_from_other_graph(self, create_info, add_options):
        graph_specification = create_info["parameters"]
        if graph_specification[0] == "custom":
            self.add_metrics_from_custom_graph(graph_specification[1], add_options)
        elif graph_specification[0] == "template":
            self.add_metrics_from_template_graph(graph_specification, add_options)
        else:
            raise MKGeneralException(_("Invalid graph type '%s'") % \
                  html.attrencode(graph_specification[0]))


    def add_metrics_from_custom_graph(self, graph_name, add_options):
        other_custom_graph = self.find_page(graph_name)
        self._["metrics"] += other_custom_graph._["metrics"] # Tolerate referencing, request is soon finished
        if add_options:
            self._["graph_options"].update(other_custom_graph._["graph_options"])


    def add_metrics_from_template_graph(self, graph_specification, add_options):
        graph_definition = create_graph_definitions_from_specification(graph_specification)[0]
        other_metrics = graph_definition["metrics"]
        for other_metric in other_metrics:
            self._["metrics"].append(self.convert_graph_metric_to_custom_metric(other_metric))

        if add_options:
            self._["graph_options"].update({
                "unit" : graph_definition["unit"],
                "explicit_vertical_range" : graph_definition["explicit_vertical_range"],
            })
            self._["title"] = graph_definition["title"]


    # Convert a metric from a graph definition (using raw perfvars, scale, site hint, etc.)
    # to a metrics definition for a custom graph (using metric names, no scale and hint)
    def convert_graph_metric_to_custom_metric(self, graph_metric):
        new_metric = {}
        for key in ( "color", "title", "line_type" ):
            new_metric[key] = graph_metric[key]
        new_metric["visible"] = True
        new_metric["expression"] = self.convert_graph_expression_to_custom_expression(graph_metric["expression"])
        return new_metric


    def convert_graph_expression_to_custom_expression(self, expression):
        if expression[0] == "rrd":
            site, host_name, service_description, perfvar, consolidation_function, scale = expression[1:]
            if consolidation_function == None:
                consolidation_function = "max"
            if service_description == "_HOST_":
                check_command = self.get_check_command_of_host(site, host_name)
            else:
                check_command = self.get_check_command_of_service(site, host_name, service_description)
            # OMG: We do not have the original number of this perfvar. It might be used
            # by some arkane translations (thouh currently not being used). We assume 0
            # here. I cannot see any way how to get this number...
            translated = perfvar_translation(0, perfvar, check_command)
            metric_name = translated["name"]
            return ("rrd", host_name, service_description, metric_name, consolidation_function)
        elif expression[0] == "operator":
            return ("operator",
                expression[1],
                [ self.convert_graph_expression_to_custom_expression(operand)
                  for operand in expression[2] ])
        else:
            return expression


    def get_check_command_of_host(self, site, host_name):
        html.live.set_only_sites([site])
        check_command = html.live.query_value(
            "GET hosts\n"
            "Columns: check_command\n"
            "Filter: host_name = %s" % host_name)
        html.live.set_only_sites(None)
        return check_command


    def get_check_command_of_service(self, site, host_name, service_description):
        html.live.set_only_sites([site])
        check_command = html.live.query_value(
            "GET services\n"
            "Columns: check_command\n"
            "Filter: host_name = %s\n"
            "Filter: description = %s" % (host_name, service_description))
        html.live.set_only_sites(None)
        return check_command


#.
#   .--Graph Designer------------------------------------------------------.
#   |                      ____                 _                          |
#   |                     / ___|_ __ __ _ _ __ | |__                       |
#   |                    | |  _| '__/ _` | '_ \| '_ \                      |
#   |                    | |_| | | | (_| | |_) | | | |                     |
#   |                     \____|_|  \__,_| .__/|_| |_|                     |
#   |                                    |_|                               |
#   |               ____            _                                      |
#   |              |  _ \  ___  ___(_) __ _ _ __   ___ _ __                |
#   |              | | | |/ _ \/ __| |/ _` | '_ \ / _ \ '__|               |
#   |              | |_| |  __/\__ \ | (_| | | | |  __/ |                  |
#   |              |____/ \___||___/_|\__, |_| |_|\___|_|                  |
#   |                                 |___/                                |
#   +----------------------------------------------------------------------+
#   |  We are still int the class CustomGraph. But the class is split up   |
#   |  into two folding sections.                                          |
#   '----------------------------------------------------------------------'

    # .--------------------------------------------------------------------.
    # | VALUESPECS                                                         |
    # '--------------------------------------------------------------------'

    def vs_graph_options(self):
        def validate_range(value, varprefix):
            if not value[1] > value[0]:
                raise MKUserError(varprefix, _("Invalid vertical range: from must be lower than to"))

        return Dictionary(
            title = _("Graph Options"),
            render = "form",
            elements = [
                ( "unit",
                   UnitChoice(
                      on_change = self.on_change_javascript(),
                )),
                ( "explicit_vertical_range",
                  Alternative(
                      title = _("Explicit vertical range"),
                      orientation = "horizontal",
                      style = "dropdown",
                      elements = [
                         FixedValue(
                             (None, None),
                             title = _("Auto - adjust to actual values"),
                             totext = "",
                         ),
                         Tuple(
                             title = _("Explicit range"),
                             orientation = "horizontal",
                             show_titles = False,
                             elements = [
                                 Float(default_value = 0.0, size=9, label="%s:" % _("Lower")),
                                 Float(default_value = 1.0, size=9, label="&nbsp;%s:" % _("Upper")),
                             ],
                             validate = validate_range,
                         ),
                      ],
                      on_change=self.on_change_javascript(),
                )),
                ( "omit_zero_metrics",
                  DropdownChoice(
                      title = _("Metrics with all zero values"),
                      choices = [
                          ( True, _("Do not graph completely zero metrics") ),
                          ( False, _("Graph them anyway") ),
                      ],
                      default_value = False,
                      on_change = self.on_change_javascript(),
                )),
            ],
            optional_keys = None,
        )


    # .--------------------------------------------------------------------.
    # | MAIN ENTRY POINT                                                   |
    # '--------------------------------------------------------------------'

    @classmethod
    def _page_handlers(self, clazz):
        return {
            "custom_graph_design" : CustomGraph.page_graph_designer,
        }

    @classmethod
    def page_graph_designer(self):
        self.load()
        graph_name = html.var("graph")
        if graph_name == "@NEW@":
            page = self.new_custom_graph()
            html.set_var("graph", page.name()) # Fix URLs that we create
        else:
            page = self.find_page(graph_name)
        if not page:
            raise MKUserError("graph", _("No graph with the ID <tt>%s</tt>") % html.attrencode(graph_name))
        if not page.may_edit():
            raise MKAuthException(_("You do not have the permissions to design this graph."))
        page.graph_designer()


    def graph_designer(self):
        html.header(_("Design Graph - %s") % self.title())
        html.begin_context_buttons()
        html.context_button(_("Back"), self.list_url(), "back")
        self.context_button_edit()
        html.end_context_buttons()

        self.do_designer_actions()
        self.remove_form_variables()

        self.show_graph_designer()
        html.footer()


    def after_create_url(self):
        return self.graph_designer_url()


    # .--------------------------------------------------------------------.
    # | ACTIONS - MODIFICATIONS TO GRAPHS                                  |
    # '--------------------------------------------------------------------'

    def do_designer_actions(self):
        if html.check_transaction():
            if html.var("_clone"):
                self.action_clone_metric()
            elif html.var("_delete"):
                self.action_delete_metric()
            elif html.var("_dissolve"):
                self.action_dissolve_operation()
            elif html.var("_add_metric"):
                self.action_add_metric()
            elif html.var("_add_constant"):
                self.action_add_constant()
            elif html.var("_move"):
                self.action_move_metric()
            elif html.var("_toggle_visibility"):
                self.action_toggle_metric_visiblity()
            else:
                operator_id = self.pressed_operator_button()
                if operator_id:
                    self.action_operator(operator_id)
                else:
                    self.action_form_submitted()
            self.save_user_instances()


    def action_delete_metric(self):
        nr = int(html.var("_delete"))
        del self._["metrics"][nr]


    def action_clone_metric(self):
        nr = int(html.var("_clone"))
        cloned_metric = self._["metrics"][nr].copy()
        cloned_metric["color"] = render_color(darken_color(parse_color(cloned_metric["color"]), 0.2))
        self._["metrics"][nr+1:nr+1] = [ cloned_metric ]
        self.clear_checkboxes()


    def action_dissolve_operation(self):
        nr = int(html.var("_dissolve"))
        operation_metric = self._["metrics"][nr]
        self._["metrics"][nr:nr+1] = operation_metric["original_metrics"][::-1]
        self.clear_checkboxes()
        self.set_checkboxes(range(nr, nr+len(operation_metric["original_metrics"])))


    def action_add_metric(self):
        site = None # No site hint available
        host_name, service_description, metric_name, consolidation_function = \
            MetricChoice().from_html_vars("new_metric")
        mi = metric_info.get(metric_name, {})
        color = parse_color_into_hexrgb(mi.get("color", get_next_random_palette_color()))
        unit = mi.get("unit", "")
        # Set unit of graph when adding first metric
        if not self._["metrics"]:
            self.graph_options()["unit"] = unit
        self._["metrics"].append({
            "title"      : mi.get("title", metric_name.title()),
            "expression" : ("rrd", host_name, service_description, metric_name, consolidation_function),
            "color"      : color,
            "line_type"  : "stack",
            "visible"    : True,
        })

    def action_add_constant(self):
        value = Float().from_html_vars("new_constant")
        self._["metrics"].append({
            "title"      : _("Constant value %.2f") % value,
            "expression" : ("constant", value),
            "color"      : "#888888",
            "line_type"  : "line",
            "visible"    : True,
        })


    def action_move_metric(self):
        nr = int(html.var("_move"))
        whence = int(html.var("_whence"))
        metric = self._["metrics"][nr]
        del self._["metrics"][nr]
        self._["metrics"][whence:whence] = [ metric ]


    def action_toggle_metric_visiblity(self):
        nr = int(html.var("_toggle_visibility"))
        metric = self._["metrics"][nr]
        metric["visible"] = not metric["visible"]


    def pressed_operator_button(self):
        for operator_id in time_series_operators:
            varname = "_operator_" + operator_id
            if html.has_var(varname):
                return operator_id


    def action_operator(self, operator_id):
        selected_metrics_indices = self.get_metric_indices_from_checkboxes()
        if not selected_metrics_indices:
            raise MKUserError("select_0", _("Please select at least one metric"))
        self.clear_checkboxes()
        first_index = selected_metrics_indices[-1]

        selected_metrics = [ self._["metrics"][i] for i in selected_metrics_indices ]
        expressions = [ m["expression"] for m in selected_metrics ]
        titles = [ m["title"] for m in selected_metrics ]
        new_metric = {
            "title"            : _("%s of %s") % (time_series_operators[operator_id][0], ", ".join(titles)),
            "expression"       : ( "operator", operator_id, expressions ),
            "original_metrics" : selected_metrics, # save for later dissolve!
            "color"            : selected_metrics[0]["color"],
            "line_type"        : selected_metrics[0]["line_type"],
            "visible"          : selected_metrics[0]["visible"],
        }
        for i in selected_metrics_indices:
            del self._["metrics"][i]
        # Insert at place of first selected metric
        self._["metrics"][first_index:first_index] = [ new_metric ]


    def get_metric_indices_from_checkboxes(self):
        indices = []
        for nr in range(len(self._["metrics"])):
            varname = "select_%d" % nr
            if html.get_checkbox(varname):
                indices.append(nr)
        return indices[::-1]


    def clear_checkboxes(self):
        html.del_all_vars(prefix = "select_")


    def set_checkboxes(self, indexes):
        for i in indexes:
            html.set_var("select_%d" % i, "on")


    def action_form_submitted(self):
        self.action_modify_options()
        self.action_modify_titles()
        self.action_modify_line_types()
        self.action_modify_colors()
        self.action_modify_metric_formulas()


    def action_modify_options(self):
        new_options = self.vs_graph_options().from_html_vars("options")
        self.vs_graph_options().validate_value(new_options, "options")
        if new_options["explicit_vertical_range"] != self.graph_options()["explicit_vertical_range"]:
            forget_manual_vertical_zoom()
        self.graph_options().update(new_options)


    def action_modify_titles(self):
        for nr, metric in enumerate(self._["metrics"]):
            metric["title"] = html.var_utf8("title_%d" % nr, "Metric %d" % (nr + 1))


    def action_modify_line_types(self):
        for nr, metric in enumerate(self._["metrics"]):
            metric["line_type"] = html.var("line_type_%d" % nr, "area")


    def action_modify_colors(self):
        for nr, metric in enumerate(self._["metrics"]):
            metric["color"] = Color().from_html_vars("color_%d" % nr)


    def action_modify_metric_formulas(self):
        for nr, metric in enumerate(self._["metrics"]):
            self.action_modify_metric_formula(nr, metric)


    def action_modify_metric_formula(self, nr, metric):
        expression = metric["expression"]
        if expression[0] == "rrd":
            try:
                varprefix = "formula_%d" % nr
                new_value = MetricChoice().from_html_vars(varprefix)
                MetricChoice().validate_value(new_value, varprefix)
                metric["expression"] = ("rrd",) + new_value
            except Exception, e:
                pass # no valid value


    def remove_form_variables(self):
        # Remove all variables from previous commits so that the table of metrics
        # reflects the settings of the saved graph, not of the URL. This makes
        # difference e.g. after dissolving operations of when the graph has been
        # changed in another browser windows
        for prefix in [ "color_", "line_type_", "title_" ]:
            html.del_all_vars(prefix=prefix)


    # .--------------------------------------------------------------------.
    # | RENDERING OF THE GRAPH DESIGNERHS                                  |
    # '--------------------------------------------------------------------'

    def show_graph_designer(self):
        html.begin_form("graph", method="POST", action="custom_graph_design.py?graph=%s" % self.name())
        self.show_graph(show_timeranges=False)
        self.show_metrics()
        html.button("apply", _("Apply"), style="display: none")
        self.show_operations()
        self.show_graph_options()
        html.hidden_fields()
        html.end_form()


    def show_operations(self):
        forms.header("Metrics")
        forms.section(_("Operation on selected metrics"))
        self.show_operator_buttons()
        forms.section(_("Add new metric"))
        self.show_new_metric_generator()
        forms.section(_("Add new constant"))
        self.show_new_constant_generator()
        forms.end()


    def show_graph_options(self):
        html.del_all_vars(prefix="options_")
        self.vs_graph_options().render_input("options", self.graph_options())


    def show_metrics(self):
        table.begin(sortable=False, searchable=False, empty_text = _("No metrics in this graph yet."), css="graph_designer_metrics")
        count = len(self._["metrics"])
        for nr, metric in enumerate(self._["metrics"][::-1]):
            self.show_metric_row(count - nr - 1, count, metric)
        table.end()


    def show_metric_row(self, nr, num_metrics, metric):
        table.row()
        table.cell("", css="buttons") # Checkbox
        html.checkbox("select_%d" % nr, onclick="fix_graph_designer_operator_visibiliy(this);")

        self.show_metric_action_buttons(nr, metric)
        self.show_metric_order_buttons(nr, num_metrics)
        table.cell(_("Color"), css="narrow")
        Color(on_change=self.on_change_javascript()).render_input("color_%d" % nr, metric["color"])
        table.cell(_("Title"), css="nobr narrow")
        html.text_input("title_%d" % nr, metric["title"], size=30, submit="apply")
        self.show_metric_visible_button(nr, metric)

        table.cell(_("Style"), css="narrow")
        self.active_line_type_selector(nr, metric["line_type"])
        table.cell(_("Formula"))
        self.show_metric_formula("formula_%d" % nr, metric)


    def show_metric_visible_button(self, nr, metric):
        table.cell(_("Visible"), css="buttons")
        toggle_url = html.makeactionuri_contextless([("graph", self.name()), ("_toggle_visibility", nr)])
        html.icon_button(toggle_url, _("Toggle the visibility of this metric"),
            "snapin_switch_" + (metric["visible"] and "on" or "off"), cssclass="onoff")


    def show_metric_formula(self, varprefix, metric, is_root=True):
        metric_type = metric["expression"][0]
        if metric_type == "rrd":
            self.show_metric_type_rrd(varprefix, metric, is_root)
        elif metric_type == "operator":
            self.show_metric_type_operator(varprefix, metric)
        elif metric_type == "constant":
            self.show_metric_type_constant(metric)
        else:
            html.write(_("Unknown metric type <tt>%s</tt>") % html.attrencode(metric_type))


    def show_metric_action_buttons(self, nr, metric):
        table.cell(_("Actions"), css="buttons")
        delete_url = html.makeactionuri_contextless([("graph", self.name()), ("_delete", nr)])
        html.icon_button(delete_url, _("Delete this metric"), "delete")
        clone_url = html.makeactionuri_contextless([("graph", self.name()), ("_clone", nr)])
        html.icon_button(clone_url, _("Clone this metric"), "clone")
        if metric["expression"][0] == "operator" and "original_metrics" in metric:
            dissolve_url = html.makeactionuri_contextless([("graph", self.name()), ("_dissolve", nr)])
            html.icon_button(dissolve_url, _("Dissolve operation"), "dissolve_operation")


    def show_metric_order_buttons(self, nr, num_metrics):
        table.cell(_("Order"), css="buttons")
        def move_url(whence):
            return html.makeactionuri_contextless([("graph", self.name()), ("_move", nr), ("_whence", whence)])

        if nr > 0:
            html.icon_button(move_url(0), _("Move this metric to the bottom"), "bottom")
            html.icon_button(move_url(nr-1), _("Move this metric one position down"), "down")
        else:
            html.empty_icon_button()
            html.empty_icon_button()

        if nr < num_metrics - 1:
            html.icon_button(move_url(nr+1), _("Move this metric one position up"), "up")
            html.icon_button(move_url(num_metrics-1), _("Move this metric to the top"), "top")
        else:
            html.empty_icon_button()
            html.empty_icon_button()


    def show_metric_type_rrd(self, varprefix, metric, is_root=True):
        expression = metric["expression"]
        value = expression[1:] # host_name, service_description, metric_name, consolidation_function
        if is_root:
            MetricChoice().render_input(varprefix, value)
        else:
            html.write(MetricChoice().value_to_text(value))


    def show_metric_type_operator(self, varprefix, metric):
        operator_id, operands = metric["expression"][1:]
        operator_title = time_series_operators[operator_id][0]
        html.write("%s:<ul class=metric_operation>" % operator_title)
        if "original_metrics" in metric:
            for nr, operand in enumerate(metric["original_metrics"]):
                html.write(render_color_icon(operand["color"]))
                html.write(" ")
                self.show_metric_formula("%s_%d" % (varprefix, nr), operand, is_root=False)
                html.write("<br>")
        else:
            for nr, operand in enumerate(operands):
                fake_operand = {
                    "expression" : operand,
                }
                html.write(render_color_icon("#c0c0c0"))
                self.show_metric_formula("%s_%d" % (varprefix, nr), fake_operand, is_root=False)
                html.write("<br>")

        html.write("</ul>")


    def show_metric_type_constant(self, metric):
        html.write("%.2f" % metric["expression"][1])


    def active_line_type_selector(self, nr, line_type):
        line_type_choices = [
            ( "line", _("Line") ),
            ( "area", _("Area") ),
            ( "stack", _("Stacked Area") ),
            ( "-line", _("Mirrored Line") ),
            ( "-area", _("Mirrored Area") ),
            ( "-stack", _("Mirrored Stacked") ),
        ]
        varname = "line_type_%d" % nr
        html.select(varname, line_type_choices, deflt=line_type, onchange=self.on_change_javascript())


    def show_new_metric_generator(self):
        html.write("<div class=new_metric>")
        MetricChoice().render_input("new_metric", ("", "", "", "max"))
        html.button("_add_metric", _("Add new metric"))
        html.write("</div>")


    def show_new_constant_generator(self):
        html.write("<div class=new_constant>")
        Float().render_input("new_constant", 1.0)
        html.button("_add_constant", _("Add new constant"))
        html.write("</div>")


    def show_operator_buttons(self):
        if self.get_metric_indices_from_checkboxes():
            display = ""
            display_off = "none"
        else:
            display = "none"
            display_off = ""
        html.write("<div id=graph_designer_operators style=\"display: %s\" class=operators>" % display)
        html.write(_("Apply operation on all checked metrics: "))
        for operator, title in self.metric_operator_choices():
            html.button("_operator_" + operator, title)
        html.write("</div>")
        html.write("<div id=graph_designer_operators_off style=\"display: %s\"><i>" % display_off)
        html.write(_("Please check the metrics you want to combine by an operation"))
        html.write("</i></div>")


    @staticmethod
    def on_change_javascript():
        return "submit_graph_designer_form();"


    @staticmethod
    def metric_operator_choices():
        return [
           (op_name, op_title)
           for (op_name, (op_title, op_func))
           in time_series_operators.items()
        ]


pagetypes.declare(CustomGraph)
graph_specification_types["custom"] = CustomGraph.create_graph_definitions
