#!/usr/bin/python
# -*- encoding: utf-8; py-indent-offset: 4 -*-
# .------------------------------------------------------------------------.
# |                ____ _               _        __  __ _  __              |
# |               / ___| |__   ___  ___| | __   |  \/  | |/ /              |
# |              | |   | '_ \ / _ \/ __| |/ /   | |\/| | ' /               |
# |              | |___| | | |  __/ (__|   <    | |  | | . \               |
# |               \____|_| |_|\___|\___|_|\_\___|_|  |_|_|\_\              |
# |                                        |_____|                         |
# |             _____       _                       _                      |
# |            | ____|_ __ | |_ ___ _ __ _ __  _ __(_)___  ___             |
# |            |  _| | '_ \| __/ _ \ '__| '_ \| '__| / __|/ _ \            |
# |            | |___| | | | ||  __/ |  | |_) | |  | \__ \  __/            |
# |            |_____|_| |_|\__\___|_|  | .__/|_|  |_|___/\___|            |
# |                                     |_|                                |
# |                     _____    _ _ _   _                                 |
# |                    | ____|__| (_) |_(_) ___  _ __                      |
# |                    |  _| / _` | | __| |/ _ \| '_ \                     |
# |                    | |__| (_| | | |_| | (_) | | | |                    |
# |                    |_____\__,_|_|\__|_|\___/|_| |_|                    |
# |                                                                        |
# | mathias-kettner.com                                 mathias-kettner.de |
# '------------------------------------------------------------------------'
#  This file is part of the Check_MK Enterprise Edition (CEE).
#  Copyright by Mathias Kettner and Mathias Kettner GmbH.  All rights reserved.
#
#  Distributed under the Check_MK Enterprise License.
#
#  You should have  received  a copy of the Check_MK Enterprise License
#  along with Check_MK. If not, email to mk@mathias-kettner.de
#  or write to the postal address provided at www.mathias-kettner.de


def render_element_availability(element, instance):

    # Put context information into URL variables. This is needed for the
    # non-Livestatus-Filters to work (e.g. BIGroup)
    html.stash_vars()
    visuals.add_context_to_uri_vars(element)
    visuals.add_context_to_uri_vars(instance["report"])

    datasource_name = element["datasource"]
    datasource = views.multisite_datasources[datasource_name]

    context = instance["context"]
    filter_headers, only_sites = visuals.get_filter_headers(datasource, instance["context"])

    # Show title of this all
    parameters = element["parameters"]
    if "title" in parameters:
        title = replace_macros(instance, parameters["title"])
        html.pdf_document.add_heading(1, title, numbers=True)

    table_layout = {}
    table_layout.update(instance["options"]["table_layout"])
    table_layout.update(parameters.get("table_layout", {}))

    # Prepare availability options. Some of the options are missing,
    # because they are not editable by the reporting-user. We need to
    # fill them in in order to avoid exceptions in the availability
    # code
    avoptions = element["avoptions"].copy()
    avoptions["timelimit"] = 5 * 60 * 3600 # Limit to 5 minutes
    avoptions["range"] = instance["range"], instance["range_title"]

    # Make sure that all options are set, even if new options had
    # been defined after the reports had been created.
    for option, value in availability.get_default_avoptions().items():
        avoptions.setdefault(option, value)

    if "aggr" in datasource["infos"]:
        av_rawdata = get_bi_availability(datasource, context, avoptions)
        what = "bi"
    else:
        if "service" in datasource["infos"]:
            what = "service"
        else:
            what = "host"

        av_rawdata = availability.get_availability_rawdata(
            what,
            filter_headers,
            None, # only_sites
            None, # single object
            True, # inlude output of check plugins
            avoptions)

    av_data = availability.compute_availability(what, av_rawdata, avoptions)
    availability_tables = availability.compute_availability_groups(what, av_data, avoptions)

    show_elements = avoptions.get("elements", [ "table" ])
    if "table" in show_elements:
        for group_title, availability_table in availability_tables:
            av_table = availability.layout_availability_table(what, group_title, availability_table, avoptions)
            render_av_table(av_table, element, instance, table_layout)

        html.pdf_document.add_margin(3)

        av_levels = avoptions["av_levels"]
        if av_levels and not "omit_av_levels" in avoptions["labelling"]:
            warn, crit = av_levels
            boxes = [
                (_("OK"), pdf.green ),
                ( "> %.3f%%" % warn, None),
                (_("Warning"), pdf.yellow ),
                ( "> %.3f%%" % crit, None),
                (_("Critical"), pdf.red ),
                ( "<= %.3f%%" % crit, None),
            ]
            render_legend(_("Availability levels:"), boxes)


        if 'display_timeline_legend' in avoptions["labelling"]:
            if what == "host":
                s_boxes = [
                  (_("Up"),          "hstate0"),
                  (_("Down"),        "hstate1"),
                  (_("Unreachable"), "hstate2"),
                  (_("Flapping"),    "flapping"),
                ]
            else:
                s_boxes = [
                  (_("OK"),          "state0"),
                  (_("Warning"),     "state1"),
                  (_("Critical"),    "state2"),
                  (_("Unknown"),     "state3"),
                  (_("Flapping"),    "flapping"),
                  (_("Host down"),   "hostdown"),
                ]
            s_boxes.append((_("Downtime"), "downtime"))
            s_boxes.append((_("OO/Service"), "ooservice"))
            s_boxes.append((_("Unmonitored"), "unmonitored"))

            boxes = [ (t, pdf.css_class_colors[s]) for (t,s) in s_boxes ]
            render_legend(_("Timeline colors:"), boxes)


    if "timeline" in show_elements or "timebar" in show_elements:
        for av_entry in av_data:
            render_availability_timeline(what, av_entry, avoptions, table_layout)


    html.unstash_vars()


def render_availability_timeline(what, av_entry, avoptions, table_layout):
    title = "%s %s" % (_("Timeline of"), availability.object_title(what, av_entry))
    html.pdf_document.add_heading(2, title, numbers=False)

    timeline_rows = av_entry["timeline"]
    if not timeline_rows:
        html.pdf_document.paragraph(_("No information available"))
        return

    timeline_layout = availability.layout_timeline(what, timeline_rows, av_entry["considered_duration"], avoptions, "standalone")
    show_elements = avoptions.get("elements", [])

    if "timebar" in show_elements:
        left, top, width, height = html.pdf_document.add_canvas(html.pdf_document.width(), 6)
        tl = InlineTimeline(timeline_layout)
        tl.render(html.pdf_document, left, top, width, height, 1, 1, 1)

    if "timeline" in show_elements:
        if "timebar" in show_elements:
            html.pdf_document.add_margin(2)
        headers = [ _("From"), _("Until"), _("Duration"), _("State"), _("Last Known Plugin Output") ]
        rows = []
        for entry in timeline_layout["table"]:
            rows.append([
                ( "", entry["from_text"] ),
                ( "", entry["until_text"] ),
                ( "", entry["duration_text"] ),
                ( entry["state"], entry["state_name"] ),
                ( "", entry.get("log_output", "") ),
            ])
        html.pdf_document.add_table(headers, rows, **table_layout)


def render_legend(title, boxes):
    box_height = 4.0 # mm
    box_width = 17
    margin = 1.2
    html.pdf_document.save_state()
    html.pdf_document.set_font_zoom(0.7)
    top = html.pdf_document.line_pos()
    left = html.pdf_document.right() - box_width
    for box_label, box_color in boxes[::-1]:
        if box_color:
            html.pdf_document.render_rect(left, top - box_height - margin, box_width, box_height, line_width=0.05, fill_color = box_color)
        html.pdf_document.render_aligned_text(left, top - box_height - margin, box_width, box_height, box_label)
        left -= box_width + margin
    html.pdf_document.render_aligned_text(html.pdf_document.left(), top - box_height - margin, 500, box_height, title, align="left")
    html.pdf_document.restore_state()
    html.pdf_document.add_margin(box_height + margin)


# The BI availability needs some extra handling, since in addition to
# the actual statehist queries, one querie for the BI aggregates
# and their structure needs to be done. This is usually hacked into
# show_view(), but very hard to extract from there. So we just repeat
# the essential things here. We really need to cleanup show_view()...
def get_bi_availability(datasource, context, avoptions):
    columns = [ "aggr_tree", "aggr_name", "aggr_group" ]
    query = ""
    only_sites = None
    limit = None
    all_active_filters = [
        visuals.get_filter(filter_name)
        for filter_name in context ]

    aggr_rows = datasource["table"](columns, query, only_sites, limit, all_active_filters)

    for filter in all_active_filters:
        aggr_rows = filter.filter_table(aggr_rows)

    spans = []
    for aggr_row in aggr_rows:
        these_spans, timewarp_tree_state = availability.get_bi_spans(aggr_row["aggr_tree"], aggr_row["aggr_group"], avoptions, timewarp=None)
        spans += these_spans
    av_rawdata = availability.spans_by_object(spans)
    return av_rawdata



def render_av_table(av_table, element, instance, table_layout):
    if not av_table["rows"]:
        return # empty table

    # Construct logical table for PDF. Sorry: in contrast to
    # availability.py, here a cell is (CSS, TEXT), not (TEXT, CSS).
    headers = av_table["object_titles"]
    if "timeline" in av_table["rows"][0]:
        headers = headers + [ _("Timeline") ]
        have_timeline = True
    else:
        have_timeline = False

    headers = headers + [ t[0] for t in av_table["cell_titles"] ]
    rows = []
    for entry in av_table["rows"]:
        row = [ ("", c[0]) for c in entry["object"] ]
        if have_timeline:
            row.append(("", ("object", InlineTimeline(entry["timeline"]))))

        row += [(c[1].split(), c[0]) for c in entry["cells"]]
        rows.append(row)

    if "summary" in av_table:
        row = []
        rows.append(row)
        row.append((["heading"], _("Summary")))

        for x in range(1, len(av_table["object_titles"])):
            row.append(("", ""))
        if have_timeline:
            row.append(("", ""))
        for text, css in av_table["summary"]:
            row.append((css.split() + ["heading"], text))

    # Now construct a nice PDF table out of each group
    html.pdf_document.add_margin(2)
    if av_table["title"]:
        html.pdf_document.save_state()
        html.pdf_document.set_font_bold()
        html.pdf_document.add_text_line(av_table["title"])
        html.pdf_document.restore_state()
        html.pdf_document.add_margin(2)

    html.pdf_document.add_table(headers, rows, **table_layout)



# Object for rendering inline timeline in reports
class InlineTimeline:
    def __init__(self, timeline):
        self._segments = []
        self._total_weight = 0.0
        for nr, text, weight, statename in timeline["spans"]:
            color = pdf.css_class_colors[statename]
            self._segments.append((color, weight))
            self._total_weight += weight

    def minimal_width(self, pdfdoc):
        return 30 # mm

    def maximal_width(self, pdfdoc):
        return 50 # mm

    def can_add_dynamic_width(self):
        return True

    def height(self, pdfdoc):
        return pdfdoc.get_line_skip()

    def set_width(self, pdfdoc, width):
        pass

    # Render itself at left/top into direction right/down by width/height
    def render(self, pdfdoc, left, top, width, height, x_padding, y_padding, row_oddeven):
        if self._total_weight > 0:
            total_weight = self._total_weight
        else:
            total_weight = len(self._segments)

        segment_left = left
        for color, weight in self._segments:
            segment_width = width * weight / total_weight
            pdfdoc.render_rect(segment_left, top - height, segment_width, height, line_width=0.05, fill_color=color)
            segment_left += segment_width


def availability_to_text(element):
    parts = []
    if element["parameters"].get("title"):
        parts.append(element["parameters"]["title"])
    parts.append(_("Datasource: %s") %
                 views.multisite_datasources[element["datasource"]]["title"])
    return ", ".join(parts)


def availability_parameters(element):
    datasource_name = element["datasource"]
    datasource = views.multisite_datasources[datasource_name]
    if "service" in datasource["infos"]:
        what = "service"
    elif "aggr_name" in datasource["infos"]:
        what = "bi"
    else:
        what = "host"

    avoption_entries = [
        (e[0], e[3])
        for e in availability.get_avoption_entries(what)
        if e[2]
    ] + [
        ( "elements",
          ListChoice(
              title = _("Elements to show"),
              choices = [
                  ( "table",    _("Table with availability figures") ),
                  ( "timebar",  _("Bar graphics representing the state history") ),
                  ( "timeline", _("Detailed timeline of events for each object" ) ),
              ],
              allow_empty = False,
              default_value = [ "table" ],
        )),
    ]
    return [
        ( "parameters",
          ViewOptions(element, title = _("Table options"))
        ),
        ( "avoptions",
          Dictionary(
              title = _("Availability options"),
              elements = avoption_entries,
              optional_keys = False,
        )),
    ]

def clone_availability_from_gui(parameters, context):
    view_name = parameters["name"]
    base_view = views.get_view_by_name(view_name)
    new_element = {
        "type" : "availability",
        "parameters" : { "title" : base_view["title"] },
        "avoptions" : views.get_availability_options_from_url(get_availability_type_from_view(base_view)),
    }
    for key in [ "datasource", "single_infos", "context", ]:
        new_element[key] = copy.copy(base_view[key])

    new_element["context"].update(context)
    return new_element


element_types["availability"] = {
    "title"           : _("Availability table"),
    "render"          : render_element_availability,
    "to_text"         : availability_to_text,
    "parameters"      : availability_parameters,
    "has_context"     : True,

    # Need to collect some information before we can create the view
    "creation_steps"  : [
        # FIXME: Hier darf man nur die DS auswählen, für die AV geht
        ( "datasource",   views.DatasourceSelection(), ),
        ( "single_infos", lambda element: visuals.SingleInfoSelection(views.multisite_datasources[element["datasource"]]["infos"]) ),
    ],
    "clone_from_gui" : clone_availability_from_gui,
}

