#!/usr/bin/python
# -*- encoding: utf-8; py-indent-offset: 4 -*-
# .------------------------------------------------------------------------.
# |                ____ _               _        __  __ _  __              |
# |               / ___| |__   ___  ___| | __   |  \/  | |/ /              |
# |              | |   | '_ \ / _ \/ __| |/ /   | |\/| | ' /               |
# |              | |___| | | |  __/ (__|   <    | |  | | . \               |
# |               \____|_| |_|\___|\___|_|\_\___|_|  |_|_|\_\              |
# |                                        |_____|                         |
# |             _____       _                       _                      |
# |            | ____|_ __ | |_ ___ _ __ _ __  _ __(_)___  ___             |
# |            |  _| | '_ \| __/ _ \ '__| '_ \| '__| / __|/ _ \            |
# |            | |___| | | | ||  __/ |  | |_) | |  | \__ \  __/            |
# |            |_____|_| |_|\__\___|_|  | .__/|_|  |_|___/\___|            |
# |                                     |_|                                |
# |                     _____    _ _ _   _                                 |
# |                    | ____|__| (_) |_(_) ___  _ __                      |
# |                    |  _| / _` | | __| |/ _ \| '_ \                     |
# |                    | |__| (_| | | |_| | (_) | | | |                    |
# |                    |_____\__,_|_|\__|_|\___/|_| |_|                    |
# |                                                                        |
# | mathias-kettner.com                                 mathias-kettner.de |
# '------------------------------------------------------------------------'
#  This file is part of the Check_MK Enterprise Edition (CEE).
#  Copyright by Mathias Kettner and Mathias Kettner GmbH.  All rights reserved.
#
#  Distributed under the Check_MK Enterprise License.
#
#  You should have  received  a copy of the Check_MK Enterprise License
#  along with Check_MK. If not, email to mk@mathias-kettner.de
#  or write to the postal address provided at www.mathias-kettner.de


def render_element_view(element, instance):
    # Our URL may contain filter variables that correspond to a context.
    # Gather context from URL
    html.stash_vars()
    visuals.add_context_to_uri_vars(element)
    visuals.add_context_to_uri_vars(instance["report"])
    datasource = views.multisite_datasources[element["datasource"]]

    # Set URL variables for time filters: The report *always* has a time
    # range. That range must automatically be applied to all history
    # table, like log entries, or mkeventd-events. If these are nailed
    # down in the element, then the variables must *not* be set of course.
    for time_filter_name in datasource.get("time_filters", []):
        if not html.has_var(time_filter_name + "_from"):
            html.set_var(time_filter_name + "_from_range", "unix")
            html.set_var(time_filter_name + "_until_range", "unix")
            from_ts, until_ts = map(int, instance["range"])
            html.set_var(time_filter_name + "_from", from_ts)
            html.set_var(time_filter_name + "_until", until_ts)

    # Add a heading
    parameters = element["parameters"]
    if "title" in parameters:
        title = replace_macros(instance, parameters["title"])
        html.pdf_document.add_heading(1, title, numbers=True)

    # Make sure that all painters for ages show absolute times. The
    # format is defined in the report or global. This overrides any
    # painter options.
    views.multisite_painter_options["ts_format"]["forced_value"] = "abs"
    views.multisite_painter_options["ts_date"]["forced_value"] = instance["options"]["date_format"]

    # Get view row-limit
    if datasource.get("ignore_limit"):
        limit = None
    else:
        limit = get_view_limit(element, instance)

    # Create closure in order to transport the two arguments parameters and instance
    # through the render_function as first two arguments.
    def close_render_func(instance):
        return lambda *args: report_render_view_table(instance, *args)

    # Note: our element bare all relevant keys (like, "columns", "sorters", e.g.), that makes it a view
    # all_filters_active: needed to make filters be used which are not configured in the
    # report but the context provides information for them and enables them
    views.show_view(element, show_heading=False, show_buttons=False,
              show_footer=False, render_function=close_render_func(instance),
              only_count=False, all_filters_active=True, limit=limit)

    # Cleanup forced painter options. They may live forth in Apache cache...
    del views.multisite_painter_options["ts_format"]["forced_value"]
    del views.multisite_painter_options["ts_date"]["forced_value"]
    html.unstash_vars()

def get_view_limit(element, instance):
    if "limit" in element["parameters"]:
        limit = element["parameters"]["limit"]
    else:
        limit = instance["options"]["view_limit"]

    return min(limit, config.reporting_view_limit)


# This function is being called by show_view() after it has fetched all rows
def report_render_view_table(instance, view, rows, datasource, group_painters, painters,
                display_options, painter_options, show_heading, show_buttons,
                show_checkboxes, layout, num_columns, show_filters, show_footer,
                browser_reload):

    html.pdf_document.save_state()

    limit = get_view_limit(view, instance)
    if limit:
        if len(rows) > limit:
            rows = rows[:limit]
            message = _("This table would contain more than %d rows. Due to a configured "
                        "limit the rest of the rows has not been fetched and rendered." % limit)
            html.pdf_document.save_state()
            html.pdf_document.set_font_bold()
            html.pdf_document.set_font_color((0.8, 0.2, 0.0))
            html.pdf_document.add_paragraph(message)
            html.pdf_document.restore_state()

    # Prepare styling information
    parameters = view["parameters"] # the view is in fact our element
    table_layout = {}
    table_layout.update(instance["options"]["table_layout"])
    table_layout.update(parameters.get("table_layout", {}))

    layout = view["parameters"].get("layout", "table")
    if type(layout) == tuple: # matrix uses ("matrix", num_columns )
        layout_name = layout[0]
        num_columns = layout[1]
    else:
        layout_name = layout
        num_columns = 1
    layout_function = view_layout_renderers[layout_name]
    layout_function(rows, group_painters, painters, table_layout, num_columns, instance["range"])

    html.pdf_document.restore_state()


def render_view_table(rows, group_painters, painters, table_layout, num_columns, time_range):
    # TODO: num_columns not supported yet
    text_table = []
    for row in rows: # view sends one row more than limit!
        text_row = []
        for p in painters:
            cell_content = prepare_pdf_paint(p, row, time_range)
            if cell_content:
                text_row.append(cell_content)
        text_table.append(text_row)

    headers = prepare_column_headers(painters, use_short=True)
    html.pdf_document.add_table(headers, text_table, **table_layout)


def render_view_dataset(rows, group_painters, painters, table_layout, num_columns, time_range):
    # TODO: num_columns not supported yet
    headers = prepare_column_headers(painters, use_short=False)
    printable_painters = [ p for p in painters if p[0].get("printable", True) != False ]

    text_table = []
    for header, painter in zip(headers, printable_painters):
        text_row = [ ('leftheading', header) ]
        text_table.append(text_row)
        for row in rows:
            cell_content = prepare_pdf_paint(painter, row, time_range)
            if cell_content:
                text_row.append(cell_content)

    html.pdf_document.add_table([], text_table, **table_layout)


def render_view_matrix(rows, group_painters, painters, table_layout, num_columns, time_range):
    for groups, unique_row_ids, matrix_cells \
        in views.create_matrices(rows, group_painters, painters, num_columns):
            text_table = []

            for painter in group_painters:
                text_row = []
                text_table.append(text_row)
                text_row.append(("leftheading", painter[0]["title"]))
                for group, group_row in groups:
                    cell_content = prepare_pdf_paint(painter, group_row, time_range)
                    if cell_content:
                        text_row.append(cell_content)

            # Now for each unique service^H^H^H^H^H^H ID column paint one row
            for row_id in unique_row_ids:
                text_row = []
                text_table.append(text_row)
                cell_content = prepare_pdf_paint(painters[0], matrix_cells[row_id].values()[0], time_range)
                if cell_content:
                    text_row.append(cell_content)

                # Create a list. For each group there is one element that is a list
                # of the the print stuff of each cell.
                cell_stacks = []

                # Now go through the groups and paint the rest of the columns
                for group_id, group_row in groups:
                    cell_row = matrix_cells[row_id].get(group_id)
                    if cell_row == None:
                        cell_stacks.append([])
                    else:
                        cell_stack = []
                        cell_stacks.append(cell_stack)
                        for painter in painters[1:]:
                            cell_content = prepare_pdf_paint(painter, cell_row, time_range)
                            if cell_content:
                                cell_stack.append(cell_content)

                # Finalize the first row
                for cell_stack in cell_stacks:
                    if cell_stack:
                        text_row.append(cell_stack[0])
                    else:
                        text_row.append(("", ""))

                # Add empty rows for each additional painter (for cells that contain more
                # than one line
                for nr in range(len(painters) - 2):
                    text_row = [("", "")]
                    text_table.append(text_row)
                    for cell_stack in cell_stacks:
                        if len(cell_stack) > nr+1:
                            text_row.append(cell_stack[nr+1])
                        else:
                            text_row.append(("", ""))

            html.pdf_document.add_table([], text_table, **table_layout)


def prepare_column_headers(painters, use_short):
    header_key = use_short and "short" or "title"
    return [ p[0].get(header_key, p[0]["title"])
             for p in painters
             if p[0].get("printable", True) != False ]


# Same as prepare_paint() for HTML output: Gets a painter and a data
# row and creates either the text for being painted or a special object
# of the type PainterPrinter - which can be rendered by the table
# algorithm of the pdf module. Might also return None. In that case that
# column is not available in PDF output.
def prepare_pdf_paint(painter, row, time_range):
    if painter[0].get("printable", True) == True:
        try:
            row = views.join_row(row, painter)
            css_classes, txt = views.paint_painter(painter[0], row)
            # if type(txt) not in (str, unicode):
            #     txt = str(txt) # What is this structure?
            txt = txt.strip()

            # row_texts.append((css_classes, html.strip_tags(txt)))
            # Handle <img...>. Our PDF writer cannot draw arbitrary
            # images, but all that we need for showing simple icons.
            # Current limitation: *one* image
            if txt.lower().startswith("<img"):
                img_filename = re.sub('.*src=["\']([^\'"]*)["\'].*', "\\1", txt)
                img_path = find_htdocs_image_path(img_filename)
                if img_path:
                    txt = ("icon", img_path)
                else:
                    txt = img_filename

            txt = html.strip_tags(txt)

            return css_classes, txt
        except Exception:
            import traceback
            raise MKGeneralException('Failed to paint "%s": %s' %
                            (painter[0]['name'], traceback.format_exc()))

    elif painter[0].get("printable", True) != False:
        # Create object that is usable in the layout algorithm instead of a text
        painter_printer = painter_printers[painter[0]["printable"]](row, time_range)
        return "", ("object", painter_printer)



view_layout_renderers = {
    "table"       : render_view_table,
    "dataset"     : render_view_dataset,
    "matrix"      : render_view_matrix,
    "boxed"       : render_view_table,
    "boxed_graph" : render_view_table,
    "tiled"       : render_view_table,
}


def find_htdocs_image_path(filename):
    dirs = [
        defaults.omd_root + "/local/share/check_mk/web/htdocs/",
        defaults.omd_root + "/share/check_mk/web/htdocs/",
    ]
    for d in dirs:
        if os.path.exists(d + filename):
            return d + filename


def view_parameters(element):
    datasource_name = element["datasource"]
    return [
        ( "parameters", ViewOptions(element) )
    ] + views.view_editor_specs(datasource_name, general_properties=False)

def view_to_text(element):
    parts = []
    if element["parameters"].get("title"):
        parts.append(element["parameters"]["title"])
    parts.append(_("Datasource: %s") %
                 views.multisite_datasources[element["datasource"]]["title"])
    parts.append(_("%d columns") % len(element["painters"]))
    return ", ".join(parts)


# This function is being called after the icon "Add this view to..."
# and selecting an existing report. We create a view element as a clone
# of an existing view here
def clone_view_from_gui(parameters, context):
    view_name = parameters["name"]
    base_view = views.get_view_by_name(view_name)
    new_view = {
        "type" : "view",
        "parameters" : {
            "title" : base_view["title"],
            "layout" : base_view["layout"] == "dataset" and "dataset" or "table",
        },
    }
    for key in [ "datasource", "single_infos", "context", "sorters",
                 "group_painters", "painters" ]:
        new_view[key] = copy.copy(base_view[key])

    new_view["context"].update(context)
    return new_view


element_types["view"] = {
    "title"           : _("A Multisite view"),
    "render"          : render_element_view,
    "to_text"         : view_to_text,
    "parameters"      : view_parameters,
    "has_context"     : True,
    "transform"       : (views.transform_view_to_valuespec_value,
                         views.transform_valuespec_value_to_view,),

    # Need to collect some information before we can create the view
    "creation_steps"  : [
        ( "datasource",   views.DatasourceSelection(), ),
        ( "single_infos", lambda element: visuals.SingleInfoSelection(views.multisite_datasources[element["datasource"]]["infos"]) ),
    ],
    "clone_from_gui" : clone_view_from_gui,
}
